/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komasync.h"

KOMAsync::~KOMAsync()
{
	while (lastMessage)
	{
		AsyncMessage* pm = lastMessage->prev;
		delete lastMessage;
		lastMessage = pm;
	}
}

int KOMAsync::getEvents(KOMString& evt, int prevId)
{
	evt.clear();
	evt.appendFormat(KOM_GROUP_SEPARATOR "%d", lastEventId);
	if (prevId < nameUpdate)
	{
		evt.append(KOM_GROUP_SEPARATOR "update_name");
	}
	if (prevId < unreadUpdate)
	{
		evt.append(KOM_GROUP_SEPARATOR "update_unread");
		newTexts = false;
	}
	AsyncMessage* msg = lastMessage;
	while (msg && msg->prev && prevId < msg->eventId)
		msg = msg->prev;
	if (msg && msg->eventId <= prevId)
		msg = msg->next;
	while (msg)
	{
		evt.appendFormat(KOM_GROUP_SEPARATOR "msg" 
			KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%d" 
			KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%s" 
			KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%d", 
			msg->eventId, msg->from, msg->to, msg->text.c_string(), 
			(int)msg->timestamp, msg->read?1:0);
		msg = msg->next;
	}
	return 0;
}

void KOMAsync::updateConfName(KOMString& str)
{
	int conf;
	KOMString oldname, newname;
	// FIXME: OOM
	if (!str.parseInt(conf) || !str.parseString(oldname) || !str.parseString(newname))
		return;
	newname.latin1ToUTF8();
	KOMConferenceStat* c = komSession->getConference(conf);
	if (c)
	{
		c->setName(newname.c_string());
		nameUpdate = ++lastEventId;
	}
	komSession->updateConferenceName(conf, newname);
}

int KOMAsync::addNewText(KOMString& str)
{
	// Parse the reply to get all confs it is in, then add the text to all confs
	int textNum;
	bool res = str.parseInt(textNum);
	if (!res)
	{
		// Force an update in all caches
		needsClearCache = true;
		return 0;
	}

	if (komSession->addTextStat(textNum, true, str))
	{
		// Force an update in all caches
		needsClearCache = true;
		return 0;
	}
	if (textNum > 0 && !komSession->isIgnored(textNum))
	{
		KOMTextStat* txt = komSession->getGlobalTextStat(textNum);
		for (KOMConferenceStat* conf = komSession->getFirstConference(); conf; conf = conf->nextConference())
		{
			if (txt->getLocalNum(conf->getNum()) > 0)
			{
				conf->setNewMessages(true);
			}
		}
	}
	return textNum;
}

void KOMAsync::updateUnread()
{
	unreadUpdate = ++lastEventId;
}

void KOMAsync::addInstantMessage(int from, int to, KOMString& text)
{
	AsyncMessage* msg = new AsyncMessage();
	if (msg)
	{
		msg->to = to;
		msg->from = from;
		msg->text.append(text.c_string());
		if (msg->to == komSession->getUserID())
			msg->to = -1;
		msg->eventId = ++lastEventId;
		msg->read = from==komSession->getUserID();
		msg->timestamp = time(NULL);
		msg->next = NULL;
		msg->prev = lastMessage;
		if (lastMessage)
			lastMessage->next = msg;
		lastMessage = msg;
	}
}

void KOMAsync::markInstantMessageRead(int eventId)
{
	for (AsyncMessage* msg = lastMessage; msg; msg = msg->prev)
	{
		if (msg->eventId == eventId)
		{
			msg->read = true;
			return;
		}
	}
}

bool KOMAsync::hasUnreadInstantMsg()
{
	for (AsyncMessage* msg = lastMessage; msg; msg = msg->prev)
	{
		if (!msg->read)
			return true;
	}
	return false;
}

bool KOMAsync::progress(KOMString& reply)
{
	int params;
	int type;
	bool res;
	int confNum;
	int textNum;
	int persNum;
	if (!reply.parseAsync())
		return false;
	res = reply.parseInt(params);
	res &= reply.parseInt(type);
	if (!res)
		return true;
	switch (type)
	{
	case 5: // conf changed name
		updateConfName(reply);
		break;
	case 12: // async message
	{
		int from, to;
		KOMString text;
		res &= reply.parseInt(to);
		res &= reply.parseInt(from);
		res &= reply.parseString(text);
		text.latin1ToUTF8();
		if (res)
		{
			addInstantMessage(from, to, text);
		}
		break;
	}
	case 15: // new text
		textNum = addNewText(reply);
		unreadUpdate = ++lastEventId;
		if (textNum > 0 && !komSession->isIgnored(textNum))
			newTexts = true;
		break;
	case 14: // text deleted
		// Delete cache from all conferences
		needsClearCache = true;
		unreadUpdate = ++lastEventId;
		break;
	case 16: // new recpt
		// FIXME: might not update the highest local num correctly
		if (reply.parseInt(textNum) && reply.parseInt(confNum))
		{
			KOMConferenceStat* c = komSession->getConference(confNum);
			if (c)
				c->setNewMessages(true);
		}
		if (textNum > 0 && !komSession->isIgnored(textNum))
			newTexts = true;
		unreadUpdate = ++lastEventId;
		break;
	case 17: // rm recpt
		if (reply.parseInt(textNum) && reply.parseInt(confNum))
		{
			// FIXME: might be better to clear the unavailable flags on new
			// Clear the cache since the permissions might have changed
			needsClearCache = true;
			// Delete cache from affected conference
			unreadUpdate = ++lastEventId;
		}
		break;
	case 8: // removed subscription
		if (reply.parseInt(confNum))
		{
			komSession->deleteConference(confNum);
			// Clear the cache since the permissions might have changed
			needsClearCache = true;
			unreadUpdate = ++lastEventId;
		}
		break;
	case 18: // new subscription
		if (reply.parseInt(persNum) && reply.parseInt(confNum))
		{
			komSession->markConferenceListDirty(true);

			// FIXME: might be better to clear the unavailable flags
			// Clear the cache since the permissions might have changed
			needsClearCache = true;
			unreadUpdate = ++lastEventId;
			newTexts = true;
		}
		break;
	default:
		break;
	}

	return true;
}

void KOMAsync::clearCacheIfNeeded()
{
	if (needsClearCache)
	{
		komSession->clearTextCache();
		needsClearCache = false;
	}
}
