/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komcache.h"

KOMHash::KOMHash(int hashId) : id(hashId), numEntries(0), numSlots(1), 
	entries(&nomemEntry), nomemEntry(NULL)
{}

KOMHash::~KOMHash()
{
	if (numSlots != 1)
		delete[] entries;
}

void KOMHash::deleteAll()
{
	for (int i = 0; i < numSlots; ++i)
	{
		while (entries[i])
		{
			KOMHashEntry* ne = entries[i]->getNextHashEntry(id);
			delete entries[i];
			entries[i] = ne;
		}
	}
	numEntries = 0;
	if (numSlots > KOM_INITIAL_HASH_SIZE)
		changeSize(KOM_INITIAL_HASH_SIZE);
}

void KOMHash::clear()
{
	for (int i = 0; i < numSlots; ++i)
	{
		entries[i] = NULL;
	}
	numEntries = 0;
	if (numSlots > KOM_INITIAL_HASH_SIZE)
		changeSize(KOM_INITIAL_HASH_SIZE);
}

void KOMHash::add(KOMHashEntry* entry)
{
	if (numEntries >= numSlots*2)
	{
		// Grow the hash table
		if (numSlots == 1)
			changeSize(KOM_INITIAL_HASH_SIZE);
		else
			changeSize(numSlots*3);
	}
	int pos = entry->getHashKey(id)%numSlots;
	entry->getNextHashEntry(id) = entries[pos];
	entries[pos] = entry;
	++numEntries;
}

KOMHashEntry* KOMHash::get(int key)
{
	if (key < 0)
		return NULL;
	int pos = key%numSlots;
	KOMHashEntry* e;
	for (e = entries[pos]; e; e = e->getNextHashEntry(id))
		if (e->getHashKey(id) == key)
			return e;
	return NULL;
}

void KOMHash::remove(KOMHashEntry* entry)
{
	int pos = entry->getHashKey(id)%numSlots;
	KOMHashEntry* e;
	KOMHashEntry** nextp = &entries[pos];
	for (e = entries[pos]; e; e = e->getNextHashEntry(id))
	{
		if (e == entry)
		{
			*nextp = e->getNextHashEntry(id);
			break;
		}
		nextp = &e->getNextHashEntry(id);
	}
	if (!e)
		return;
	--numEntries;
	// Shring the hash if needed
	if (numEntries < (numSlots/3) && numSlots > KOM_INITIAL_HASH_SIZE)
	{
		// Shrink the hash table
		changeSize(numSlots/3);
	}
}

void KOMHash::changeSize(int newSize)
{
	KOMHashEntry** nent = new KOMHashEntry*[newSize];
	if (nent)
	{
		// Clear all entries to start with
		for (int i = 0; i < newSize; ++i)
			nent[i] = NULL;
		int oldSize = numSlots;
		KOMHashEntry** oldEnt = entries;
		entries = nent;
		numEntries = 0;
		numSlots = newSize;
		// Add old old entries, rehash
		for (int i = 0; i < oldSize; ++i)
		{
			while (oldEnt[i])
			{
				KOMHashEntry* nextent = oldEnt[i]->getNextHashEntry(id);
				add(oldEnt[i]);
				oldEnt[i] = nextent;
			}
		}
		if (oldSize != 1)
			delete[] oldEnt;
	}
}

KOMTextStat::KOMTextStat() : globalNum(-1), author(-1), 
		to(NULL), numTo(0), cc(NULL), numCc(0),	timeval(0), 
		commentTo(-1), firstComment(-1), nextComment(-1), 
		numComments(0), textLen(0), unavailable(false), utf8(false), 
		reformat(false)
{}

KOMTextStat::~KOMTextStat()
{
	delete[] to;
	delete[] cc;
	localNums.deleteAll();
}

int KOMTextStat::getLocalNum(int conf)
{
	LocalNum* loc = (LocalNum*)localNums.get(conf);
	return loc?loc->localNum:0;
}

KOMConferenceStat::KOMConferenceStat(int number) : refCount(1), conferenceNum(number), read(true), highestLocal(0), invite(false), secret(false), passive(false), instant_passive(false), numReadRanges(0), allocatedReadRanges(0), readRanges(NULL), firstCached(-1), oldestCached(false), new_msgs(false)
{}

KOMConferenceStat::~KOMConferenceStat()
{
	ignoredTexts.deleteAll();
	localToGlobalTextNum.deleteAll();
	delete[] readRanges;
}

int KOMConferenceStat::setName(const char* n)
{
	name.clear();
	if (n)
	{
		return name.append(n);
	}
	return 0;
}

void KOMConferenceStat::clearTextCache()
{
	localToGlobalTextNum.deleteAll();
	firstCached = -1;
	oldestCached = false;
}

void KOMConferenceStat::updateCacheRange(int first, bool foundFirst)
{
	if (firstCached < 0)
	{
		firstCached = first;
	}
	else
	{
		if (first < firstCached)
			firstCached = first;
	}
	oldestCached |= foundFirst;
}

int KOMConferenceStat::addReadText(int textNum)
{
	if (isTextRead(textNum, false))
		return 0;
	for (int i = 0; i < numReadRanges; ++i)
	{
		bool added = false;
		if (textNum == readRanges[i*2]-1)
		{
			--readRanges[i*2];
			added = true;
		}
		else if (textNum == readRanges[i*2+1]+1)
		{
			++readRanges[i*2+1];
			added = true;
		}
		if (added)
		{
			for (int j = i+1; j < numReadRanges; ++j)
			{
				bool merged = false;
				if (textNum == readRanges[j*2]-1)
				{
					// Merge the ranges
					readRanges[i*2+1] = readRanges[j*2+1];
					merged = true;
				}
				else if (textNum == readRanges[j*2+1]+1)
				{
					// Merge the ranges
					readRanges[i*2] = readRanges[j*2];
					merged = true;
				}
				if (merged)
				{
					--numReadRanges;
					for (int k = j; k < numReadRanges; ++k)
					{
						readRanges[k*2] = readRanges[(k+1)*2];
						readRanges[k*2+1] = readRanges[(k+1)*2+1];
					}
					return 0;
				}
			}
			return 0;
		}
	}
	if (numReadRanges >= allocatedReadRanges)
	{
		int* nr = new int[(allocatedReadRanges+16)*2];
		if (!nr)
			return -1;
		for (int i = 0; i < numReadRanges; ++i)
		{
			nr[i*2] = readRanges[i*2];
			nr[i*2+1] = readRanges[i*2+1];
		}
		delete[] readRanges;
		readRanges = nr;
		allocatedReadRanges += 16;
	}
	readRanges[numReadRanges*2] = textNum;
	readRanges[numReadRanges*2+1] = textNum;
	++numReadRanges;
	return 0;
}

bool KOMConferenceStat::isTextRead(int textNum, bool checkIgnoreList)
{
	if (textNum <= 0)
		return true;
	if (checkIgnoreList && isIgnored(textNum))
		return true;
	for (int i = 0; i < numReadRanges; ++i)
	{
		if (textNum >= readRanges[i*2] && textNum <= readRanges[i*2+1])
			return true;
	}
	return false;
}

int KOMConferenceStat::findNextLowerUnread(int textNum)
{
	if (textNum <= 0)
		return -1;
	for (int i = 0; i < numReadRanges; ++i)
	{
		if (textNum >= readRanges[i*2] && textNum <= readRanges[i*2+1])
		{
			int nextUnread = readRanges[i*2]-1;
			if (isIgnored(nextUnread))
				return findNextLowerUnread(nextUnread-1);
			return nextUnread;
		}
	}
	return textNum;

}

int KOMConferenceStat::addGlobalTextNum(int localNum, int globalNum)
{
	LocToGlobTextNum* newnum = (LocToGlobTextNum*)localToGlobalTextNum.get(localNum);
	if (newnum)
		newnum->globalNum = globalNum;
	else
	{
		newnum = new LocToGlobTextNum(localNum, globalNum);
		if (!newnum)
			return -1;
		localToGlobalTextNum.add(newnum);
	}
	return 0;
}

int KOMConferenceStat::getGlobalTextNum(int localNum)
{
	LocToGlobTextNum* newnum = (LocToGlobTextNum*)localToGlobalTextNum.get(localNum);
	return newnum?newnum->globalNum:0;
}


KOMCache::KOMCache() : confsDirty(true), confs(NULL), firstTextData(NULL), 
		lastTextData(NULL), textCacheSize(0), firstCachedName(NULL)
{}

KOMCache::~KOMCache()
{
	resetCache();
	favoritesCache.deleteAll();
	ignoredTexts.deleteAll();
	textData.deleteAll();
}

void KOMCache::resetCache()
{
	while (confs)
	{
		KOMConferenceStat* nc = confs->next;
		confs->decRef();
		confs = nc;
	}
	nameCache.deleteAll();
	globalTextNum.deleteAll();
	replyTranslations.deleteAll();
	
	// Restore the initial state
	confsDirty = true;
	firstCachedName = NULL;
}

KOMConferenceStat* KOMCache::getConference(int conf)
{
	for (KOMConferenceStat* c = confs; c; c = c->next)
	{
		if (c->getNum() == conf)
			return c;
	}
	return NULL;
}

void KOMCache::addConference(KOMConferenceStat* c)
{
	c->next = NULL;
	if (!confs)
	{
		confs = c;
		return;
	}
	KOMConferenceStat* pc = confs;
	while (pc->next)
		pc = pc->next;
	pc->next = c;
}

void KOMCache::deleteConference(int conf)
{
	if (confs && confs->getNum() == conf)
	{
		KOMConferenceStat* nc = confs->next;
		confs->decRef();
		confs = nc;
		return;
	}
	KOMConferenceStat* pc = confs;
	while (pc->next && pc->next->getNum() != conf)
		pc = pc->next;
	if (pc->next)
	{
		KOMConferenceStat* c = pc->next;
		pc->next = c->next;
		c->decRef();
	}
}

void KOMCache::clearTextCache()
{
	for (KOMConferenceStat* c = confs; c; c = c->next)
		c->clearTextCache();
	globalTextNum.deleteAll();
}

#define KOM_TEXT_DATA_CACHE_SIZE 10*1024*1024
int KOMCache::addTextData(KOMTextData* d)
{
	// Make sure the same text is not added twise
	if (textData.get(d->textNum))
		return 1;
	textData.add(d);
	d->prevList = NULL;
	if (firstTextData)
		firstTextData->prevList = d;
	d->nextList = firstTextData;
	firstTextData = d;
	if (!lastTextData)
		lastTextData = firstTextData;
	textCacheSize += d->text.len();
	// Clear the list from the back
	while (textCacheSize > KOM_TEXT_DATA_CACHE_SIZE && lastTextData && lastTextData != firstTextData)
	{
		KOMTextData* del = lastTextData;
		lastTextData = del->prevList;
		textCacheSize -= del->text.len();
		textData.remove(del);
		delete del;
	}
	return 0;
}

KOMTextData* KOMCache::getTextData(int text)
{
	KOMTextData* d = (KOMTextData*)textData.get(text);
	if (d && d != firstTextData)
	{
		// Move d to front of the list
		d->prevList->nextList = d->nextList; 
		if (d->nextList)
			d->nextList->prevList = d->prevList;
		else
			lastTextData = d->prevList;
		d->prevList = NULL;
		d->nextList = firstTextData;
		firstTextData->prevList = d;
		firstTextData = d;
	}
	return d;
}

const KOMString* KOMCache::getConferenceName(int id)
{
	KOMConfNameCacheEntry* ne = (KOMConfNameCacheEntry*)nameCache.get(id);
	if (!ne)
		return NULL;
	return &(ne->name);
}

void KOMCache::updateConferenceName(int id, const KOMString& name)
{
	KOMConfNameCacheEntry* ne = (KOMConfNameCacheEntry*)nameCache.get(id);
	if (ne)
	{
		ne->name.clear();
		ne->name.append(name.c_string());
	}
	else
	{
		ne = new KOMConfNameCacheEntry(firstCachedName);
		ne->num = id;
		ne->name.append(name.c_string());
		firstCachedName = ne;
		nameCache.add(ne);
	}
}

int KOMCache::addEmptyTextStat(int globalNum, int localNum, int confNum)
{
	if (globalNum < 0)
		return 1; // Invalid text
	KOMTextStat* txt = getGlobalTextStat(globalNum);
	if (!txt)
	{
		txt = new KOMTextStat();
		if (!txt)
		{
			return -1;
		}
		txt->globalNum = globalNum;

		// Do not add texts with negative values
		globalTextNum.add(txt);
	}
	if (localNum > 0 && !txt->localNums.get(confNum))
	{
		KOMTextStat::LocalNum* loc = new KOMTextStat::LocalNum();
		if (!loc)
			return -1;
		loc->confNum = confNum;
		loc->localNum = localNum;
		txt->localNums.add(loc);
		// When this happens to or cc is changed, so reload this stat
		txt->author = -1;
	}
	if (getConference(confNum))
	{
		getConference(confNum)->addGlobalTextNum(localNum, globalNum);
		if (ignoredTexts.get(globalNum))
		{
			getConference(confNum)->addIgnored(localNum);
		}
	}
	return 0;
}

int KOMCache::addTextStat(int globalNum, bool is_ok, KOMString& str)
{
	if (globalNum < 0)
		return 1; // Invalid text
	KOMTextStat* txt = getGlobalTextStat(globalNum);
	if (txt && txt->hasStat())
		return 0;

	if (!is_ok)
	{
		// FIXME: check which error this was
		if (!txt)
		{
			txt = new KOMTextStat();
			if (!txt)
				return -1;
			txt->globalNum = globalNum;
			globalTextNum.add(txt);
		}
		// Remove the link from the parent to this comment since it is not available
		if (txt->commentTo >= 0)
		{
			KOMTextStat* parent = getGlobalTextStat(txt->commentTo);
			if (parent && parent->hasStat())
			{
				if (parent->firstComment == txt->globalNum)
				{
					parent->firstComment = txt->nextComment;
					--parent->numComments;
				}
				else
				{
					KOMTextStat* child = getGlobalTextStat(parent->firstComment);
					while (child && child->nextComment != txt->globalNum && child->nextComment >= 0)
						child = getGlobalTextStat(child->nextComment);
					if (child && child->nextComment == txt->globalNum)
					{
						child->nextComment = txt->nextComment;
						--parent->numComments;
					}
				}
			}
		}
		txt->unavailable = true;
		return 0;
	}
	time_t timeval;
	int author;
	int commentTo = -1;
	int numComments = 0;
	int* comments = NULL;
	int textLen;
	int numTo = 0;
	int numCc = 0;
	int* toAndCc = NULL;

	int temp;
	// Check if text or text stat was requested
	bool res = str.parseTime(timeval);
	res &= str.parseInt(author);
	// lines
	res &= str.parseInt(temp);
	// chars
	res &= str.parseInt(textLen);
	// marks
	res &= str.parseInt(temp);
	// Misc info array
	int miscItems;
	res &= str.parseArrayStart(miscItems);
	comments = new int[miscItems];
	if (!comments)
		return -1;
	toAndCc = new int[miscItems];
	if (!toAndCc)
	{
		delete[] comments;
		return -1;
	}
	int curConf = 0;
	KOMTextStat::LocalNum* locals = NULL;
	for (int i = 0; res && i < miscItems; ++i)
	{
		int tnum;
		// type
		res &= str.parseInt(temp);
		if (temp == 7 || temp == 9)
		{
			time_t timeval;
			res &= str.parseTime(timeval);
		}
		else
			res &= str.parseInt(tnum);
		if (res && (temp == 2 || temp == 4) && tnum > 0)
		{
			/*if (commentTo >= 0 && commentTo!=tnum)
			{
				return 1;
			}*/
			// Only allow commenting older texts to avoid loops
			// FIXME: currently only takes the first text, should do something more clever (store all?)
			if (tnum < globalNum && commentTo < 0)
				commentTo = tnum;
		}
		else if (res && (temp == 3 || temp == 5) && tnum > 0)
		{
			// Only add children which have higher number to avoid loops
			if (tnum > 0 && tnum > globalNum)
			{
				comments[numComments] = tnum;
				++numComments;
			}
		}
		else if (res && (temp == 0 || temp == 1 || temp == 15))
		{
			curConf = tnum;
			if (temp == 0)
			{
				toAndCc[numTo] = tnum;
				++numTo;
			}
			else if (temp == 1)
			{
				++numCc;
				toAndCc[miscItems-numCc] = tnum;
			}
		}
		else if (res && curConf > 0 && temp == 6 && tnum > 0)
		{
			KOMTextStat::LocalNum* loc = new KOMTextStat::LocalNum();
			if (!loc)
			{
				delete[] comments;
				delete[] toAndCc;
				while (locals)
				{
					loc = (KOMTextStat::LocalNum*)locals->next;
					delete locals;
					locals = loc;
				}
				return -1;
			}
			loc->confNum = curConf;
			loc->localNum = tnum;
			loc->next = locals;
			locals = loc;
		}

	}
	res &= str.parseArrayEnd(miscItems);
	// aux array
	int auxItems;
	bool utf8 = false;
	bool reformat = false;
	KOMString contentType;
	contentType.append("text/plain");
	res &= str.parseArrayStart(auxItems);
	for (int i = 0; res && i < auxItems; ++i)
	{
		int auxNum;
		int auxTag;
		int auxCreator;
		time_t auxTime;
		unsigned int auxFlags;
		int auxInherit;
		KOMString auxData;
		res &= str.parseInt(auxNum);
		res &= str.parseInt(auxTag);
		res &= str.parseInt(auxCreator);
		res &= str.parseTime(auxTime);
		res &= str.parseBitfield(auxFlags);
		res &= str.parseInt(auxInherit);
		res &= str.parseString(auxData);
		if (res && auxTag == 1)
		{
			// Content type
			reformat = auxData.parseLiteral("text/x-kom-basic") ||
						auxData.parseLiteral("x-kom/text");
			contentType.clear();
			if (reformat)
			{
				contentType.append("text/plain");
			}
			else
			{
				contentType.append(auxData.c_string());
				contentType.terminateAt(';');
				contentType.terminateAt(' ');
			}
			auxData.advanceToAfter(';');
			if (auxData.parseLiteral("charset=utf-8"))
				utf8 = true;
		}
	}

	if (!res)
	{
		delete[] comments;
		delete[] toAndCc;
		while (locals)
		{
			KOMTextStat::LocalNum* loc = (KOMTextStat::LocalNum*)locals->next;
			delete locals;
			locals = loc;
		}
		return 1;
	}
	if (!txt)
	{
		txt = new KOMTextStat();
		if (!txt)
		{
			delete[] comments;
			delete[] toAndCc;
			while (locals)
			{
				KOMTextStat::LocalNum* loc = (KOMTextStat::LocalNum*)locals->next;
				delete locals;
				locals = loc;
			}
			return -1;
		}
		txt->globalNum = globalNum;
		// Do not add texts with negative values
		globalTextNum.add(txt);
	}
	else
	{
		txt->localNums.deleteAll();
	}
	// Do not replace the commentTo if it is already set since it could be set to something else
	if (txt->commentTo < 0)
		txt->commentTo = commentTo;
	bool parent_ignored = txt->commentTo>0 && isIgnored(txt->commentTo);
	// Comments to ignored texts are also ignored
	if (parent_ignored)
		addIgnored(txt->globalNum);
	while (locals)
	{
		KOMTextStat::LocalNum* loc = (KOMTextStat::LocalNum*)locals->next;
		txt->localNums.add(locals);
		if (getConference(locals->confNum))
		{
			getConference(locals->confNum)->addGlobalTextNum(locals->localNum, globalNum);
			if (parent_ignored)
			{
				getConference(locals->confNum)->addIgnored(locals->localNum);
			}
		}
		locals = loc;
	}
	txt->author = author;
	delete[] txt->to;
	txt->to = NULL;
	delete[] txt->cc;
	txt->cc = NULL;
	if (numTo)
	{
		txt->to = new int[numTo];
		if (!txt->to)
		{
			delete[] comments;
			delete[] toAndCc;
			return -1;
		}
		for (int i = 0; i < numTo; ++i)
			txt->to[i] = toAndCc[i];
	}
	txt->numTo = numTo;
	if (numCc)
	{
		txt->cc = new int[numCc];
		if (!txt->cc)
		{
			delete[] comments;
			delete[] toAndCc;
			return -1;
		}
		for (int i = 0; i < numCc; ++i)
			txt->cc[i] = toAndCc[miscItems-1-i];
	}
	delete[] toAndCc;
	txt->numCc = numCc;
	txt->timeval = timeval;
	txt->utf8 = utf8;
	txt->reformat = reformat;
	txt->contentType.clear();
	txt->contentType.append(contentType.c_string());
	txt->textLen = textLen;
	txt->firstComment = -1;
	txt->numComments = numComments;
	if (numComments)
	{
		KOMTextStat* firstct = NULL;
		KOMTextStat* prevct = NULL;
		for (int i = 0; i < numComments; ++i)
		{
			KOMTextStat* ct = getGlobalTextStat(comments[i]);
			if (!ct)
			{
				ct = new KOMTextStat();
				if (!ct)
				{
					// in this case the cache is broken
					clearTextCache();
					delete[] comments;
					return -1;
				}
				ct->globalNum = comments[i];
				// Do not add texts with negative values
				globalTextNum.add(ct);
			}
			if (ct->commentTo >= 0 && ct->commentTo != globalNum)
			{
				--txt->numComments;
				// Make sure the first comment field is set correctly
				if (comments[i] == comments[0] &&  i < numComments-1)
					comments[0] = comments[i+1];
			}
			else
			{
				ct->commentTo = globalNum;
				ct->nextComment = -1;
				if (prevct)
					prevct->nextComment = ct->globalNum;
				if (!firstct)
					firstct = ct;
				prevct = ct;
			}
		}
		if (firstct)
			txt->firstComment = firstct->globalNum;
	}
	delete[] comments;
	if (txt->commentTo >= 0)
	{
		KOMTextStat* parent = getGlobalTextStat(txt->commentTo);
		if (parent && !parent->isUnavailable() && parent->hasStat() && parent->firstComment != txt->globalNum)
		{
			// Make sure the parent has a pointer to this
			// If this is not done new texts will never be 
			// visible since the parent did not have a pointer
			// to this text yet when the stat for the parent
			// was requested
			if (parent->firstComment < 0)
			{
				parent->numComments = 1;
				parent->firstComment = txt->globalNum;
				txt->nextComment = -1;
			}
			else
			{
				KOMTextStat* lastChild = getGlobalTextStat(parent->firstComment);
				while (lastChild && lastChild->nextComment >= 0 && lastChild->nextComment != txt->globalNum)
					lastChild = getGlobalTextStat(lastChild->nextComment);
				if (lastChild && lastChild->nextComment < 0)
				{
					++parent->numComments;
					lastChild->nextComment = txt->globalNum;
					txt->nextComment = -1;
				}
			}
		}
	}
	return 0;
}

class KOMFavoriteCacheEntry : public KOMHashEntry
{
public:
	KOMFavoriteCacheEntry(int tn) : textNum(tn), next(NULL)
	{}
	virtual ~KOMFavoriteCacheEntry()
	{}
	virtual int getHashKey(int hashId){return textNum;}
	virtual KOMHashEntry* &getNextHashEntry(int hashId){return next;}

	int textNum;
	KOMHashEntry* next;
};

void KOMCache::addFavorite(int globalNum)
{
	if (isFavorite(globalNum))
		return;
	KOMFavoriteCacheEntry* ce = new KOMFavoriteCacheEntry(globalNum);
	if (ce)
		favoritesCache.add(ce);
}

void KOMCache::removeFavorite(int globalNum)
{
	KOMFavoriteCacheEntry* ce = (KOMFavoriteCacheEntry*)favoritesCache.get(globalNum);
	if (ce)
	{
		favoritesCache.remove(ce);
		delete ce;
	}
}

void KOMCache::clearFavorites()
{
	favoritesCache.deleteAll();
}

bool KOMCache::isFavorite(int globalNum)
{
	return favoritesCache.get(globalNum)!=NULL;
}

void KOMCache::addIgnored(int globalNum)
{
	if (ignoredTexts.get(globalNum)!=NULL)
		return;
	KOMFavoriteCacheEntry* ce = new KOMFavoriteCacheEntry(globalNum);
	if (ce)
		ignoredTexts.add(ce);
	KOMTextStat* txt = getGlobalTextStat(globalNum);
	if (txt)
	{
		if (txt->commentTo > 0)
			addIgnored(txt->commentTo);
		if (txt->nextComment > 0)
			addIgnored(txt->nextComment);
		if (txt->firstComment > 0)
			addIgnored(txt->firstComment);

		for (KOMConferenceStat* c = getFirstConference(); c; c = c->nextConference())
		{
			int loc = txt->getLocalNum(c->getNum());
			if (loc > 0)
				c->addIgnored(loc);
		}
	}
}
bool KOMCache::isIgnored(int globalNum)
{
	return (ignoredTexts.get(globalNum)!=NULL);
}
void KOMConferenceStat::addIgnored(int localNum)
{
	if (ignoredTexts.get(localNum)!=NULL)
		return;
	KOMFavoriteCacheEntry* ce = new KOMFavoriteCacheEntry(localNum);
	if (ce)
		ignoredTexts.add(ce);
}

bool KOMConferenceStat::isIgnored(int localNum)
{
	return (ignoredTexts.get(localNum)!=NULL);
}

class KOMReplyTranslation : public KOMHashEntry
{
public:
	virtual ~KOMReplyTranslation(){}
	virtual int getHashKey(int hashId){return conf;}
	virtual KOMHashEntry* &getNextHashEntry(int hashId){return next;}

	int conf;
	int reconf;
	KOMHashEntry* next;
};
int KOMCache::getReplyConference(int conf)
{
	KOMReplyTranslation* trans = (KOMReplyTranslation*)replyTranslations.get(conf);
	if (!trans)
		return 0;
	return trans->reconf;
}

void KOMCache::setReplyConference(int conf, int reconf)
{
	KOMReplyTranslation* trans = new KOMReplyTranslation();
	if (!trans)
		return;
	trans->conf = conf;
	trans->reconf = reconf;
	replyTranslations.add(trans);
}
