/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#ifndef KOMCACHE_H
#define KOMCACHE_H

#include "komstring.h"

/** An entry in the hash table. */
class KOMHashEntry
{
public:
	virtual ~KOMHashEntry(){}
	/** Get the key for this hash entry. */
	virtual int getHashKey(int hashId) = 0;
	/** Get a reference to the next hash entry pointer. */
	virtual KOMHashEntry* &getNextHashEntry(int hashId) = 0;
};
// Hash size should be prime
#define KOM_INITIAL_HASH_SIZE 101
/** Hash table used for kom caching. */
class KOMHash
{
public:
	KOMHash(int hashId = 0);
	~KOMHash();

	/** Add an entry to the hash table.
	 * @param entry the entry to add. */
	void add(KOMHashEntry* entry);
	/** Get the entry for a key, or NULL if the key is not in the table. */
	KOMHashEntry* get(int key);
	/** Remove an entry from the hash table. */
	void remove(KOMHashEntry* entry);

	/** Delete all entries stored in this hash table. */
	void deleteAll();
	/** Remove all entries from the cache without deleting them. */
	void clear();
private:
	/** Change the size of the hash table and rehash. */
	void changeSize(int newSize);
	int id;
	int numEntries;
	int numSlots;
	KOMHashEntry** entries;
	KOMHashEntry* nomemEntry;
};

/** A cached text status. */
class KOMTextStat : public KOMHashEntry
{
public:
	KOMTextStat();
	virtual ~KOMTextStat();

	// Hash entry functions
	virtual int getHashKey(int hashId){return globalNum;}
	virtual KOMHashEntry* &getNextHashEntry(int hashId){return nextGlobal;}

	/** @returns true if the stat of this text stat is initialized. */
	bool hasStat(){return author>=0;}
	/** @ returns true if the text is not available due to permission or
	 * that it has been deleted. */
	bool isUnavailable(){return unavailable;}
	/** Mark this text as unavailable. */
	//void setUnavailable(){unavailable = true;}

	int getGlobalNum(){return globalNum;}
	int getLocalNum(int conf);
	int getCommentTo(){return commentTo;}
	int getAuthor(){return author;}
	int getNumTo(){return numTo;}
	int getTo(int i){return to[i];}
	int getNumCc(){return numCc;}
	int getCc(int i){return cc[i];}
	time_t getTimeval(){return timeval;}
	int getNumComments(){return numComments;}
	int getFirstComment(){return firstComment;}
	int getNextComment(){return nextComment;}
	int getTextLength(){return textLen;}

	bool isUTF8(){return utf8;}
	bool shouldReformat(){return reformat;}
	KOMString& getContentType(){return contentType;}
private:
	KOMHashEntry* nextGlobal;
	class LocalNum : public KOMHashEntry
	{
	public:
		virtual ~LocalNum()
		{}
		virtual int getHashKey(int hashId){return confNum;}
		virtual KOMHashEntry* &getNextHashEntry(int hashId){return next;}
		int confNum;
		int localNum;
		KOMHashEntry* next;
	};

	int globalNum;
	KOMHash localNums;

	int author;
	int* to;
	int numTo;
	int* cc;
	int numCc;
	time_t timeval;
	int commentTo;
	int firstComment;
	int nextComment;
	int numComments;
	int textLen;
	bool unavailable;
	bool utf8;
	bool reformat;
	KOMString contentType;

	friend class KOMCache;
};

/** A text data entry containing the actual text data. */
class KOMTextData : public KOMHashEntry
{
public:
	KOMTextData(int tnum) : textNum(tnum), nextList(NULL), prevList(NULL)
	{}
	virtual ~KOMTextData(){}

	virtual int getHashKey(int hashId){return textNum;}
	virtual KOMHashEntry* &getNextHashEntry(int hashId){return nextHashEntry;}

	KOMString text;

private:
	int textNum;

	KOMHashEntry* nextHashEntry;
	KOMTextData* nextList;
	KOMTextData* prevList;

	friend class KOMCache;
};

/** Cached conference. Includes name and texts. */
class KOMConferenceStat
{
public:
	KOMConferenceStat(int number);

	void addRef(){++refCount;}
	void decRef(){--refCount;if (refCount<=0) delete this;}

	/** Get the next conference in the conference list. */
	KOMConferenceStat* nextConference(){return next;}
	/** Get the name of the conference. */
	const char* getName(){return name.len()?name.c_string():NULL;}
	/** Set the name of the conference. Set NULL to clear conference. */
	int setName(const char* n);
	/** Get the number of the conference. */
	int getNum(){return conferenceNum;}

	/** Set the highest local number found in this conference. */
	void setHighestLocal(int hl){highestLocal = hl;}
	/** Get the highest local number found in this conference. */
	int getHighestLocal(){return highestLocal;}

	/** Set the read status of this conference. */
	void setRead(bool r){read = r;}
	/** Check if this conference is read. */
	bool isRead(){return read;}

	/** Set the invite status of this conference. */
	void setInvite(bool inv){invite = inv;}
	/** Check if this conference is an invite or regular subscription. */
	bool isInvite(){return invite;}
	/** Set the secret status of this conference. */
	void setSecret(bool sec){secret = sec;}
	/** Check if this conference is secret or regular subscription. */
	bool isSecret(){return secret;}
	/** Set the passive status of this conference. */
	void setPassive(bool pass, bool msgpassinv){passive = pass; instant_passive = ((pass&&!msgpassinv)||(!pass&&msgpassinv));}
	/** Check if this conference is passive. */
	bool isPassive(){return passive;}
	/** Check if the instant message part of this conference is passive. */
	bool isInstantPassive(){return instant_passive;}

	/** Add  text to the read range of this conference. */
	int addReadText(int textNum);
	/** Check if a local text in this conference is read. */
	bool isTextRead(int textNum, bool checkIgnoreList = true);
	/** Find the next unread local text with number lower 
	 * than or equal to textNum. */
	int findNextLowerUnread(int textNum);

	/** Update the range of cached entries. */
	void updateCacheRange(int first, bool foundFirst);
	int getFirstCached(){return firstCached;}
	bool isOldestCached(){return oldestCached;}

	/** Translate a local number to a global one. */
	int getGlobalTextNum(int localNum);

	bool hasNewMessages(){return new_msgs;}
	void setNewMessages(bool m){new_msgs = m;}

	bool isIgnored(int localNum);
	void addIgnored(int localNum);
private:
	/** Clear the text cache, delete all entries. Should be called when the
	 * cache is made invalid. */
	void clearTextCache();
	/** Add a global to local text translation. */
	int addGlobalTextNum(int localNum, int globalNum);
	class LocToGlobTextNum : public KOMHashEntry
	{
	public:
		LocToGlobTextNum(int local, int global) : localNum(local), globalNum(global), next(NULL)
		{}
		~LocToGlobTextNum()
		{}
		virtual int getHashKey(int hashId){return localNum;}
		virtual KOMHashEntry* &getNextHashEntry(int hashId){return next;}
		int localNum;
		int globalNum;
		KOMHashEntry* next;
	};
	virtual ~KOMConferenceStat();
	int refCount;

	int conferenceNum;
	KOMString name;
	bool read;
	int highestLocal;

	bool invite;
	bool secret;
	bool passive;
	bool instant_passive;

	int numReadRanges;
	int allocatedReadRanges;
	int* readRanges;

	int firstCached;
	bool oldestCached;

	KOMHash localToGlobalTextNum;
	KOMHash ignoredTexts;

	bool new_msgs;
	KOMConferenceStat* next;
	friend class KOMCache;
};

class KOMConfNameCacheEntry : public KOMHashEntry
{
public:
	KOMConfNameCacheEntry(KOMConfNameCacheEntry* nl) : next(NULL), nextInList(nl)
	{}
	virtual ~KOMConfNameCacheEntry()
	{}
	virtual int getHashKey(int hashId){return num;}
	/** Get a reference to the next hash entry pointer. */
	virtual KOMHashEntry* &getNextHashEntry(int hashId){return next;}

	KOMConfNameCacheEntry* getNextInList(){return nextInList;}

	int num;
	KOMString name;
private:
	KOMHashEntry* next;
	KOMConfNameCacheEntry* nextInList;
};

/** Tha main cache class of KOMclean. This handles text data and conference 
 * cacheing. KOMSession extends this class so the cache functions are availbel 
 * in the kom session object. */
class KOMCache
{
public:
	KOMCache();
	virtual ~KOMCache();
	
	/** Reset the cache to only contain values considered safe when reconnecting. */
	void resetCache();

	/** Checks if the conference list is dirty or not. */
	bool isConferenceListDirty(){return confsDirty;}
	/** Mark a conference list as dirty or not dirty. */
	void markConferenceListDirty(bool dirty){confsDirty = dirty;}
	/** Get a conference with the specified number. */
	KOMConferenceStat* getConference(int conf);
	/** Get the first cached conference. */
	KOMConferenceStat* getFirstConference(){return confs;}
	/** Add a conference to the cache. */
	void addConference(KOMConferenceStat* conf);
	/** Delete a conference from the cache. */
	void deleteConference(int conf);

	/** Get a conference name from the name cache. */
	const KOMString* getConferenceName(int id);
	/** Update a conference name in the name cache. */
	void updateConferenceName(int id, const KOMString& name);
	KOMConfNameCacheEntry* getFirstCachedName(){return firstCachedName;}

	/** Clear all cached text stats from a conference. -1 means all. */
	void clearTextCache();

	/** Add a new text data entry to the cache. */
	int addTextData(KOMTextData* d);
	/** Get a text data entry from the cache. */
	KOMTextData* getTextData(int text);

	int addTextStat(int globalNum, bool is_ok, KOMString& str);
	int addEmptyTextStat(int globalNum, int localNum, int confNum);
	KOMTextStat* getGlobalTextStat(int num){return (KOMTextStat*)globalTextNum.get(num);}

	void addFavorite(int globalNum);
	void removeFavorite(int globalNum);
	void clearFavorites();
	bool isFavorite(int globalNum);

	void addIgnored(int globalNum);
	bool isIgnored(int globalNum);

	/** @returns the conference which replies for conf should be sent to, or 0 if it is not yet known. */
	int getReplyConference(int conf);
	/** Cache the redirection of replies from conf to reconf. */
	void setReplyConference(int conf, int reconf);
private:
	bool confsDirty;
	KOMConferenceStat* confs;

	KOMHash globalTextNum;

	KOMTextData* firstTextData;
	KOMTextData* lastTextData;
	KOMHash textData;
	int textCacheSize;
	KOMHash nameCache;
	KOMConfNameCacheEntry* firstCachedName;

	KOMHash favoritesCache;
	KOMHash ignoredTexts;

	KOMHash replyTranslations;
};

#endif // KOMCACHE_H

