/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#define _CRT_SECURE_NO_WARNINGS

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

// Required socket includes
#ifdef WIN32
#include <winsock.h>
#define close(s) closesocket(s)
#define socket_errno WSAGetLastError()
#define EAGAIN WSAEWOULDBLOCK
#define UINT32 unsigned long
#define WM_SOCKET WM_USER
extern HWND g_win;
#else // !WIN32
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#define socket_errno errno
#define UINT32 u_int32_t
#endif // !WIN32

#include "komclient.h"
#include "komasync.h"

// Commands
#include "komlogin.h"
#include "komlistsubconf.h"
#include "komlistallconf.h"
#include "komsubscribe.h"
#include "komlistthreads.h"
#include "komlistmessages.h"
#include "komsendmsg.h"
#include "komsendinstantmsg.h"
#include "kommarkread.h"
#include "komnextunreadthread.h"
#include "komconfinfo.h"
#include "komcreateconf.h"
#include "komchangepresentation.h"
#include "komlookupname.h"
#include "komfavorites.h"

#define KOM_DEBUG
#ifdef KOM_DEBUG
FILE* logfile = NULL;
#endif // KOM_DEBUG

static inline void writeReply(KOMCommandListener* l, int sid, bool ok, const char* msg)
{
	KOMString reply;
	if (reply.append(ok?"OK":("ERROR" KOM_GROUP_SEPARATOR)) ||
			reply.append(msg))
		l->onReply(sid, "ERROR" KOM_GROUP_SEPARATOR "Out of memory");
	else
		l->onReply(sid, reply.c_string());
}

KOMSession::KOMSession() : socket(-1), delayedCommands(NULL), connecting(false), disconnecting(false), reconnecting(false), protocolVersion(11), commandID(2), async(NULL), lastActive(0), curConfNum(0)
{
#ifdef KOM_DEBUG
	if (!logfile)
		logfile = fopen("/tmp/komlog.txt", "w");
#endif // KOM_DEBUG
}

KOMSession::~KOMSession()
{
	while (delayedCommands)
	{
		KOMDelayedCommand* nc = delayedCommands->next;
		writeReply(delayedCommands->listener, delayedCommands->sessionId, false, "");
		delete delayedCommands;
		delayedCommands = nc;
	}
	if (socket >= 0)
		close(socket);
	delete async;
	KOMCommand* nextCmd = NULL;
	for (KOMCommand* cmd = commands.first; cmd; cmd = nextCmd)
	{
		nextCmd = cmd->next;
		cmd->error(KOM_ERROR_NOT_LOGGEDIN);
	}
#ifdef KOM_DEBUG
	if (logfile)
	{
		fclose(logfile);
		logfile = NULL;
	}
#endif // KOM_DEBUG
}

int KOMSession::connect(const char* srv, const char* user, const char* passwd)
{
	// Make sure there are no old replies pending
	reply.clear();
	struct hostent* addr;
	struct sockaddr_in komaddr;
	if (isConnected())
	{
		if (getFirstCommand())
			getFirstCommand()->error(KOM_ERROR_ALREADY_CONNECTED);
		return -1;
	}

	addr = gethostbyname(srv);
	if (!addr || addr->h_length < 1)
	{
		if (getFirstCommand())
			getFirstCommand()->error(KOM_ERROR_SERVER_NOT_FOUND);
		return -1;
	}
	socket = ::socket(PF_INET, SOCK_STREAM, 0);
	if (socket < 0)
	{
		if (getFirstCommand())
			getFirstCommand()->error(KOM_ERROR_SOCKET);
		return -1;
	}
	komaddr.sin_family = AF_INET;
	komaddr.sin_port = htons(4894);
	komaddr.sin_addr.s_addr = *(UINT32*)(addr->h_addr_list[0]);
	if (::connect(socket, (struct sockaddr*)&komaddr, sizeof(komaddr))<0)
	{
		if (getFirstCommand())
			getFirstCommand()->error(KOM_ERROR_SERVER_TIMEOUT);
		disconnect(reconnecting);
		return -1;
	}
	if (currentServer.c_string() != srv)
	{
		currentServer.clear();
		currentServer.append(srv);
	}
	if (currentUser.c_string() != user)
	{
		currentUser.clear();
		currentUser.append(user);
	}
	if (currentPasswd.c_string() != passwd)
	{
		currentPasswd.clear();
		currentPasswd.append(passwd);
	}
	// Write the hello string
	connecting = true;
	KOMString hello;
	if (hello.appendFormat("A%dH%s%%web\n", strlen(user)+4, user))
	{
		if (getFirstCommand())
			getFirstCommand()->error(KOM_ERROR_OOM);
		return -1;
	}
#ifdef WIN32
	// AsyncSelect sets the socket in non blocking mode
	WSAAsyncSelect(socket, g_win, WM_SOCKET, FD_READ|FD_CLOSE);
#else // !WIN32
	int flags;
#ifdef O_NONBLOCK
	if ((flags = fcntl(socket, F_GETFL, 0)) == -1)
		flags = 0;
	fcntl(socket, F_SETFL, flags | O_NONBLOCK);
#else
	flags = 1;
	ioctl(socket, FIOBIO, &flags);
#endif
#endif // !WIN32
#ifdef KOM_DEBUG
	if (logfile)
	{
		fprintf(logfile, "SEND: %s", hello.c_string());
		fflush(logfile);
	}
#endif // KOM_DEBUG
	if (send(socket, hello.c_string(), hello.len(), 0) < 0)
	{
		disconnect(false);
		return -1;
	}
	return 0;
}

int KOMSession::reconnect()
{
	if (isConnected() || !isReconnecting())
	{
		if (getFirstCommand())
			getFirstCommand()->error(KOM_ERROR_ALREADY_CONNECTED);
		return -1;
	}
	return connect(currentServer.c_string(), currentUser.c_string(), currentPasswd.c_string());
}


void KOMSession::disconnect(bool reconnect)
{
	if (socket >= 0)
		close(socket);
	socket = -1;
	connecting = false;
	if (getFirstCommand())
		getFirstCommand()->error(KOM_ERROR_SERVER_TIMEOUT);
	while (delayedCommands)
	{
		KOMDelayedCommand* dc = delayedCommands;
		delayedCommands = delayedCommands->next;
		writeReply(dc->listener, dc->sessionId, false, KOM_ERROR_SERVER_TIMEOUT);
		delete dc;
	}
	// The cache will be invalid if we reconnect since new messages while disconnected were missed
	resetCache();
	reconnecting = reconnect;
}

void KOMSession::processData()
{
	while (1)
	{
		char tmpreply[1024];
		int replyLen = recv(socket, tmpreply, sizeof(tmpreply)-1, 0);
		if (replyLen > 0)
		{
			tmpreply[replyLen] = 0;
			reply.append(tmpreply, replyLen);
#ifdef KOM_DEBUG
			if (logfile && *tmpreply)
			{
				fprintf(logfile, "RECV: %s", tmpreply);
				fflush(logfile);
			}
#endif // KOM_DEBUG
		}
		else
		{
			if (replyLen == 0 || socket_errno != EAGAIN)
			{
				disconnect(true);
			}
			break;
		}
	}
	// sometimes there are multiple replies in one packet
	while (reply.hasCompleteReply())
	{
		reply.saveParserPosition();
		bool is_ok;
		if (reply.parseAsync())
		{
			reply.restoreParserPosition();
			if (!async)
				async = new KOMAsync(this);
			// Store a list of all pending async commands
			if (async)
				async->progress(reply);
		}
		else if (reply.parseProtocolError())
		{
			if (getFirstCommand())
				getFirstCommand()->error(KOM_ERROR_PROTOCOL);
		}
		else if (reply.parseStatus(is_ok) || connecting)
		{
			reply.restoreParserPosition();
			KOMCommand* nextCmd = NULL;
			for (KOMCommand* cmd = commands.first; cmd; cmd = nextCmd)
			{
				nextCmd = cmd->next;
				if (cmd->progress(reply))
					break;
				// Restore all changes the command did before it noticed it should not handle this
				reply.restoreParserPosition();
			}
		}
		// Delete to the parsed position and then until the first newline
		reply.deleteParsed();
		reply.advanceToAfter('\n');

		if (async && !getFirstCommand())
		{
			// There are not active commands, it is safe to clear the cache
			async->clearCacheIfNeeded();
		}

		if (disconnecting)
		{
			disconnect(true);
			disconnecting = false;
		}

		while (!getFirstCommand() && delayedCommands)
		{
			KOMDelayedCommand* dc = delayedCommands;
			delayedCommands = delayedCommands->next;
			if (createCommand(dc->listener, dc->sessionId, dc->cmd.c_string()))
			{
				writeReply(dc->listener, dc->sessionId, false, KOM_ERROR_OOM);
			}
			delete dc;
		}
	}
}

void KOMSession::pingServer()
{
	KOMString cmdstr;
	cmdstr.appendFormat("%d 56\n", getCommandID());
	sendCommand(cmdstr.c_string());
}

void KOMSession::sendCommand(const char* cmd)
{
	while (1)
	{
		char tmpreply[1024];
		int replyLen = recv(socket, tmpreply, sizeof(tmpreply)-1, 0);
		if (replyLen > 0)
		{
			tmpreply[replyLen] = 0;
			reply.append(tmpreply, replyLen);
#ifdef KOM_DEBUG
			if (logfile && *tmpreply)
			{
				fprintf(logfile, "RECV: %s", tmpreply);
				fflush(logfile);
			}
#endif // KOM_DEBUG
		}
		else
		{
			if (replyLen == 0 || socket_errno != EAGAIN)
			{
				disconnecting = true;
			}
			break;
		}
	}
#ifdef KOM_DEBUG
	if (logfile)
	{
		int temp1, temp2;
		if (sscanf(cmd, "%d 89 %d", &temp1, &temp2) == 2 ||
			sscanf(cmd, "%d 62 %d", &temp1, &temp2) == 2)
			fprintf(logfile, "SEND: \"login\"\n");
		else
			fprintf(logfile, "SEND: %s", cmd);
		fflush(logfile);
	}
#endif // KOM_DEBUG
	if (send(socket, cmd, strlen(cmd), 0) < 0)
	{
		disconnecting = true;
	}
}

int KOMSession::createCommand(KOMCommandListener* l, int sid, const char* str)
{
	if (getFirstCommand())
	{
		KOMDelayedCommand* dc = new KOMDelayedCommand();
		dc->listener = l;
		dc->sessionId = sid;
		dc->cmd.append(str);
		dc->next = NULL;
		KOMDelayedCommand** nextc = &delayedCommands;
		KOMDelayedCommand* lc = delayedCommands;
		while (lc)
		{
			nextc = &lc->next;
			lc = lc->next;
		}
		*nextc = dc;
		return 0;
	}
	KOMCommand* cmd = NULL;
	KOMString cmdstr;
	bool send_user_active = true;
	cmdstr.append(str);
#ifdef KOM_DEBUG
	if (logfile && !cmdstr.parseLiteral("login") && !cmdstr.parseLiteral("poll_async"))
	{
		fprintf(logfile, "CMD[%d]: %s\n", (int)time(NULL), cmdstr.c_string());
		fflush(logfile);
	}
	cmdstr.resetParserPosition();
#endif // KOM_DEBUG
	// Check which command was issued and perform the apropriate task
	if (cmdstr.parseLiteral("login") || cmdstr.parseLiteral("reconnect"))
	{
		cmd = new KOMLoginCommand(l, sid, this);
		send_user_active = false;
	}
	else
	{
		if (!isLoggedIn())
		{
			if (isReconnecting() && cmdstr.parseLiteral("logout"))
			{
				disconnect(false);
				writeReply(l, sid, true, "");
			}
			else
			{
				// invalid
				writeReply(l, sid, false, KOM_ERROR_NOT_LOGGEDIN);
			}
			return 0;
		}
		int prevId;
		if (cmdstr.parseLiteral("poll_async") && cmdstr.parseInt(prevId))
		{
			KOMString evt;
			if (async)
			{
				if (async->getEvents(evt, prevId))
				{
					writeReply(l, sid, false, KOM_ERROR_OOM);
					return 0;
				}
			}
			else
				evt.appendFormat(KOM_GROUP_SEPARATOR "%d", prevId);
			writeReply(l, sid, true, evt.c_string());
			return 0;
		}
		else if (cmdstr.parseLiteral("mark_instant_message_read") && cmdstr.parseInt(prevId))
		{
			if (async)
			{
				async->markInstantMessageRead(prevId);
			}
			send_user_active = false;
		}
		else if (cmdstr.parseLiteral(KOMListSubscribedConferencesCommand::commandString))
		{
			cmd = new KOMListSubscribedConferencesCommand(l, sid, this);
			send_user_active = false;
		}
		else if (cmdstr.parseLiteral(KOMListAllConferencesCommand::commandString))
		{
			cmd = new KOMListAllConferencesCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMListAllOnlineUsersCommand::commandString))
		{
			cmd = new KOMListAllOnlineUsersCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMLookupNameFromIDCommand::commandString))
		{
			cmd = new KOMLookupNameFromIDCommand(l, sid, this);
			send_user_active = false;
		}
		else if (cmdstr.parseLiteral(KOMListCachedNamesCommand::commandString))
		{
			cmd = new KOMListCachedNamesCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMPrefetchNamesCommand::commandString))
		{
			cmd = new KOMPrefetchNamesCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMListThreadsCommand::commandString))
		{
			cmd = new KOMListThreadsCommand(l, sid, this);
			send_user_active = false;
		}
		else if (cmdstr.parseLiteral(KOMListMessagesCommand::commandString))
		{
			cmd = new KOMListMessagesCommand(l, sid, this);
			send_user_active = false;
		}
		else if (cmdstr.parseLiteral(KOMSendMsgCommand::commandString))
		{
			cmd = new KOMSendMsgCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMGetReplyRecptCommand::commandString))
		{
			cmd = new KOMGetReplyRecptCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMSendInstantMsgCommand::commandString))
		{
			cmd = new KOMSendInstantMsgCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMMarkReadCommand::commandString))
		{
			cmd = new KOMMarkReadCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMIgnoreThreadCommand::commandString))
		{
			cmd = new KOMIgnoreThreadCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMSetFavoriteCommand::commandString))
		{
			cmd = new KOMSetFavoriteCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMListFavoritesCommand::commandString))
		{
			cmd = new KOMListFavoritesCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMSubscribeCommand::commandString))
		{
			cmd = new KOMSubscribeCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMChangeSubscriptionTypeCommand::commandString))
		{
			cmd = new KOMChangeSubscriptionTypeCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMNextUnreadThreadCommand::commandString))
		{
			cmd = new KOMNextUnreadThreadCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMConferenceInfoCommand::commandString))
		{
			cmd = new KOMConferenceInfoCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMCreateConferenceCommand::commandString))
		{
			cmd = new KOMCreateConferenceCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMChangeConferenceTypeCommand::commandString))
		{
			cmd = new KOMChangeConferenceTypeCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMRenameConferenceCommand::commandString))
		{
			cmd = new KOMRenameConferenceCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMChangeSupervisorCommand::commandString))
		{
			cmd = new KOMChangeSupervisorCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral(KOMChangePresentationCommand::commandString))
		{
			cmd = new KOMChangePresentationCommand(l, sid, this);
		}
		else if (cmdstr.parseLiteral("logout"))
		{
			disconnect(false);
			writeReply(l, sid, true, "");
			return 0;
		}
		else
		{
			writeReply(l, sid, false, KOM_ERROR_UNKNOWN_COMMAND);
			return 0;
		}
	}
	if (cmd)
	{
		if (send_user_active)
			userActive();
		cmdstr.resetParserPosition();
		int retval = cmd->Construct(cmdstr);
		if (disconnecting)
		{
			disconnect(true);
			disconnecting = false;
		}
		return retval;
	}
	writeReply(l, sid, false, KOM_ERROR_OOM);
	return 0;
}

void KOMSession::updateUnread()
{
	if (async)
		async->updateUnread();
}

bool KOMSession::hasUnread()
{
	for (KOMConferenceStat* conf = getFirstConference(); conf; conf = conf->nextConference())
	{
		if (!conf->isRead())
			return true;
	}
	if (async)
		return async->hasNewTexts();
	return false;
}

bool KOMSession::hasUnreadInstantMsg()
{
	if (async)
		return async->hasUnreadInstantMsg();
	return false;
}

void KOMSession::addInstantMessage(int from, int to, KOMString& text)
{
	if (async)
		async->addInstantMessage(from, to, text);
}

void KOMSession::changeCurrentConference(int conf)
{
	if (conf != curConfNum)
	{
		userActive();
		KOMString cmdstr;
		cmdstr.appendFormat("%d 2 %d\n", getCommandID(), conf);
		sendCommand(cmdstr.c_string());
		curConfNum = conf;
	}
}

void KOMSession::userActive()
{
	time_t t = time(NULL);
	if (t-lastActive >= 30)
	{
		KOMString cmdstr;
		cmdstr.appendFormat("%d 82\n", getCommandID());
		sendCommand(cmdstr.c_string());
		lastActive = t;
	}
}

KOMCommand* KOMSession::getFirstCommand()
{
	return commands.first;
}

KOMCommand::KOMCommand(KOMCommandListener* l, int sid, KOMSession* kom) : prev(NULL), next(NULL), listener(l), sessionid(sid), komSession(kom)
{
	prev = komSession->commands.last;
	if (komSession->commands.last)
		komSession->commands.last->next = this;
	else
		komSession->commands.first = this;
	komSession->commands.last = this;
}

KOMCommand::~KOMCommand()
{
	if (!prev)
		komSession->commands.first = next;
	else
		prev->next = next;
	if (!next)
		komSession->commands.last = prev;
	else
		next->prev = prev;
}

void KOMCommand::ok(const char* msg)
{
#ifdef KOM_DEBUG
	if (logfile)
	{
		fprintf(logfile, "OK : %s", msg);
		fflush(logfile);
	}
#endif // KOM_DEBUG
	writeReply(listener, sessionid, true, msg);
	delete this;
}

void KOMCommand::error(const char* msg)
{
#ifdef KOM_DEBUG
	if (logfile)
	{
		fprintf(logfile, "ERR: %s", msg);
		fflush(logfile);
	}
#endif // KOM_DEBUG
	writeReply(listener, sessionid, false, msg);
	delete this;
}

