/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#ifndef KOMCLIENT_H
#define KOMCLIENT_H

#include "komcache.h"
#include "komstring.h"

// Some predefined error replies
#define KOM_ERROR_ALREADY_CONNECTED "Already connected"
#define KOM_ERROR_OOM "Out of memory"
#define KOM_ERROR_SERVER_NOT_FOUND "Could not locate server"
#define KOM_ERROR_SERVER_TIMEOUT "Could not connect to server"
#define KOM_ERROR_SOCKET "Socket error"
#define KOM_ERROR_PROTOCOL "Protocol error"
#define KOM_ERROR_INVALID_USER "No such user"
#define KOM_ERROR_SYNTAX "Syntax error"
#define KOM_ERROR_NOT_LOGGEDIN "Not logged in"
#define KOM_ERROR_UNKNOWN_COMMAND "Unknown command"
#define KOM_ERROR_INVALID_CHARACTERS "Invalid (non Latin1) characters in the string"

#define KOM_CONTENT_TYPE "text/plain; charset=iso-8859-1"
#define KOM_CONTENT_TYPE_LEN 30
#define KOM_CONTENT_TYPE_UTF8 "text/plain; charset=utf-8"
#define KOM_CONTENT_TYPE_UTF8_LEN 25

class KOMAsync;
class KOMCommand;
class KOMCommandListener;

/** A kom session which is communicating with KOM. This is the main KOM client
 * class of KOMclean. */
class KOMSession : public KOMCache
{
public:
	KOMSession();
	~KOMSession();

	/** Connect to a server. */
	int connect(const char* srv, const char* user, const char* passwd);
	int reconnect();
	/** Set loged in status and user id. */
	void setLoggedIn(int uid){connecting = false; userID = uid;}
	/** Disconnect from the server. */
	void disconnect(bool reconnect);
	bool isConnected(){return socket >= 0;}
	bool isLoggedIn(){return isConnected() && !connecting;}
	bool isReconnecting(){return !isLoggedIn() && reconnecting;}
	/** Set the protocol version used by the kom server. Returns false if 
	 * the protocol version is not supported. */
	bool setProtocolVersion(int version){protocolVersion = version;return version>=10;}
	int getProtocolVersion(){return protocolVersion;}

	/** Process a data response from KOM .*/
	void processData();
	
	/** Create a new command. */
	int createCommand(KOMCommandListener* l, int sid, const char* str);

	/** Send a request to the server to keep the connection alive and detect disconnections. */
	void pingServer();

	/** Send a command string to kom. */
	void sendCommand(const char* cmd);

	/** Get the next available command id. */
	int getCommandID(){return commandID++;}
	/** Get the id of the current user. */
	int getUserID(){return userID;}

	bool hasUnread();
	bool hasUnreadInstantMsg();

	void updateUnread();
	void addInstantMessage(int from, int to, KOMString& text);

	/** Get the first command in the command list. */
	KOMCommand* getFirstCommand();
	class CommandList
	{
	public:
		CommandList() : first(NULL), last(NULL)
		{}
		KOMCommand* first;
		KOMCommand* last;
	};
	CommandList commands;
	int socket;

	void changeCurrentConference(int conf);
private:
	/** Tell the server that the user is active if that has not been done the last 30 seconds. */
	void userActive();

	struct KOMDelayedCommand
	{
		KOMString cmd;
		KOMCommandListener* listener;
		int sessionId;
		KOMDelayedCommand* next;
	};
	KOMDelayedCommand* delayedCommands;
	bool connecting;
	bool disconnecting;
	bool reconnecting;
	int userID;
	int protocolVersion;

	int commandID;

	KOMAsync* async;
	KOMString reply;
	time_t lastActive;
	int curConfNum;
public:
	KOMString currentServer;
	KOMString currentUser;
	KOMString currentPasswd;
};

/** Listener which receives the reply from an issued command. */
class KOMCommandListener
{
public:
	virtual ~KOMCommandListener(){}
	/** The reply is ready and can be presented to the user. */
	virtual void onReply(int sid, const char* str) = 0;
};

/** A representation of a command in kom. */
class KOMCommand
{
public:
	KOMCommand(KOMCommandListener* l, int id, KOMSession* kom);
	virtual ~KOMCommand();
	
	/** Construct the command and send the initial request to the server.
	 * @returns 0 on success. */
	virtual int Construct(KOMString& cmd) = 0;
	/** Progress the command. If the command is done, ok or error must be 
	 * called. If the command does not belong to this command, false must
	 * be returned so the command can be passed to another command. */
	virtual bool progress(KOMString& reply) = 0;

	/** Reply ok to this command. */
	void ok(const char* msg);
	/** Reply to this command with an error. */
	void error(const char* msg);
	
	KOMCommand* prev;
	KOMCommand* next;
protected:
	KOMCommandListener* listener;
	int sessionid;
	KOMSession* komSession;
};

#endif // KOMCLIENT_H

