/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komconfinfo.h"

const char* const KOMConferenceInfoCommand::commandString = "conference_info";

int KOMConferenceInfoCommand::Construct(KOMString& cmd)
{
	if (!cmd.parseLiteral(commandString) || !cmd.parseInt(confNum))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	presentationNum = 0;
	presentationUTF8 = false;
	presentationReformat = false;
	
	cmdID = komSession->getCommandID();

	KOMString cmdstr;
	if (cmdstr.appendFormat("%d 91 %d\n", cmdID, confNum))
	{
		error(KOM_ERROR_OOM);
		return 0;
	}

	state = CONF_INFO_STAT;
	komSession->sendCommand(cmdstr.c_string());
	return 0;
}

void KOMConferenceInfoCommand::sendOk(KOMString& userListReply)
{
	KOMTextData* txt = NULL;
	if (presentationNum > 0)
		txt = komSession->getTextData(presentationNum);
	KOMString re;
	// Format is editable supervisor, description, protected, original, 
	// secret, allow anonymous, forbid secret
	re.appendFormat(KOM_GROUP_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%d" 
		KOM_RECORD_SEPARATOR "%s"
		KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%d"
		KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%d"
		KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR, 
		(supervisor == komSession->getUserID())?1:0,
		supervisor, txt?txt->text.c_string():"", (confType>>7)&1,
		((confType)>>6)&1, (confType>>5)&1, (confType>>3)&1, 
		((confType)>>2)&1);
	int numUsers;
	if (!userListReply.parseArrayStart(numUsers))
	{
		error(KOM_ERROR_SYNTAX);
		return;
	}
	for (int i = 0; i < numUsers; ++i)
	{
		int user;
		if (!userListReply.parseInt(user))
		{
			error(KOM_ERROR_SYNTAX);
			return;
		}
		if (i > 0)
			re.appendFormat(",%d", user);
		else
			re.appendFormat("%d", user);
		time_t timeval;
		unsigned int type;
		if (!userListReply.parseInt(user) || !userListReply.parseTime(timeval) || !userListReply.parseBitfield(type))
		{
			error(KOM_ERROR_SYNTAX);
			return;
		}
	}
	ok(re.c_string());
}

bool KOMConferenceInfoCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (id != cmdID)
		return false;
	// If stat fails, fall through to get text which handles errors and continues
	if (!is_ok && state == CONF_INFO_PRESENTATION_STAT)
		state = CONF_INFO_PRESENTATION;
	if (!is_ok && state != CONF_INFO_PRESENTATION)
		error(reply.parseErrorCode());
	else if (state == CONF_INFO_USERLIST)
	{
		sendOk(reply);
	}
	else if (state == CONF_INFO_PRESENTATION_STAT)
	{
		time_t t;
		int temp;
		int textSize;
		is_ok = reply.parseTime(t);
		is_ok &= reply.parseInt(temp);
		is_ok &= reply.parseInt(temp);
		is_ok &= reply.parseInt(textSize);
		is_ok &= reply.parseInt(temp);
		int items;
		is_ok &= reply.parseArrayStart(items);
		for (int i = 0; is_ok && i < items; ++i)
		{
			int type;
			is_ok &= reply.parseInt(type);
			switch (type)
			{
			case 7:
			case 9:
				is_ok &= reply.parseTime(t);
				break;
			default:
				is_ok &= reply.parseInt(temp);
			}
		}
		is_ok &= reply.parseArrayEnd(items);
		is_ok &= reply.parseArrayStart(items);
		for (int i = 0; is_ok && i < items; ++i)
		{
			int auxNum;
			int auxTag;
			int auxCreator;
			time_t auxTime;
			unsigned int auxFlags;
			int auxInherit;
			KOMString auxData;
			is_ok &= reply.parseInt(auxNum);
			is_ok &= reply.parseInt(auxTag);
			is_ok &= reply.parseInt(auxCreator);
			is_ok &= reply.parseTime(auxTime);
			is_ok &= reply.parseBitfield(auxFlags);
			is_ok &= reply.parseInt(auxInherit);
			is_ok &= reply.parseString(auxData);
			if (is_ok && auxTag == 1)
			{
				presentationReformat = auxData.parseLiteral("text/x-kom-basic");
				// Content type
				auxData.advanceToAfter(';');
				if (auxData.parseLiteral("charset=utf-8"))
					presentationUTF8 = true;
			}
		}
		is_ok &= reply.parseArrayEnd(items);
		if (!is_ok)
		{
			error(KOM_ERROR_OOM);
			return true;
		}

		KOMString cmdstr;
		cmdID = komSession->getCommandID();
		if (cmdstr.appendFormat("%d 25 %d 0 %d\n", cmdID, presentationNum, textSize))
		{
			error(KOM_ERROR_OOM);
			return true;
		}
		state = CONF_INFO_PRESENTATION;
		komSession->sendCommand(cmdstr.c_string());
	}
	else if (state == CONF_INFO_PRESENTATION)
	{
		if (is_ok && !komSession->getTextData(presentationNum))
		{
			KOMTextData* text = new KOMTextData(presentationNum);
			if (!text)
			{
				error(KOM_ERROR_OOM);
				return true;
			}
			if (!reply.parseString(text->text))
			{
				delete text;
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			if (!presentationUTF8)
				text->text.latin1ToUTF8();
			if (presentationReformat)
				text->text.komToPlainText();
			komSession->addTextData(text);
		}
		if (numMembers)
		{
			KOMString cmdstr;
			cmdID = komSession->getCommandID();
			if (cmdstr.appendFormat("%d 101 %d 0 32768\n", cmdID, confNum))
			{
				error(KOM_ERROR_OOM);
				return true;
			}
			state = CONF_INFO_USERLIST;
			komSession->sendCommand(cmdstr.c_string());
			return true;
		}
		KOMString fakeReply;
		fakeReply.append("0 *");
		sendOk(fakeReply);
	}
	else
	{
		// Parse the conference info
		// The most interesting stuff at the moment is supervisor and presentation
		KOMString str;
		time_t timeval;
		int temp;
		// Name
		is_ok = reply.parseString(str);
		// Type
		if (is_ok)
			is_ok = reply.parseBitfield(confType);
		// Creation time
		if (is_ok)
			is_ok = reply.parseTime(timeval);
		// last written
		if (is_ok)
			is_ok = reply.parseTime(timeval);
		// Creator
		if (is_ok)
			is_ok = reply.parseInt(temp);
		// presentation
		if (is_ok)
			is_ok = reply.parseInt(presentationNum);
		// supervisor
		if (is_ok)
			is_ok = reply.parseInt(supervisor);
		// Permitted submitters
		if (is_ok)
			is_ok = reply.parseInt(temp);
		// Super conf
		if (is_ok)
			is_ok = reply.parseInt(temp);
		// motd
		if (is_ok)
			is_ok = reply.parseInt(temp);
		// nice
		if (is_ok)
			is_ok = reply.parseInt(temp);
		// keep commented
		if (is_ok)
			is_ok = reply.parseInt(temp);
		// num members
		if (is_ok)
			is_ok = reply.parseInt(numMembers);
		if (!is_ok)
		{
			error(KOM_ERROR_SYNTAX);
			return true;
		}
		if (presentationNum > 0)
		{
			if (!komSession->getTextData(presentationNum))
			{
				KOMString cmdstr;
				cmdID = komSession->getCommandID();
				if (cmdstr.appendFormat("%d 90 %d\n", cmdID, presentationNum))
				{
					error(KOM_ERROR_OOM);
					return true;
				}
				state = CONF_INFO_PRESENTATION_STAT;
				komSession->sendCommand(cmdstr.c_string());
				return true;
			}
		}
		if (numMembers)
		{
			KOMString cmdstr;
			cmdID = komSession->getCommandID();
			if (cmdstr.appendFormat("%d 101 %d 0 %d\n", cmdID, confNum, numMembers))
			{
				error(KOM_ERROR_OOM);
				return true;
			}
			state = CONF_INFO_USERLIST;
			komSession->sendCommand(cmdstr.c_string());
			return true;
		}
		KOMString fakeReply;
		fakeReply.append("0 *");
		sendOk(fakeReply);
	}
	return true;
}

