/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2009 Tim Johansson.  All rights reserved.
**
*/

#include "komfavorites.h"

const char* const KOMSetFavoriteCommand::commandString = "set_favorite";
const char* const KOMListFavoritesCommand::commandString = "list_favorites";

int KOMSetFavoriteCommand::Construct(KOMString& cmd)
{
	if (!cmd.parseLiteral(commandString) || !cmd.parseInt(textNum) || !cmd.parseInt(markType))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}

	cmdID = komSession->getCommandID();

	int err = 0;
	KOMString cmdstr;
	if (markType <= 0)
	{
		err = cmdstr.appendFormat("%d 73 %d\n", cmdID, textNum);
	}
	else
	{
		err = cmdstr.appendFormat("%d 72 %d 100\n", cmdID, textNum);
	}
	if (err)
	{
		error(KOM_ERROR_OOM);
		return 0;
	}

	komSession->sendCommand(cmdstr.c_string());
	return 0;
}

bool KOMSetFavoriteCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (id != cmdID)
		return false;
	if (!is_ok)
		error(reply.parseErrorCode());
	else
	{
		if (markType <= 0)
			komSession->removeFavorite(textNum);
		else
			komSession->addFavorite(textNum);
		ok("");
	}
	return true;
}

int KOMListFavoritesCommand::Construct(KOMString& cmd)
{
	if (!cmd.parseLiteral(commandString))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}

	cmdID = komSession->getCommandID();

	KOMString cmdstr;
	if (cmdstr.appendFormat("%d 23\n", cmdID))
	{
		error(KOM_ERROR_OOM);
		return 0;
	}

	numRequests = 0;
	komSession->sendCommand(cmdstr.c_string());
	return 0;
}

bool KOMListFavoritesCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	for (int i = 0; i < numFavorites; ++i)
	{
		if (id == favorites[i].statCmd)
		{
			komSession->addTextStat(favorites[i].textNum, is_ok, reply);
			favorites[i].statCmd = -1;
			KOMTextStat* txt = komSession->getGlobalTextStat(favorites[i].textNum);
			if (txt && txt->hasStat() && !txt->isUnavailable() && !komSession->getTextData(favorites[i].textNum))
			{
				// request text data
				favorites[i].textCmd = komSession->getCommandID();
				KOMString cmdstr;
				cmdstr.appendFormat("%d 25 %d 0 %d\n", favorites[i].textCmd, txt->getGlobalNum(), txt->getTextLength());
				komSession->sendCommand(cmdstr.c_string());
			}
			else
				--numRequests;
			if (!numRequests)
			{
				sendReply();
			}
			return true;
		}
		else if (id == favorites[i].textCmd)
		{
			if (is_ok && !komSession->getTextData(favorites[i].textNum))
			{
				KOMTextData* text = new KOMTextData(favorites[i].textNum);
				if (!text)
				{
					error(KOM_ERROR_OOM);
					return true;
				}
				if (!reply.parseString(text->text))
				{
					error(KOM_ERROR_PROTOCOL);
					return true;
				}
				KOMTextStat* txt = komSession->getGlobalTextStat(favorites[i].textNum);
				if (txt && !txt->getContentType().parseLiteral("text/"))
				{
					KOMString temp;
					temp.append(text->text.c_string(), text->text.len());
					temp.advanceToAfter('\n');
					temp.base64Encode();
					text->text.terminateAt('\n');
					text->text.append('\n');
					text->text.append(temp.c_string());
				}
				if (!txt || !txt->isUTF8())
					text->text.latin1ToUTF8();
				if (txt && txt->shouldReformat())
					text->text.komToPlainText();

				komSession->addTextData(text);
			}
			favorites[i].textCmd = -1;
			--numRequests;
			if (!numRequests)
			{
				sendReply();
			}
			return true;
		}
	}
	if (id != cmdID)
		return false;
	if (!is_ok)
		error(reply.parseErrorCode());
	else
	{
		if (!reply.parseArrayStart(numFavorites) || numFavorites < 0)
		{
			error(KOM_ERROR_SYNTAX);
			return true;
		}
		if (numFavorites == 0)
		{
			sendReply();
			return true;
		}
		favorites = new KOMFav[numFavorites];
		if (!favorites)
		{
			error(KOM_ERROR_OOM);
			return true;
		}
		for (int i = 0; i < numFavorites; ++i)
		{
			favorites[i].statCmd = -1;
			favorites[i].textCmd = -1;
			if (!reply.parseInt(favorites[i].textNum) || !reply.parseInt(favorites[i].markType))
			{
				error(KOM_ERROR_SYNTAX);
				return true;
			}
			KOMTextStat* txt = komSession->getGlobalTextStat(favorites[i].textNum);
			if (!txt || (!txt->hasStat() && !txt->isUnavailable()))
			{
				// request the text stat
				favorites[i].statCmd = komSession->getCommandID();
				KOMString cmdstr;
				cmdstr.appendFormat("%d 90 %d\n", favorites[i].statCmd, favorites[i].textNum);
				komSession->sendCommand(cmdstr.c_string());
				++numRequests;
			}
			else if (!txt->isUnavailable() && !komSession->getTextData(favorites[i].textNum))
			{
				// request the text data
				favorites[i].textCmd = komSession->getCommandID();
				KOMString cmdstr;
				cmdstr.appendFormat("%d 25 %d 0 %d\n", favorites[i].textCmd, txt->getGlobalNum(), txt->getTextLength());
				komSession->sendCommand(cmdstr.c_string());
				++numRequests;
			}
		}
		if (!numRequests)
		{
			sendReply();
		}
	}
	return true;
}

void KOMListFavoritesCommand::sendReply()
{
	KOMString reply;
	komSession->clearFavorites();
	for (int i = 0; i < numFavorites; ++i)
	{
		komSession->addFavorite(favorites[i].textNum);
		KOMTextStat* txt = komSession->getGlobalTextStat(favorites[i].textNum);
		if (!txt || !txt->hasStat() || txt->isUnavailable())
			continue;
		KOMTextData* txtdata = komSession->getTextData(favorites[i].textNum);
		KOMString title;
		if (txtdata)
		{
			title.append(txtdata->text.c_string());
			title.terminateAt('\n');
		}
		else
		{
			title.append("*Unavailable*");
		}
		reply.appendFormat(KOM_GROUP_SEPARATOR "%d"
				KOM_RECORD_SEPARATOR "%d"
				KOM_RECORD_SEPARATOR "%u"
				KOM_RECORD_SEPARATOR "%s" 
				KOM_RECORD_SEPARATOR "1",
				txt->getGlobalNum(), txt->getAuthor(), 
				(unsigned int)txt->getTimeval(), 
				title.c_string());
	}
	reply.appendFormat(KOM_GROUP_SEPARATOR "0");
	ok(reply.c_string());
}
