/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komlistallconf.h"

const char* const KOMListAllConferencesCommand::commandString = "list_all_conferences";
const char* const KOMListAllOnlineUsersCommand::commandString = "list_all_online_users";

KOMListAllConferencesCommand::~KOMListAllConferencesCommand()
{}

int KOMListAllConferencesCommand::Construct(KOMString& cmd)
{
	KOMString cmdstr;
	int confs, users;
	if (!cmd.parseLiteral(commandString) || !cmd.parseInt(confs) || !cmd.parseInt(users))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	KOMString search;
	if (!cmd.parseCommandString(search))
		search.clear();
	if (!search.UTF8ToLatin1())
	{
		error(KOM_ERROR_INVALID_CHARACTERS);
		return 0;
	}

	cmdID = komSession->getCommandID();
	if (cmdstr.appendFormat("%d 76 %dH%s %d %d\n", cmdID, search.len(), search.c_string(), users?1:0, confs?1:0))
	{
		error(KOM_ERROR_OOM);
		return 0;
	}
	komSession->sendCommand(cmdstr.c_string());
	return 0;
}

struct ConfInfo
{
	KOMString name;
	KOMString strippedName;
	unsigned int type;
	int num;
};
int conf_info_compare(const void* c1, const void* c2)
{
	return ((ConfInfo*)c1)->strippedName.compare(((ConfInfo*)c2)->strippedName);
}
bool KOMListAllConferencesCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (id != cmdID)
		return false;
	if (!is_ok)
		error(reply.parseErrorCode());
	else
	{
		KOMString restr;
		int numEntries;
		if (!reply.parseArrayStart(numEntries) || !numEntries)
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		ConfInfo* info = new ConfInfo[numEntries];
		int curEntry = 0;
		while (curEntry < numEntries)
		{
			if (!reply.parseString(info[curEntry].name))
				break;
			info[curEntry].name.latin1ToUTF8();
			if (!reply.parseBitfield(info[curEntry].type))
				break;
			if (!reply.parseInt(info[curEntry].num))
				break;
			info[curEntry].strippedName.append(info[curEntry].name.c_string());
			info[curEntry].strippedName.stripConfName();
			komSession->updateConferenceName(info[curEntry].num, info[curEntry].name);
			++curEntry;
		}
		if (curEntry < numEntries)
		{
			delete[] info;
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		qsort(info, numEntries, sizeof(ConfInfo), conf_info_compare);
		curEntry = 0;
		while (curEntry < numEntries)
		{
			// highest bit of type is the "protected" bit
			if (restr.appendFormat(KOM_GROUP_SEPARATOR "%d" 
					       KOM_RECORD_SEPARATOR "%s"
					       KOM_RECORD_SEPARATOR "%d"
					       KOM_RECORD_SEPARATOR "%d",
					       info[curEntry].num, info[curEntry].name.c_string(), ((info[curEntry].type>>7)&1), ((info[curEntry].type>>4)&1)))
			{
				delete[] info;
				error(KOM_ERROR_OOM);
				return true;
			}
			++curEntry;
		}
		delete[] info;
		ok(restr.c_string());
	}
	return true;
}

KOMListAllOnlineUsersCommand::~KOMListAllOnlineUsersCommand()
{}

int KOMListAllOnlineUsersCommand::Construct(KOMString& cmd)
{
	KOMString cmdstr;
	if (!cmd.parseLiteral(commandString))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	cmdID = komSession->getCommandID();
	if (cmdstr.appendFormat("%d 83 1 0 0\n", cmdID))
	{
		error(KOM_ERROR_OOM);
		return 0;
	}
	komSession->sendCommand(cmdstr.c_string());
	return 0;
}

bool KOMListAllOnlineUsersCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (id != cmdID)
		return false;
	if (!is_ok)
		error(reply.parseErrorCode());
	else
	{
		KOMString restr;
		int numEntries;
		if (!reply.parseArrayStart(numEntries) || !numEntries)
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		for (int i = 0; i < numEntries; ++i)
		{
			int session, person, conf, idle;
			unsigned int flags;
			KOMString doing;
			if (!reply.parseInt(session) || !reply.parseInt(person) || !reply.parseInt(conf) || 
				!reply.parseInt(idle) || !reply.parseBitfield(flags) || !reply.parseString(doing))
			{
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			doing.latin1ToUTF8();
			// second highest bit signals that idle is valid
			if (restr.appendFormat(KOM_GROUP_SEPARATOR "%d" 
					       KOM_RECORD_SEPARATOR "%d"
					       KOM_RECORD_SEPARATOR "%d"
					       KOM_RECORD_SEPARATOR "%s",
					       person, conf, ((flags>>6)&1)?idle:-1, doing.c_string()))
			{
				error(KOM_ERROR_OOM);
				return true;
			}
		}
		ok(restr.c_string());
	}
	return true;
}

