/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komlistmessages.h"

const char* const KOMListMessagesCommand::commandString = "list_messages";

KOMListMessagesCommand::~KOMListMessagesCommand()
{
	while (requests)
	{
		MessageRequest* nr = requests->next;
		delete requests;
		requests = nr;
	}
}

#define MAX_MESSAGE_REQUESTS 128

int KOMListMessagesCommand::Construct(KOMString& cmd)
{
	// 107 (query-read-texts)
	// 90 (get-text-stat)
	// 25 (get-text)
	KOMString cmdstr;
	if (!cmd.parseLiteral(commandString) || !cmd.parseInt(baseTextNum) || !cmd.parseInt(conferenceNum))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	if (komSession->getConference(conferenceNum))
		komSession->changeCurrentConference(conferenceNum);

	numRequests = 0;

	requestedBaseTextNum = baseTextNum;
	rootTextRequest = -1;
	findRootText();
	return 0;
}

void KOMListMessagesCommand::findRootText()
{
	KOMTextStat* txt = komSession->getGlobalTextStat(baseTextNum);
	if (txt && txt->isUnavailable())
	{
		ok(KOM_GROUP_SEPARATOR "0");
		return;
	}
	if (!txt || !txt->hasStat())
	{
		// Request this text
		rootTextRequest = komSession->getCommandID();
		requestedRootNum = baseTextNum;
		KOMString cmdstr;
		cmdstr.appendFormat("%d 90 %d\n", rootTextRequest, baseTextNum);
		komSession->sendCommand(cmdstr.c_string());
		return;
	}
	while (true)
	{
		if (txt->getCommentTo() <= 0)
		{
			sendTextRequests();
			return;
		}
		requestedRootNum = txt->getCommentTo();
		txt = komSession->getGlobalTextStat(requestedRootNum);
		if (txt && txt->isUnavailable())
		{
			sendTextRequests();
			return;
		}
		if (!txt || !txt->hasStat())
		{
			// Request the text
			rootTextRequest = komSession->getCommandID();
			KOMString cmdstr;
			cmdstr.appendFormat("%d 90 %d\n", rootTextRequest, requestedRootNum);
			komSession->sendCommand(cmdstr.c_string());
			return;
		}
		baseTextNum = txt->getGlobalNum();
	}
}

void KOMListMessagesCommand::sendTextRequests()
{
	KOMTextStat* txt = komSession->getGlobalTextStat(baseTextNum);
	if (txt && txt->isUnavailable())
	{
		// FIXME: not sure what should be done in this case, send an error?
		ok(KOM_GROUP_SEPARATOR "0");
		return;
	}
	if (!txt || !txt->hasStat())
	{
		// request
		MessageRequest* req = new MessageRequest();
		if (!req)
		{
			error(KOM_ERROR_OOM);
			return;
		}
		req->cmdId = komSession->getCommandID();
		req->textNum = baseTextNum;
		req->stat = true;
		req->next = requests;
		requests = req;
		++numRequests;
		KOMString cmdstr;
		cmdstr.appendFormat("%d 90 %d\n", req->cmdId, req->textNum);
		komSession->sendCommand(cmdstr.c_string());
		return;
	}
	// To traverse the tree the following is done:
	// Start at the root
	// Go to the first child while there is one
	// If there is no first child, go t the next sibling if there is one
	// If there is no next sibling go up to the parent until there is a 
	// next sibling and go to that
	// This results in a depth first traverse of the tree
	// FIXME: this will break if the tree is broken, so the tree should be verified better
	while (numRequests < MAX_MESSAGE_REQUESTS)
	{
		if (!txt->hasStat())
		{
			// request
			// must make sure this request is not sent already
			bool new_request = true;
			for (MessageRequest* r = requests; r; r = r->next)
				if (r->textNum == txt->getGlobalNum())
					new_request = false;
			if (new_request)
			{
				MessageRequest* req = new MessageRequest();
				if (!req)
				{
					error(KOM_ERROR_OOM);
					return;
				}
				req->cmdId = komSession->getCommandID();
				req->textNum = txt->getGlobalNum();
				req->stat = true;
				req->next = requests;
				requests = req;
				++numRequests;
				KOMString cmdstr;
				cmdstr.appendFormat("%d 90 %d\n", req->cmdId, req->textNum);
				komSession->sendCommand(cmdstr.c_string());
			}
			// go back one step and continue to a next comment
			while (txt && txt->getCommentTo() >= 0 && txt->getNextComment() < 0)
				txt = komSession->getGlobalTextStat(txt->getCommentTo());
			if (!txt)
			{
				error(KOM_ERROR_PROTOCOL);
				return;
			}
			if (txt->getCommentTo() < 0)
				break;
			txt = komSession->getGlobalTextStat(txt->getNextComment());
		}
		else if (txt->getFirstComment() >= 0)
		{
			// Follow first comment
			txt = komSession->getGlobalTextStat(txt->getFirstComment());
		}
		else if (txt->getNextComment() >= 0)
		{
			// go to next comment
			txt = komSession->getGlobalTextStat(txt->getNextComment());
		}
		else if (txt->getCommentTo() >= 0)
		{
			// go back one step and continue to a next comment
			while (txt->getCommentTo() >= 0 && txt->getNextComment() < 0)
				txt = komSession->getGlobalTextStat(txt->getCommentTo());
			if (!txt)
			{
				error(KOM_ERROR_PROTOCOL);
				return;
			}
			if (txt->getCommentTo() < 0)
				break;
			txt = komSession->getGlobalTextStat(txt->getNextComment());
		}
		else
			break;
		if (!txt)
		{
			error(KOM_ERROR_PROTOCOL);
			return;
		}
	}
	// If we could traverse the tree and there are not active requests for stats, send all requests for texts
	if (!requests)
	{
		// All stats are in the cache, request the texts
		KOMTextStat* txt = komSession->getGlobalTextStat(baseTextNum);
		while (numRequests < MAX_MESSAGE_REQUESTS)
		{
			if (!komSession->getTextData(txt->getGlobalNum()))
			{
				// request
				MessageRequest* req = new MessageRequest();
				if (!req)
				{
					error(KOM_ERROR_OOM);
					return;
				}
				req->cmdId = komSession->getCommandID();
				req->textNum = txt->getGlobalNum();
				req->stat = false;
				req->next = requests;
				requests = req;
				++numRequests;
				KOMString cmdstr;
				cmdstr.appendFormat("%d 25 %d 0 %d\n", req->cmdId, txt->getGlobalNum(), txt->getTextLength());
				komSession->sendCommand(cmdstr.c_string());
			}
			if (txt->getFirstComment() >= 0)
			{
				// Follow first comment
				txt = komSession->getGlobalTextStat(txt->getFirstComment());
			}
			else if (txt->getNextComment() >= 0)
			{
				// go to next comment
				txt = komSession->getGlobalTextStat(txt->getNextComment());
			}
			else if (txt->getCommentTo() >= 0)
			{
				// go back one step and continue to a next comment
				while (txt->getCommentTo() >= 0 && txt->getNextComment() < 0)
					txt = komSession->getGlobalTextStat(txt->getCommentTo());
				if (!txt)
				{
					error(KOM_ERROR_PROTOCOL);
					return;
				}
				if (txt->getCommentTo() < 0)
					break;
				txt = komSession->getGlobalTextStat(txt->getNextComment());
			}
			else
				break;
			if (!txt)
			{
				error(KOM_ERROR_PROTOCOL);
				return;
			}
		}
		if (!requests)
			sendReply();
	}
}

void KOMListMessagesCommand::progressStat(KOMString& re, bool is_ok, MessageRequest* req)
{
	// Text stat
	komSession->addTextStat(req->textNum, is_ok, re);
	MessageRequest** nextp = &requests;
	for (MessageRequest* r = requests; r; r = r->next)
	{
		if (r == req)
		{
			*nextp = req->next;
			--numRequests;
			break;
		}
		nextp = &r->next;
	}
	delete req;
	sendTextRequests();
}

void KOMListMessagesCommand::progressText(KOMString& re, bool is_ok, MessageRequest* req)
{
	if (is_ok && !komSession->getTextData(req->textNum))
	{
		KOMTextData* text = new KOMTextData(req->textNum);
		if (!text)
		{
			error(KOM_ERROR_OOM);
			return;
		}
		if (!re.parseString(text->text))
		{
			error(KOM_ERROR_PROTOCOL);
			return;
		}
		KOMTextStat* txt = komSession->getGlobalTextStat(req->textNum);
		if (txt && !txt->getContentType().parseLiteral("text/"))
		{
			KOMString temp;
			temp.append(text->text.c_string(), text->text.len());
			temp.advanceToAfter('\n');
			temp.base64Encode();
			text->text.terminateAt('\n');
			text->text.append('\n');
			text->text.append(temp.c_string());
		}
		if (!txt || !txt->isUTF8())
			text->text.latin1ToUTF8();
		if (txt && txt->shouldReformat())
			text->text.komToPlainText();

		komSession->addTextData(text);
	}

	// Remove the request from the lsit and delete it
	MessageRequest** nextp = &requests;
	for (MessageRequest* r = requests; r; r = r->next)
	{
		if (r == req)
		{
			*nextp = req->next;
			--numRequests;
			break;
		}
		nextp = &r->next;
	}
	delete req;
	sendTextRequests();
}

bool KOMListMessagesCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (id == rootTextRequest)
	{
		komSession->addTextStat(requestedRootNum, is_ok, reply);
		rootTextRequest = -1;
		findRootText();
		return true;
	}
	MessageRequest* r = requests;
	while (r && r->cmdId != id)
		r = r->next;
	if (!r)
	{
		return false;
	}
	// First build the tree of all messages, then get all the message texts
	if (r->stat)
		progressStat(reply, is_ok, r);
	else
		progressText(reply, is_ok, r);
	return true;
}

void KOMListMessagesCommand::sendReply()
{
	KOMString reply;
	// All stats are in the cache, request the texts
	KOMTextStat* txt = komSession->getGlobalTextStat(baseTextNum);
	KOMConferenceStat* conf = komSession->getConference(conferenceNum);
	reply.appendFormat(KOM_GROUP_SEPARATOR "%d", komSession->isIgnored(txt->getGlobalNum()));
	while (1)
	{
		// FIXME: OOM
		reply.appendFormat(KOM_GROUP_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR,
			txt->getGlobalNum(), txt->getAuthor());
		if (txt->getNumTo()>0)
		{
			reply.appendFormat("%d", txt->getTo(0));
		}
		for (int i = 1; i < txt->getNumTo(); ++i)
		{
			reply.appendFormat(",%d", txt->getTo(i));
		}
		reply.append(KOM_RECORD_SEPARATOR);
		if (txt->getNumCc()>0)
		{
			reply.appendFormat("%d", txt->getCc(0));
		}
		for (int i = 1; i < txt->getNumCc(); ++i)
		{
			reply.appendFormat(",%d", txt->getCc(i));
		}
		int is_read = 0;
		if (conf)
		{
			is_read = conf->isTextRead(txt->getLocalNum(conf->getNum()), false)?1:0;
			// Ignored texts are marked with a flag
			if (conf->isIgnored(txt->getLocalNum(conf->getNum())))
				is_read += 2;
		}
		else if (conferenceNum <= 0)
		{
			// When viewing a text by number the requested text should be the equivalent of unread
			is_read = txt->getGlobalNum() == requestedBaseTextNum?0:1;
		}
		reply.appendFormat(KOM_RECORD_SEPARATOR "%u" KOM_RECORD_SEPARATOR "%d" 
			KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%d" KOM_RECORD_SEPARATOR,  
			(unsigned int)txt->getTimeval(), is_read, komSession->isFavorite(txt->getGlobalNum())?1:0, txt->getNumComments());

		reply.appendFormat("%s" KOM_RECORD_SEPARATOR, txt->getContentType().c_string());
		// FIXME: OOM
		KOMTextData* td = komSession->getTextData(txt->getGlobalNum());
		if (td)
			reply.append(td->text.c_string());
		else
			reply.append("*Unavailable*");
		if (txt->getFirstComment() >= 0)
		{
			// Follow first comment
			txt = komSession->getGlobalTextStat(txt->getFirstComment());
		}
		else if (txt->getNextComment() >= 0)
		{
			// go to next comment
			txt = komSession->getGlobalTextStat(txt->getNextComment());
		}
		else if (txt->getCommentTo() >= 0)
		{
			// go back one step and continue to a next comment
			while (txt->getCommentTo() >= 0 && txt->getNextComment() < 0)
				txt = komSession->getGlobalTextStat(txt->getCommentTo());
			if (!txt)
			{
				error(KOM_ERROR_PROTOCOL);
				return;
			}
			if (txt->getCommentTo() < 0)
				break;
			txt = komSession->getGlobalTextStat(txt->getNextComment());
		}
		else
			break;
		if (!txt)
		{
			error(KOM_ERROR_PROTOCOL);
			return;
		}
	}
	ok(reply.c_string());
}

