/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komlistsubconf.h"

const char* const KOMListSubscribedConferencesCommand::commandString = "list_subscribed_conferences";

KOMListSubscribedConferencesCommand::~KOMListSubscribedConferencesCommand()
{
	delete[] confRequests;
}

int KOMListSubscribedConferencesCommand::Construct(KOMString& cmd)
{
	listCmdId = -1;
	unreadCmdId = -1;
	numConfRequests = 0;
	// Mark all as read
	int numConfs = 0;
	for (KOMConferenceStat* c = komSession->getFirstConference(); c; c = c->nextConference())
	{
		c->setRead(true);
		++numConfs;
	}
	int err;
	if (komSession->isConferenceListDirty())
	{
		KOMString cmdstr;
		status = UPDATE_CONF_LIST;
		listCmdId = komSession->getCommandID();
		if (komSession->getProtocolVersion()<11)
			err = cmdstr.appendFormat("%d 99 %d 0 1000 1\n", listCmdId, komSession->getUserID());
		else
			err = cmdstr.appendFormat("%d 108 %d 0 1000 1 0\n", listCmdId, komSession->getUserID());
		if (!err)
			komSession->sendCommand(cmdstr.c_string());
	}
	else
	{
		status = UPDATE_CONF_MESSAGE_LIST;
		confRequests = new int[numConfs*2];
		if (!confRequests)
		{
			error(KOM_ERROR_OOM);
			return 0;
		}
		err = 0;
		for (KOMConferenceStat* c = komSession->getFirstConference(); c && !err; c = c->nextConference())
		{
			// Always update passive meetings since there is no notification of new messages there
			if (c->hasNewMessages() || c->isPassive())
			{
				c->setNewMessages(false);
				KOMString cmdstr;
				confRequests[numConfRequests*2] = komSession->getCommandID();
				confRequests[numConfRequests*2+1] = c->getNum();
				err = cmdstr.appendFormat("%d 103 %d %d 255\n", confRequests[numConfRequests*2], c->getNum(), c->getHighestLocal()+1);
				if (!err)
					komSession->sendCommand(cmdstr.c_string());
				++numConfRequests;
			}
		}
		if (!numConfRequests)
		{
			status = UPDATE_CONF_UNREAD;
			// send query read texts
			unreadCmdId = komSession->getCommandID();
			KOMString restr;
			if (restr.appendFormat("%d 52 %d\n", unreadCmdId, komSession->getUserID()))
				error(KOM_ERROR_OOM);
			else
				komSession->sendCommand(restr.c_string());
		}
	}
	if (err)
	{
		error(KOM_ERROR_OOM);
		return 0;
	}
	return 0;
}

bool KOMListSubscribedConferencesCommand::progressConfList(int id, KOMString& re, bool is_ok)
{
	if (!is_ok)
	{
		error(re.parseErrorCode());
		return true;
	}
	int numConfs;
	if (!re.parseArrayStart(numConfs))
	{
		error(KOM_ERROR_PROTOCOL);
		return true;
	}
	if (!numConfs)
	{
		komSession->markConferenceListDirty(false);
		ok("");
		return true;
	}
	confRequests = new int[numConfs*2];
	if (!confRequests)
	{
		error(KOM_ERROR_OOM);
		return true;
	}
	for (int i = 0; i < numConfs; ++i)
	{
		int confNum;
		int temp;
		// position
		bool res = re.parseInt(temp);
		// time
		time_t timeval;
		res &= re.parseTime(timeval);
		// id
		res &= re.parseInt(confNum);
		// prio
		res &= re.parseInt(temp);

		if (!res)
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		KOMConferenceStat* c = komSession->getConference(confNum);
		if (c && c->getName())
			c = NULL;
		else
		{
			if (!c)
			{
				c = new KOMConferenceStat(confNum);
				if (!c)
				{
					error(KOM_ERROR_OOM);
					return true;
				}
			}
			c->setRead(true);
			komSession->addConference(c);
			
			status = UPDATE_CONF_NAME;
			KOMString cmdstr;
			confRequests[numConfRequests*2] = komSession->getCommandID();
			confRequests[numConfRequests*2+1] = confNum;
			if (cmdstr.appendFormat("%d 78 %d\n", confRequests[numConfRequests*2], confNum))
			{
				error(KOM_ERROR_OOM);
				return true;
			}
			komSession->sendCommand(cmdstr.c_string());
			++numConfRequests;
		}
		// read range
		if (komSession->getProtocolVersion() < 11)
		{
			// last read text
			res &= re.parseInt(temp);
			if (c)
			{
				for (int lt = 1; lt <= temp; ++lt)
					c->addReadText(lt);
			}
			//res &= re.parseSizeOnlyArray(temp);
			int smallarray;
			res &= re.parseArrayStart(smallarray);
			for (int j = 0; j < smallarray; ++j)
			{
				// first read
				res &= re.parseInt(temp);
				if (res && c)
					c->addReadText(temp);
			}
			res &= re.parseArrayEnd(smallarray);
		}
		else
		{
			int smallarray;
			//res &= re.parseSizeOnlyArray(smallarray);
			res &= re.parseArrayStart(smallarray);
			for (int j = 0; j < smallarray; ++j)
			{
				int first, last;
				// first read
				res &= re.parseInt(first);
				// last read
				res &= re.parseInt(last);
				if (res && c)
				{
					for (int tn = first; tn <= last; ++tn)
						c->addReadText(tn);
				}
			}
			res &= re.parseArrayEnd(smallarray);
		}
		// added by
		res &= re.parseInt(temp);
		// time
		res &= re.parseTime(timeval);
		// type (bitfield)
		unsigned int type;
		res &= re.parseBitfield(type);
		if (!res)
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		c = komSession->getConference(confNum);
		c->setInvite((type>>7)&1);
		c->setPassive((type>>6)&1, (type>>4)&1);
		c->setSecret((type>>5)&1);
	}

	if (!numConfRequests)
	{
		status = UPDATE_CONF_UNREAD;
		// send query read texts
		unreadCmdId = komSession->getCommandID();
		KOMString restr;
		if (restr.appendFormat("%d 52 %d\n", unreadCmdId, komSession->getUserID()))
			error(KOM_ERROR_OOM);
		else
			komSession->sendCommand(restr.c_string());
	}
	return true;
}

bool KOMListSubscribedConferencesCommand::progressConfName(int id, KOMString& re, bool is_ok)
{
	for (int i = 0; i < numConfRequests; ++i)
	{
		if (confRequests[i*2] == id)
		{
			if (!is_ok)
			{
				error(re.parseErrorCode());
				return true;
			}
			KOMString name;
			unsigned int type;
			int highest_local;
			if (!re.parseString(name) || !re.parseBitfield(type) || !re.parseInt(highest_local))
			{
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			name.latin1ToUTF8();
			komSession->updateConferenceName(confRequests[i*2+1], name);
			KOMConferenceStat* c = komSession->getConference(confRequests[i*2+1]);
			c->setName(name.c_string());
			c->setHighestLocal(highest_local);
			if (i != numConfRequests-1)
			{
				confRequests[i*2] = confRequests[(numConfRequests-1)*2];
				confRequests[i*2+1] = confRequests[(numConfRequests-1)*2+1];
			}
			--numConfRequests;
			if (!numConfRequests)
			{
				status = UPDATE_CONF_UNREAD;
				// send query read texts
				unreadCmdId = komSession->getCommandID();
				KOMString restr;
				if (restr.appendFormat("%d 52 %d\n", unreadCmdId, komSession->getUserID()))
					error(KOM_ERROR_OOM);
				else
					komSession->sendCommand(restr.c_string());
			}
			return true;
		}
	}
	return false;
}

bool KOMListSubscribedConferencesCommand::progressConfMessageList(int id, KOMString& re, bool is_ok)
{
	for (int i = 0; i < numConfRequests; ++i)
	{
		if (confRequests[i*2] == id)
		{
			if (is_ok)
			{
				KOMConferenceStat* conf = komSession->getConference(confRequests[i*2+1]);

				int firstText, lastText, more, dense;
				if (!re.parseInt(firstText) ||
						!re.parseInt(lastText) ||
						!re.parseInt(more) ||
						!re.parseInt(dense))
				{
					error(KOM_ERROR_PROTOCOL);
					return true;
				}
				if (dense)
				{
					int firstLocal, numTexts;
					if (!re.parseInt(firstLocal) || !re.parseArrayStart(numTexts))
					{
						error(KOM_ERROR_PROTOCOL);
						return true;
					}
					for (int i = 0; i < numTexts; ++i)
					{
						int textNum;
						if (!re.parseInt(textNum))
						{
							error(KOM_ERROR_PROTOCOL);
							return true;
						}
						if (textNum > 0)
						{
							if (komSession->addEmptyTextStat(textNum, firstLocal, conf->getNum()))
							{
								error(KOM_ERROR_OOM);
								return true;
							}
						}
						++firstLocal;
					}
				}
				else
				{
					int numTexts;
					if (!re.parseArrayStart(numTexts))
					{
						error(KOM_ERROR_PROTOCOL);
						return true;
					}
					for (int i = 0; i < numTexts; ++i)
					{
						int localNum, textNum;
						if (!re.parseInt(localNum) ||
								!re.parseInt(textNum))
						{
							error(KOM_ERROR_PROTOCOL);
							return true;
						}
						if (textNum > 0 && localNum > 0)
						{
							if (komSession->addEmptyTextStat(textNum, localNum, conf->getNum()))
							{
								error(KOM_ERROR_OOM);
								return true;
							}
						}
					}
				}
	
				conf->updateCacheRange(conf->getHighestLocal()+1, conf->getHighestLocal() <= 0);
				conf->setHighestLocal(lastText-1);
				// If there are even more texts, send a request for them
				if (more)
				{
					KOMString cmdstr;
					confRequests[i*2] = komSession->getCommandID();
					int err = cmdstr.appendFormat("%d 103 %d %d 255\n", confRequests[i*2], confRequests[i*2+1], conf->getHighestLocal()+1);
					if (err)
					{
						error(KOM_ERROR_OOM);
						return true;
					}
					komSession->sendCommand(cmdstr.c_string());
					return true;
				}
			}

			if (i != numConfRequests-1)
			{
				confRequests[i*2] = confRequests[(numConfRequests-1)*2];
				confRequests[i*2+1] = confRequests[(numConfRequests-1)*2+1];
			}
			--numConfRequests;
			if (!numConfRequests)
			{
				status = UPDATE_CONF_UNREAD;
				// send query read texts
				unreadCmdId = komSession->getCommandID();
				KOMString restr;
				if (restr.appendFormat("%d 52 %d\n", unreadCmdId, komSession->getUserID()))
					error(KOM_ERROR_OOM);
				else
					komSession->sendCommand(restr.c_string());
			}
			return true;
		}
	}
	return false;
}

bool KOMListSubscribedConferencesCommand::progressUnread(int id, KOMString& re, bool is_ok)
{
	if (!is_ok)
	{
		error(re.parseErrorCode());
		return true;
	}
	int num_unread;
	if (!re.parseArrayStart(num_unread))
	{
		error(KOM_ERROR_PROTOCOL);
		return true;
	}
	for (int i = 0; i < num_unread; ++i)
	{
		int unread_num;
		if (!re.parseInt(unread_num))
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		KOMConferenceStat* c = komSession->getConference(unread_num);
		if (c)
		{
			c->setRead(false);
		}
	}

	numConfRequests = 0;
	status = UPDATE_CONF_VERIFY_UNREAD;
	verifyUnread();
	return true;
}

bool KOMListSubscribedConferencesCommand::progressVerifyUnread(int id, KOMString& re, bool is_ok)
{
	for (int i = 0; i < numConfRequests; ++i)
	{
		if (confRequests[i*2] == id)
		{
			if (!is_ok)
			{
				error(re.parseErrorCode());
				return true;
			}
			KOMConferenceStat* conf = komSession->getConference(confRequests[i*2+1]);

			// update the cache
			int firstText, lastText, more, dense;
			if (!re.parseInt(firstText) ||
					!re.parseInt(lastText) ||
					!re.parseInt(more) ||
					!re.parseInt(dense))
			{
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			if (dense)
			{
				int firstLocal, numTexts;
				if (!re.parseInt(firstLocal) || !re.parseArrayStart(numTexts))
				{
					error(KOM_ERROR_PROTOCOL);
					return true;
				}
				for (int i = 0; i < numTexts; ++i)
				{
					int textNum;
					if (!re.parseInt(textNum))
					{
						error(KOM_ERROR_PROTOCOL);
						return true;
					}
					if (textNum > 0)
					{
						if (komSession->addEmptyTextStat(textNum, firstLocal, conf->getNum()))
						{
							error(KOM_ERROR_OOM);
							return true;
						}
					}
					++firstLocal;
				}
			}
			else
			{
				int numTexts;
				if (!re.parseArrayStart(numTexts))
				{
					error(KOM_ERROR_PROTOCOL);
					return true;
				}
				for (int i = 0; i < numTexts; ++i)
				{
					int localNum, textNum;
					if (!re.parseInt(localNum) ||
							!re.parseInt(textNum))
					{
						error(KOM_ERROR_PROTOCOL);
						return true;
					}
					if (textNum > 0 && localNum > 0)
					{
						if (komSession->addEmptyTextStat(textNum, localNum, conf->getNum()))
						{
							error(KOM_ERROR_OOM);
							return true;
						}
					}
				}
			}
			
			if (komSession->getProtocolVersion() < 11)
			{
				int firstReqLocToGlob = conf->getFirstCached()>=0?conf->getFirstCached():conf->getHighestLocal();
				firstReqLocToGlob -= 254;
				if (firstReqLocToGlob < 1)
					firstReqLocToGlob = 1;
				conf->updateCacheRange(firstReqLocToGlob, firstReqLocToGlob <= 1);
			}
			else
			{
				conf->updateCacheRange(firstText, !more);
			}

			if (i != numConfRequests-1)
			{
				confRequests[i*2] = confRequests[(numConfRequests-1)*2];
				confRequests[i*2+1] = confRequests[(numConfRequests-1)*2+1];
			}
			--numConfRequests;

			verifyUnread(conf->getNum());
			return true;
		}
	}
	return false;
}
void KOMListSubscribedConferencesCommand::verifyUnread(int confNum)
{
	// Verify that there really was unread messages in the conferences marked as unread
	for (KOMConferenceStat* c = komSession->getFirstConference(); c; c = c->nextConference())
	{
		if (confNum >= 0 && confNum != c->getNum())
			continue;
		if (!c->isRead())
		{
			int first_unread = c->findNextLowerUnread(c->getHighestLocal());
			while (first_unread > 0)
			{
				if (c->getFirstCached() < 0 || first_unread < c->getFirstCached())
				{
					if (c->isOldestCached())
					{
						first_unread = -1;
					}
					else
					{
						KOMString cmdstr;
						// request more texts to the cache and then retry
						confRequests[numConfRequests*2] = komSession->getCommandID();
						confRequests[numConfRequests*2+1] = c->getNum();
						if (komSession->getProtocolVersion() < 11)
						{
							int firstReqLocToGlob = c->getFirstCached()>=0?c->getFirstCached():c->getHighestLocal();
							firstReqLocToGlob -= 254;
							if (firstReqLocToGlob < 1)
								firstReqLocToGlob = 1;
							// FIXME: OOM
							cmdstr.appendFormat("%d 103 %d %d 255\n", confRequests[numConfRequests*2], c->getNum(), firstReqLocToGlob);
						}
						else
							// FIXME: OOM
							cmdstr.appendFormat("%d 121 %d %d 255\n", confRequests[numConfRequests*2], c->getNum(), first_unread+1);
						komSession->sendCommand(cmdstr.c_string());
						++numConfRequests;
					}
					break;
				}
				if (komSession->getGlobalTextStat(c->getGlobalTextNum(first_unread)) && 
					!komSession->getGlobalTextStat(c->getGlobalTextNum(first_unread))->isUnavailable())
					break; // Found an unread text, there are unread in this conference
				first_unread = c->findNextLowerUnread(first_unread - 1);
			}
			if (first_unread <= 0)
				c->setRead(true);
		}
	}

	if (!numConfRequests)
	{
		KOMString restr;
		int err = 0;
		for (KOMConferenceStat* c = komSession->getFirstConference(); c && !err; c = c->nextConference())
		{
			err = restr.appendFormat(KOM_GROUP_SEPARATOR "%d"
					KOM_RECORD_SEPARATOR "%s"
					KOM_RECORD_SEPARATOR "%d"
					KOM_RECORD_SEPARATOR "%d" 
					KOM_RECORD_SEPARATOR "%d" 
					KOM_RECORD_SEPARATOR "%d" 
					KOM_RECORD_SEPARATOR "%d", 
					c->getNum(), c->getName(), c->isRead()?1:0, 
					c->isInvite()?1:0, c->isSecret()?1:0, c->isPassive()?1:0,
					c->isInstantPassive()?1:0);
		}
		if (err)
			error(KOM_ERROR_OOM);
		else
		{
			komSession->markConferenceListDirty(false);
			ok(restr.c_string());
		}
	}
}

bool KOMListSubscribedConferencesCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (status == UPDATE_CONF_LIST)
	{
		return progressConfList(id, reply, is_ok);
	}
	else if (status == UPDATE_CONF_UNREAD)
	{
		return progressUnread(id, reply, is_ok);
	}
	else if (status == UPDATE_CONF_VERIFY_UNREAD)
	{
		return progressVerifyUnread(id, reply, is_ok);
	}
	else if (status == UPDATE_CONF_MESSAGE_LIST)
	{
		return progressConfMessageList(id, reply, is_ok);
	}
	return progressConfName(id, reply, is_ok);
}

