/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komlistthreads.h"

class KOMThread : public KOMHashEntry
{
public:
	KOMThread(int textNum) : baseTextNum(textNum), newestLocalNum(-1)
	{}
	virtual ~KOMThread()
	{}
	virtual int getHashKey(int id){return baseTextNum;}
	virtual KOMHashEntry* &getNextHashEntry(int id){return next;}

	int baseTextNum;
	int newestLocalNum;
	KOMHashEntry* next;
};

const char* const KOMListThreadsCommand::commandString = "list_threads";

KOMListThreadsCommand::~KOMListThreadsCommand()
{
	threads.deleteAll();
	delete[] requests;
	if (conf)
	{
		conf->decRef();
	}
}

int KOMListThreadsCommand::Construct(KOMString& cmd)
{
	// 107 (query-read-texts)
	// 121 (local-to-global-reverse)
	// 90 (get-text-stat)
	// 25 (get-text)
	KOMString cmdstr;
	int conferenceNum;
	if (!cmd.parseLiteral(commandString) || !cmd.parseInt(conferenceNum) || !cmd.parseInt(firstThread) || !cmd.parseInt(numThreads))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	
	conf = komSession->getConference(conferenceNum);
	if (!conf)
	{
		// FIXME: Not really a syntax error
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	komSession->changeCurrentConference(conferenceNum);
	conf->addRef();

	// We need to find all threads before the requested ones, we need at least this many requests
	// Try to get one more than needed so we can find out if there are more threads after these
	maxRequests = firstThread+numThreads+1;
	activeRequests = 0;
	requests = new TextRequest[maxRequests];
	if (!requests)
	{
		error(KOM_ERROR_OOM);
		return 0;
	}
	for (int i = 0; i < maxRequests; ++i)
	{
		requests[i].cmdId = -1;
		requests[i].reqTextNum = -1;
		requests[i].baseLocalNum = -1;
	}

	/*
	Go through all entries from newest to oldest, if is is in the rage
	which should be returned, build the entire thread to check if it is read
	or not. Optimization would be to only check if the newest is read.
	*/
	locToGlobCmdId = komSession->getCommandID();
	newestLocal = conf->getHighestLocal();
	currentText = newestLocal;
	if (conf->getFirstCached()<0)
	{
		// Get the newest texts
		if (komSession->getProtocolVersion() < 11)
		{
			firstReqLocToGlob = newestLocal-254;
			if (firstReqLocToGlob < 1)
				firstReqLocToGlob = 1;
			// FIXME: OOM
			cmdstr.appendFormat("%d 103 %d %d 255\n", locToGlobCmdId, conf->getNum(), firstReqLocToGlob);
		}
		else
			// FIXME: OOM
			cmdstr.appendFormat("%d 121 %d 0 255\n", locToGlobCmdId, conf->getNum());
		komSession->sendCommand(cmdstr.c_string());
	}
	else
	{
		sendStatRequests();
		if (activeRequests == 0 && maxRequests <= 0)
		{
			// All requests are finished and all required threads has been found
			sendTitleRequests();
		}
	}
	return 0;
}

bool KOMListThreadsCommand::progressLocToGlob(KOMString& re, bool is_ok)
{
	if (!is_ok)
	{
		error(re.parseErrorCode());
		return true;
	}
	int firstText, lastText, more, dense;
	locToGlobCmdId = -1;
	if (!re.parseInt(firstText) ||
			!re.parseInt(lastText) ||
			!re.parseInt(more) ||
			!re.parseInt(dense))
	{
		error(KOM_ERROR_PROTOCOL);
		return true;
	}
	if (dense)
	{
		int firstLocal, numTexts;
		if (!re.parseInt(firstLocal) || !re.parseArrayStart(numTexts))
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		for (int i = 0; i < numTexts; ++i)
		{
			int textNum;
			if (!re.parseInt(textNum))
			{
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			if (textNum > 0)
			{
				if (komSession->addEmptyTextStat(textNum, firstLocal, conf->getNum()))
				{
					error(KOM_ERROR_OOM);
					return true;
				}
			}
			++firstLocal;
		}
	}
	else
	{
		int numTexts;
		if (!re.parseArrayStart(numTexts))
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		for (int i = 0; i < numTexts; ++i)
		{
			int localNum, textNum;
			if (!re.parseInt(localNum) ||
					!re.parseInt(textNum))
			{
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			if (textNum > 0 && localNum > 0)
			{
				if (komSession->addEmptyTextStat(textNum, localNum, conf->getNum()))
				{
					error(KOM_ERROR_OOM);
					return true;
				}
			}
		}
	}
	
	if (komSession->getProtocolVersion() < 11)
	{
		conf->updateCacheRange(firstReqLocToGlob, firstReqLocToGlob <= 1);
	}
	else
	{
		conf->updateCacheRange(firstText, !more);
	}
	sendStatRequests();
	if (activeRequests == 0 && maxRequests <= 0)
	{
		// All requests are finished and all required threads has been found
		sendTitleRequests();
	}
	return true;
}

void KOMListThreadsCommand::sendStatRequests()
{
	// Start from the end, the newest messages
	while (activeRequests < maxRequests)
	{
		if (currentText < conf->getFirstCached())
		{
			// If there are more entries to find
			if (conf->isOldestCached())
				maxRequests = 0; // No more requests can be made
			else
			{
				// Send a new local to global reverse from firstText-1
				locToGlobCmdId = komSession->getCommandID();
				// Get the newest texts
				// FIXME: OOM
				KOMString cmdstr;
				if (komSession->getProtocolVersion() < 11)
				{
					firstReqLocToGlob = conf->getFirstCached()-254;
					if (firstReqLocToGlob < 1)
						firstReqLocToGlob = 1;
					// FIXME: OOM
					cmdstr.appendFormat("%d 103 %d %d 255\n", locToGlobCmdId, conf->getNum(), firstReqLocToGlob);
				}
				else
					// FIXME: OOM
					cmdstr.appendFormat("%d 121 %d %d 255\n", locToGlobCmdId, conf->getNum(), conf->getFirstCached());
				komSession->sendCommand(cmdstr.c_string());
			}
			return;
		}
		KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
		if (txt && !txt->isUnavailable())
		{
			requests[activeRequests].reqTextNum = txt->getGlobalNum();
			// New text to request stat for (if there is no stat for it already)
			requests[activeRequests].baseLocalNum = currentText;
			KOMTextStat* prevtxt = NULL;
			while (txt && txt->hasStat() && !txt->isUnavailable() && txt->getCommentTo()>=0)
			{
				prevtxt = txt;
				requests[activeRequests].reqTextNum = txt->getCommentTo();
				txt = komSession->getGlobalTextStat(txt->getCommentTo());
			}
			if (txt && txt->isUnavailable())
				txt = prevtxt;
			if (txt && txt->hasStat())
			{
				// Check if this was a new thread or not
				KOMThread* th = (KOMThread*)threads.get(txt->getGlobalNum());
				if (th)
				{
					if (currentText > th->newestLocalNum)
						th->newestLocalNum = currentText;
				}
				else
				{
					// Found a new thread
					--maxRequests;
					KOMThread* th = new KOMThread(txt->getGlobalNum());
					if (!th)
					{
						error(KOM_ERROR_OOM);
						return; // FIXME: must abort
					}
					threads.add(th);
					th->newestLocalNum = currentText;
				}
			}
			else
			{
				// if there is already a request for the text we are about to request, this cannot possibly be a new thread
				bool new_request = true;
				for (int r = 0; r < activeRequests && new_request; ++r)
				{
					if (requests[activeRequests].reqTextNum == requests[r].reqTextNum)
					{
						// If this is a newer text, hijack the old request
						if (requests[activeRequests].baseLocalNum > requests[r].baseLocalNum)
							requests[r].baseLocalNum = requests[activeRequests].baseLocalNum;
						new_request = false;
					}
				}
				if (new_request)
				{
					// send the request for text stat
					requests[activeRequests].cmdId = komSession->getCommandID();
					// FIXME: OOM
					KOMString cmdstr;
					cmdstr.appendFormat("%d 90 %d\n", requests[activeRequests].cmdId, requests[activeRequests].reqTextNum);
					komSession->sendCommand(cmdstr.c_string());
					++activeRequests;
				}
			}
		}
		--currentText;
	}
}

void KOMListThreadsCommand::sendTitleRequests()
{
	currentText = newestLocal;
	int foundThreads = 0;
	// Find the threads we don't really care about
	while (foundThreads < firstThread)
	{
		KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
		KOMTextStat* prevtxt = NULL;
		while (txt && !txt->isUnavailable() && txt->getCommentTo() >= 0)
		{
			prevtxt = txt;
			txt = komSession->getGlobalTextStat(txt->getCommentTo());
		}
		if (txt && txt->isUnavailable())
			txt = prevtxt;
		KOMThread* e;
		if (txt && (e = (KOMThread*)threads.get(txt->getGlobalNum())) && e->newestLocalNum == currentText)
		{
			++foundThreads;
		}
		--currentText;
		if (currentText < conf->getFirstCached())
		{
			ok(KOM_GROUP_SEPARATOR "0");
			return;
		}
	}
	foundThreads = 0;
	activeRequests = 0;
	while (foundThreads < numThreads)
	{
		KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
		KOMTextStat* prevtxt = NULL;
		while (txt && !txt->isUnavailable() && txt->getCommentTo() >= 0)
		{
			prevtxt = txt;
			txt = komSession->getGlobalTextStat(txt->getCommentTo());
		}
		if (txt && txt->isUnavailable())
			txt = prevtxt;
		KOMThread* e;
		if (txt && (e = (KOMThread*)threads.get(txt->getGlobalNum())) && e->newestLocalNum == currentText)
		{
			txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
			if (!komSession->getTextData(txt->getGlobalNum()))
			{
				requests[activeRequests].reqTextNum = txt->getGlobalNum();
				requests[activeRequests].baseLocalNum = -1;
	
				requests[activeRequests].cmdId = komSession->getCommandID();
				KOMString cmdstr;
				cmdstr.appendFormat("%d 25 %d 0 %d\n", requests[activeRequests].cmdId, txt->getGlobalNum(), txt->getTextLength());
				komSession->sendCommand(cmdstr.c_string());
				++activeRequests;
			}
			++foundThreads;
		}
		--currentText;
		if (currentText < conf->getFirstCached())
			break;
	}
	if (!activeRequests)
		sendReply();

}

bool KOMListThreadsCommand::progressStat(KOMString& re, int id, bool is_ok)
{
	int r;
	for (r = 0; r < activeRequests; ++r)
	{
		if (requests[r].cmdId == id)
			break;
	}
	if (r >= activeRequests)
		return false;
	
	if (requests[r].baseLocalNum < 0)
	{
		if (is_ok)
		{
			// title request
			if (!komSession->getTextData(requests[r].reqTextNum))
			{
				KOMTextData* text = new KOMTextData(requests[r].reqTextNum);
				if (!re.parseString(text->text))
				{
					error(KOM_ERROR_PROTOCOL);
					return true;
				}
				KOMTextStat* txt = komSession->getGlobalTextStat(requests[r].reqTextNum);
				if (txt && !txt->getContentType().parseLiteral("text/"))
				{
					KOMString temp;
					temp.append(text->text.c_string(), text->text.len());
					temp.advanceToAfter('\n');
					temp.base64Encode();
					text->text.terminateAt('\n');
					text->text.append('\n');
					text->text.append(temp.c_string());
				}
				if (!txt || !txt->isUTF8())
					text->text.latin1ToUTF8();
				if (txt && txt->shouldReformat())
					text->text.komToPlainText();
				komSession->addTextData(text);
			}
		}
		// FIXME: else, assumes the text is unavailable if this happens, maybe mark the text stat as deleted too? 
		if (r != activeRequests-1)
		{
			requests[r] = requests[activeRequests-1];
		}
		--activeRequests;
		if (activeRequests <= 0)
			sendReply();
		return true;
	}

	int status = komSession->addTextStat(requests[r].reqTextNum, is_ok, re);
	if (status > 0)
	{
		error(KOM_ERROR_PROTOCOL);
		return true;
	}
	else if (status < 0)
	{
		error(KOM_ERROR_OOM);
		return true;
	}

	KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(requests[r].baseLocalNum));

	if (!txt || txt->isUnavailable())
	{
		// The initial text of the thread cound not be found or was unavailable
		if (r != activeRequests-1)
		{
			requests[r] = requests[activeRequests-1];
		}
		--activeRequests;
		// Launch new stat requests if needed
		sendStatRequests();
	}
	else
	{
		// Find next text to request stat for
		KOMTextStat* prevtxt = NULL;
		while (txt && txt->hasStat() && !txt->isUnavailable() && txt->getCommentTo() >= 0)
		{
			prevtxt = txt;
			requests[r].reqTextNum = txt->getCommentTo();
			txt = komSession->getGlobalTextStat(txt->getCommentTo());
		}
		if (txt && txt->isUnavailable())
			txt = prevtxt;
		// If top level comment
		if (txt && txt->hasStat())
		{
			// Found a thread
			KOMThread* th = (KOMThread*)threads.get(txt->getGlobalNum());
			if (th)
			{
				if (requests[r].baseLocalNum > th->newestLocalNum)
					th->newestLocalNum = requests[r].baseLocalNum;
			}
			else
			{
				// Found a new thread
				--maxRequests;
				KOMThread* th = new KOMThread(txt->getGlobalNum());
				if (!th)
				{
					error(KOM_ERROR_OOM);
					return true; // FIXME: must abort
				}
				threads.add(th);
				th->newestLocalNum = requests[r].baseLocalNum;
			}
			if (r != activeRequests-1)
			{
				requests[r] = requests[activeRequests-1];
			}
			--activeRequests;
			// Launch new stat requests if needed
			sendStatRequests();
		}
		else
		{
			// if there is already a request for the text we are about to request, this cannot possibly be a new thread
			bool new_request = true;
			for (int tr = 0; tr < activeRequests && new_request; ++tr)
			{
				if (r != tr && requests[r].reqTextNum == requests[tr].reqTextNum)
				{
					// If this is a newer text, hijack the old request
					if (requests[r].baseLocalNum > requests[tr].baseLocalNum)
						requests[tr].baseLocalNum = requests[r].baseLocalNum;
					new_request = false;
				}
			}
			if (new_request)
			{
				KOMString cmdstr;
				// else request stat for the new text
				requests[r].cmdId = komSession->getCommandID();
				cmdstr.appendFormat("%d 90 %d\n", requests[r].cmdId, requests[r].reqTextNum);
				komSession->sendCommand(cmdstr.c_string());
			}
			else
			{
				// This request thread is no longer active
				if (r != activeRequests-1)
				{
					requests[r] = requests[activeRequests-1];
				}
				--activeRequests;
				// Launch new stat requests if needed
				sendStatRequests();
			}
		}
	}

	if (activeRequests == 0 && maxRequests <= 0)
	{
		// All requests are finished and all required threads has been found
		sendTitleRequests();
	}

	return true;
}

bool KOMListThreadsCommand::progress(KOMString& reply)
{
	int id;
	bool is_ok;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (progressStat(reply, id, is_ok))
		return true;
	if (id == locToGlobCmdId)
		return progressLocToGlob(reply, is_ok);
	return false;
}

void KOMListThreadsCommand::sendReply()
{
	KOMString reply;
	currentText = newestLocal;
	int foundThreads = 0;
	// Find the threads we don't really care about
	while (foundThreads < firstThread)
	{
		KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
		KOMTextStat* prevtxt = NULL;
		while (txt && !txt->isUnavailable() && txt->getCommentTo() >= 0)
		{
			prevtxt = txt;
			txt = komSession->getGlobalTextStat(txt->getCommentTo());
		}
		if (txt && txt->isUnavailable())
			txt = prevtxt;
		KOMThread* e;
		if (txt && (e = (KOMThread*)threads.get(txt->getGlobalNum())) && e->newestLocalNum == currentText)
		{
			++foundThreads;
		}
		--currentText;
		if (currentText < conf->getFirstCached())
		{
			ok(KOM_GROUP_SEPARATOR "0");
			return;
		}
	}
	foundThreads = 0;
	while (foundThreads < numThreads)
	{
		KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
		KOMTextStat* prevtxt = NULL;
		while (txt && !txt->isUnavailable() && txt->getCommentTo() >= 0)
		{
			prevtxt = txt;
			txt = komSession->getGlobalTextStat(txt->getCommentTo());
		}
		if (txt && txt->isUnavailable())
			txt = prevtxt;
		KOMThread* e;
		if (txt && (e = (KOMThread*)threads.get(txt->getGlobalNum())) && e->newestLocalNum == currentText)
		{
			int globalNum = txt->getGlobalNum();
			++foundThreads;
			KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
			KOMTextData* txtdata = komSession->getTextData(txt->getGlobalNum());
			KOMString title;
			if (txtdata)
			{
				title.append(txtdata->text.c_string());
				title.terminateAt('\n');
			}
			else
			{
				title.append("*Unavailable*");
			}
			// FIXME: OOM
			reply.appendFormat(KOM_GROUP_SEPARATOR "%d"
					KOM_RECORD_SEPARATOR "%d"
					KOM_RECORD_SEPARATOR "%u"
					KOM_RECORD_SEPARATOR "%s"
					KOM_RECORD_SEPARATOR "%d", globalNum, txt->getAuthor(), (unsigned int)txt->getTimeval(), title.c_string(), (conf->isTextRead(currentText)?1:0)); // last is read
		}
		--currentText;
		if (currentText < conf->getFirstCached())
		{
			// FIXME: OOM
			reply.append(KOM_GROUP_SEPARATOR "0");
			ok(reply.c_string());
			return;
		}
	}
	while (1)
	{
		KOMTextStat* txt = komSession->getGlobalTextStat(conf->getGlobalTextNum(currentText));
		KOMTextStat* prevtxt = NULL;
		while (txt && !txt->isUnavailable() && txt->getCommentTo() >= 0)
		{
			prevtxt = txt;
			txt = komSession->getGlobalTextStat(txt->getCommentTo());
		}
		if (txt && txt->isUnavailable())
			txt = prevtxt;
		KOMThread* e;
		if (txt && (e = (KOMThread*)threads.get(txt->getGlobalNum())) && e->newestLocalNum == currentText)
		{
			// FIXME: OOM
			reply.append(KOM_GROUP_SEPARATOR "1");
			ok(reply.c_string());
			return;
		}
		--currentText;
		if (currentText < conf->getFirstCached())
		{
			// FIXME: OOM
			reply.append(KOM_GROUP_SEPARATOR "0");
			ok(reply.c_string());
			return;
		}
	}
}

