/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komlogin.h"
#include "version.h"


int KOMLoginCommand::Construct(KOMString& cmd)
{
	KOMString server;
	if (cmd.parseLiteral("reconnect"))
	{
		createNew = 0;
		user.clear();
		user.append(komSession->currentUser.c_string());
		passwd.clear();
		passwd.append(komSession->currentPasswd.c_string());
		if (komSession->reconnect())
		{
			// The connect call will send an error
			return 0;
		}
		state = LOGIN_CONNECT;
		return 0;
	}
	if (!cmd.parseLiteral("login") || !cmd.parseCommandString(server) || !cmd.parseCommandString(user) || !cmd.parseCommandString(passwd) || !cmd.parseInt(createNew))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	if (!server.UTF8ToLatin1() || !user.UTF8ToLatin1() || !passwd.UTF8ToLatin1())
	{
		error(KOM_ERROR_INVALID_CHARACTERS);
		return 0;
	}
	if (komSession->connect(server.c_string(), user.c_string(), passwd.c_string()))
	{
		// The connect call will send an error
		return 0;
	}
	state = LOGIN_CONNECT;
	return 0;
}

bool KOMLoginCommand::progress(KOMString& reply)
{
	// Store the session on the stack since members cannot be used after calling error
	KOMSession* sess = komSession;
	if (state == LOGIN_CONNECT)
	{
		if (!reply.parseLiteral("LysKOM"))
		{
			error(KOM_ERROR_PROTOCOL);
			sess->disconnect(false);
			return true;
		}
		// for create user
		KOMString cmd;
		cmdID = komSession->getCommandID();
		if (createNew)
		{
			if (cmd.appendFormat("%d 89 %dH%s %dH%s 00000000 0 { }\n", cmdID, user.len(), user.c_string(), passwd.len(), passwd.c_string()))
			{
				error(KOM_ERROR_OOM);
				sess->disconnect(false);
				return true;
			}
			state = LOGIN_CREATE_USER;
		}
		else
		{
			if (cmd.appendFormat("%d 76 %dH%s 1 0\n", cmdID, user.len(), user.c_string()))
			{
				error(KOM_ERROR_OOM);
				sess->disconnect(false);
				return true;
			}
			state = LOGIN_LOOKUP_USER;
		}
		komSession->sendCommand(cmd.c_string());
		return true;
	}
	int id;
	bool is_ok;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
	{
		error(KOM_ERROR_PROTOCOL);
		sess->disconnect(false);
		return true;
	}
	if (id != cmdID)
		return false;
	if (!is_ok)
	{
		error(reply.parseErrorCode());
		sess->disconnect(false);
		return true;
	}

	switch (state)
	{
	case LOGIN_LOOKUP_USER:
	{
		int numEntries;
		if (!reply.parseArrayStart(numEntries))
		{
			error(KOM_ERROR_PROTOCOL);
			sess->disconnect(false);
			return true;
		}
		int firstUser = -1;
		while (numEntries)
		{
			KOMString name;
			int foo;
			if (!reply.parseString(name))
				break;
			if (!reply.parseInt(foo))
				break;
			if (!reply.parseInt(userID))
				break;
			if (firstUser < 0)
				firstUser = userID;
			if (name == user)
			{
				KOMString cmd;
				cmdID = komSession->getCommandID();
				if (cmd.appendFormat("%d 62 %d %dH%s 0\n", cmdID, userID, passwd.len(), passwd.c_string()))
				{
					error(KOM_ERROR_OOM);
					sess->disconnect(false);
					return true;
				}
				komSession->sendCommand(cmd.c_string());
				state = LOGIN_LOGIN;
				return true;
			}
			--numEntries;
		}
		if (firstUser >= 0)
		{
			userID = firstUser;
			KOMString cmd;
			cmdID = komSession->getCommandID();
			if (cmd.appendFormat("%d 62 %d %dH%s 0\n", cmdID, userID, passwd.len(), passwd.c_string()))
			{
				error(KOM_ERROR_OOM);
				sess->disconnect(false);
				return true;
			}
			komSession->sendCommand(cmd.c_string());
			state = LOGIN_LOGIN;
			return true;
		}
		error(KOM_ERROR_INVALID_USER);
		sess->disconnect(false);
		break;
	}
	case LOGIN_CREATE_USER:
		if (reply.parseInt(userID))
		{
			KOMString cmd;
			cmdID = komSession->getCommandID();
			if (cmd.appendFormat("%d 62 %d %dH%s 0\n", cmdID, userID, passwd.len(), passwd.c_string()))
			{
				error(KOM_ERROR_OOM);
				sess->disconnect(false);
				return true;
			}
			komSession->sendCommand(cmd.c_string());
			state = LOGIN_LOGIN;
			return true;
		}
		error(KOM_ERROR_PROTOCOL);
		sess->disconnect(false);
		return true;
	case LOGIN_LOGIN:
	{
		komSession->setLoggedIn(userID);
		KOMString cmd;
		cmdID = komSession->getCommandID();
		if (cmd.appendFormat("%d 80 8 { 5 8 12 14 15 16 17 18 }\n", cmdID))
		{
			error(KOM_ERROR_OOM);
			sess->disconnect(false);
			return true;
		}
		komSession->sendCommand(cmd.c_string());
		state = LOGIN_ACCEPT_ASYNC;
		break;
	}
	case LOGIN_ACCEPT_ASYNC:
	{
		KOMString cmd;
		cmdID = komSession->getCommandID();
		if (cmd.appendFormat("%d 75\n", cmdID))
		{
			error(KOM_ERROR_OOM);
			sess->disconnect(false);
			return true;
		}
		komSession->sendCommand(cmd.c_string());
		state = LOGIN_CHECK_VERSION;
		break;
	}
	case LOGIN_CHECK_VERSION:
	{
		int protocolVersion;
		if (!reply.parseInt(protocolVersion))
		{
			error(KOM_ERROR_PROTOCOL);
			sess->disconnect(false);
		}
		else
		{
			if (!komSession->setProtocolVersion(protocolVersion))
			{
				error(KOM_ERROR_PROTOCOL);
				sess->disconnect(false);
			}
			else
			{
				KOMString cmd;
				cmdID = komSession->getCommandID();
				KOMString ver;
				ver.append(KOMCLEAN_VERSION_STRING);
				if (cmd.appendFormat("%d 69 8HKOMclean %dH%s\n", cmdID, ver.len(), ver.c_string()))
				{
					error(KOM_ERROR_OOM);
					sess->disconnect(false);
					return true;
				}
				komSession->sendCommand(cmd.c_string());
				state = LOGIN_SET_CLIENT;
			}
		}
		break;
	}
	case LOGIN_SET_CLIENT:
		{
			KOMString cmd;
			cmdID = komSession->getCommandID();
			if (cmd.appendFormat("%d 120 1\n", cmdID))
			{
				error(KOM_ERROR_OOM);
				sess->disconnect(false);
				return true;
			}
			komSession->sendCommand(cmd.c_string());
			state = LOGIN_SET_UTC;
		}
		break;
	case LOGIN_SET_UTC:
		{
			KOMString cmd;
			cmdID = komSession->getCommandID();
			if (cmd.appendFormat("%d 23\n", cmdID))
			{
				error(KOM_ERROR_OOM);
				sess->disconnect(false);
				return true;
			}
			komSession->sendCommand(cmd.c_string());
			state = LOGIN_INIT_FAVORITE_CACHE;
		}
		break;
	case LOGIN_INIT_FAVORITE_CACHE:
		{
			int numFav;
			komSession->clearFavorites();
			if (reply.parseArrayStart(numFav))
			{
				for (int i = 0; i < numFav; ++i)
				{
					int textNum, markType;
					if (!reply.parseInt(textNum) || !reply.parseInt(markType))
						break;
					komSession->addFavorite(textNum);
				}
			}
			ok("");
		}
		break;
	default:
		error(KOM_ERROR_PROTOCOL); // should never happen, so internal error is better
		sess->disconnect(false);
		break;
	}

	return true;
}

