/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2009 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komlookupname.h"

const char* const KOMLookupNameFromIDCommand::commandString = "lookup_name_from_id";
const char* const KOMListCachedNamesCommand::commandString = "list_cached_names";
const char* const KOMPrefetchNamesCommand::commandString = "prefetch_names";

KOMLookupNameFromIDCommand::~KOMLookupNameFromIDCommand()
{}

int KOMLookupNameFromIDCommand::Construct(KOMString& cmd)
{
	KOMString cmdstr;
	if (!cmd.parseLiteral(commandString) || !cmd.parseInt(conf))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	const KOMString* name = komSession->getConferenceName(conf);
	if (name)
	{
		if (cmdstr.appendFormat(KOM_GROUP_SEPARATOR "%d" 
				       KOM_RECORD_SEPARATOR "%s",
				       conf, name->c_string()))
		{
			error(KOM_ERROR_OOM);
			return true;
		}
		ok(cmdstr.c_string());
		return 0;
	}
	cmdID = komSession->getCommandID();
	if (cmdstr.appendFormat("%d 78 %d\n", cmdID, conf))
	{
		error(KOM_ERROR_OOM);
		return 0;
	}

	komSession->sendCommand(cmdstr.c_string());
	return 0;
}


bool KOMLookupNameFromIDCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (id != cmdID)
		return false;
	if (!is_ok)
		error(reply.parseErrorCode());
	else
	{
		KOMString restr;
		KOMString name;
		if (!reply.parseString(name))
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		name.latin1ToUTF8();
		komSession->updateConferenceName(conf, name);
		if (restr.appendFormat(KOM_GROUP_SEPARATOR "%d" 
				       KOM_RECORD_SEPARATOR "%s",
				       conf, name.c_string()))
		{
			error(KOM_ERROR_OOM);
			return true;
		}
		ok(restr.c_string());
	}
	return true;
}

int KOMListCachedNamesCommand::Construct(KOMString& cmd)
{
	KOMString cmdstr;
	if (!cmd.parseLiteral(commandString))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	KOMString reply;

	for (KOMConfNameCacheEntry* cn = komSession->getFirstCachedName(); cn; cn = cn->getNextInList())
	{
		reply.appendFormat(KOM_GROUP_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%s", cn->num, cn->name.c_string());
	}

	ok(reply.c_string());
	return 0;
}

KOMPrefetchNamesCommand::~KOMPrefetchNamesCommand()
{
	while (requests)
	{
		PFReq* r = requests->next;
		delete requests;
		requests = r;
	}
}

int KOMPrefetchNamesCommand::Construct(KOMString& cmd)
{
	if (!cmd.parseLiteral(commandString))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	int conf;
	requests = NULL;
	activeRequests = 0;
	while (cmd.parseInt(conf))
	{
		PFReq* r = new PFReq();
		if (!r)
		{
			error(KOM_ERROR_OOM);
			return 0;
		}
		const KOMString* name = komSession->getConferenceName(conf);
		r->reqId = 0;
		r->confId = conf;
		if (!name)
		{
			KOMString cmdstr;
			int cmdID = komSession->getCommandID();
			if (cmdstr.appendFormat("%d 78 %d\n", cmdID, conf))
			{
				delete r;
				error(KOM_ERROR_OOM);
				return 0;
			}
			r->reqId = cmdID;
			++activeRequests;
			komSession->sendCommand(cmdstr.c_string());
		}
		r->next = requests;
		requests = r;
	}

	return 0;
}


bool KOMPrefetchNamesCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	for (PFReq* r = requests; r; r = r->next)
	{
		if (r->reqId == id)
		{
			if (!is_ok)
			{
				error(reply.parseErrorCode());
				return true;
			}
			else
			{
				KOMString name;
				if (!reply.parseString(name))
				{
					error(KOM_ERROR_PROTOCOL);
					return true;
				}
				name.latin1ToUTF8();
				komSession->updateConferenceName(r->confId, name);
			}
			--activeRequests;
			if (!activeRequests)
			{
				// reply
				KOMString restr;

				for (r = requests; r; r = r->next)
				{
					const KOMString* name = komSession->getConferenceName(r->confId);
					if (name)
						restr.appendFormat(KOM_GROUP_SEPARATOR "%d" KOM_RECORD_SEPARATOR "%s", r->confId, name->c_string());
				}

				ok(restr.c_string());
			}
			return true;
		}
	}
	return false;
}
