/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#include <stdlib.h>
#include "komsendmsg.h"
#include "version.h"

const char* const KOMSendMsgCommand::commandString = "send_message";
const char* const KOMGetReplyRecptCommand::commandString = "get_reply_recpt";

int KOMSendMsgCommand::Construct(KOMString& cmd)
{
	int to[128];
	int cc[128];
	int numTo = 0;
	int numCc = 0;
	KOMString cmdstr;
	bool toSelf = false;
	firstSubConf = 0;

	if (!cmd.parseLiteral(commandString))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}

	replyTo = -1;
	if (cmd.parseLiteral("reply"))
	{
		if (!cmd.parseInt(replyTo))
		{
			error(KOM_ERROR_SYNTAX);
			return 0;
		}
	}
	if (cmd.parseLiteral("to"))
	{
		while (cmd.parseInt(to[numTo]))
		{
			if (komSession->getConference(to[numTo]))
			{
				toSelf = true;
				if (firstSubConf <= 0)
					firstSubConf = to[numTo];
			}
			++numTo;
			if (numTo >= (int)(sizeof(to)/sizeof(to[0])))
			{
				error(KOM_ERROR_SYNTAX);
				return 0;
			}
		}
	}
	if (!numTo)
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	if (cmd.parseLiteral("cc"))
	{
		while (cmd.parseInt(cc[numCc]))
		{
			if (komSession->getConference(cc[numCc]))
			{
				toSelf = true;
				if (firstSubConf <= 0)
					firstSubConf = to[numTo];
			}
			++numCc;
			if (numCc >= (int)(sizeof(cc)/sizeof(cc[0])))
			{
				error(KOM_ERROR_SYNTAX);
				return 0;
			}
		}
	}
	if (!cmd.parseLiteral("text"))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	if (!cmd.parseCommandString(text))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	
	cmdID = komSession->getCommandID();
	int err;
	KOMString auxArray;
	KOMString version;
	version.appendFormat("KOMclean %s", KOMCLEAN_VERSION_STRING);
	auxArray.appendFormat("2 { 15 00000 1 %dH%s 1 00000 1 ", version.len(), version.c_string());
	if (!text.UTF8ToLatin1())
	{
		auxArray.appendFormat("%dH%s }", KOM_CONTENT_TYPE_UTF8_LEN, KOM_CONTENT_TYPE_UTF8);
	}
	else
	{
		auxArray.appendFormat("%dH%s }", KOM_CONTENT_TYPE_LEN, KOM_CONTENT_TYPE);
	}
	err = cmdstr.appendFormat("%d 86 %dH%s %d { ", cmdID, text.len(), text.c_string(), numTo+numCc+(toSelf?0:1)+(replyTo>0?1:0));
	if (replyTo > 0 && !err)
	{
		err = cmdstr.appendFormat("2 %d ", replyTo);
	}
	for (int i = 0; i < numTo && !err; ++i)
	{
		err = cmdstr.appendFormat("0 %d ", to[i]);
	}
	if (!err && !toSelf)
	{
		firstSubConf = komSession->getUserID();
		err = cmdstr.appendFormat("0 %d ", komSession->getUserID());
	}
	for (int i = 0; i < numCc && !err; ++i)
	{
		err = cmdstr.appendFormat("1 %d ", cc[i]);
	}
	if (!err)
		err = cmdstr.appendFormat("} %s\n", auxArray.c_string());
	// Clear text so the next reply is handled as a ok
	text.clear();
	if (err)
	{
		error(KOM_ERROR_OOM);
		return 0;
	}
	komSession->sendCommand(cmdstr.c_string());
	return 0;
}

bool KOMSendMsgCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	if (id != cmdID)
		return false;
	if (!is_ok)
		error(reply.parseErrorCode());
	else
	{
		KOMString response;
		int textNum;
		if (!reply.parseInt(textNum))
			textNum = 0;
		response.appendFormat(KOM_GROUP_SEPARATOR "%d" KOM_GROUP_SEPARATOR "%d", textNum, firstSubConf);
		ok(response.c_string());
	}
	return true;
}

int KOMGetReplyRecptCommand::Construct(KOMString& cmd)
{
	if (!cmd.parseLiteral(commandString) || !cmd.parseLiteral("from") || !cmd.parseInt(from) || !cmd.parseLiteral("to"))
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	int allocTo = 128;
	to = new Recpt[allocTo];
	numTo = 0;
	toModified = false;
	toSelf = false;
	fromCmd = 0;
	while (cmd.parseInt(to[numTo].conf))
	{
		to[numTo].cmdId = 0;
		++numTo;
		if (numTo >= allocTo)
		{
			Recpt* newTo = new Recpt[allocTo+128];
			if (!newTo)
			{
				error(KOM_ERROR_OOM);
				return 0;
			}
			for (int i = 0; i < numTo; ++i)
				newTo[i] = to[i];
			delete[] to;
			to = newTo;
			allocTo += 128;
		}
	}

	if (!numTo)
	{
		error(KOM_ERROR_SYNTAX);
		return 0;
	}
	pendingCmds = 0;
	for (int i = 0; i < numTo; ++i)
	{
		int reconf = komSession->getReplyConference(to[i].conf);
		if (reconf > 0)
		{
			toModified |= (reconf != to[i].conf);
			to[i].conf = reconf;
		}
		else
		{
			KOMString req;
			to[i].cmdId = komSession->getCommandID();
			req.appendFormat("%d 91 %d\n", to[i].cmdId, to[i].conf);
			komSession->sendCommand(req.c_string());
			++pendingCmds;
		}
	}

	if (!pendingCmds)
		toListFinished();

	return 0;
}

bool KOMGetReplyRecptCommand::progress(KOMString& reply)
{
	bool is_ok;
	int id;
	if (!reply.parseStatus(is_ok) || !reply.parseInt(id))
		return false;
	for (int i = 0; i < numTo; ++i)
	{
		if (to[i].cmdId == id)
		{
			if (!is_ok)
			{
				error(reply.parseErrorCode());
				return true;
			}

			// Parse the conference info
			KOMString tmpstr;
			int tmpint;
			time_t tmptime;
			unsigned int type;
			int superconf;
			bool res = reply.parseString(tmpstr);
			res &= reply.parseBitfield(type);
			res &= reply.parseTime(tmptime);
			res &= reply.parseTime(tmptime);
			res &= reply.parseInt(tmpint);
			res &= reply.parseInt(tmpint);
			res &= reply.parseInt(tmpint);
			res &= reply.parseInt(tmpint);
			res &= reply.parseInt(superconf);

			if (!res)
			{
				error(KOM_ERROR_SYNTAX);
				return true;
			}
			int reconf = to[i].conf;
			if (((type>>6) & 1) && superconf > 0)
			{
				toModified |= (to[i].conf != superconf);
				reconf = superconf;
			}
			komSession->setReplyConference(to[i].conf, reconf);
			to[i].conf = reconf;

			to[i].cmdId = 0;
			--pendingCmds;
			if (pendingCmds == 0)
			{
				toListFinished();
			}
			return true;
		}
	}
	if (id == fromCmd)
	{
		if (!is_ok)
		{
			error(reply.parseErrorCode());
			return true;
		}

		bool toFrom = false;
		int numConfs;
		if (!reply.parseArrayStart(numConfs))
		{
			error(KOM_ERROR_PROTOCOL);
			return true;
		}
		for (int i = 0; i < numConfs && !toFrom; ++i)
		{
			int confNum;
			int temp;
			// position
			bool res = reply.parseInt(temp);
			// time
			time_t timeval;
			res &= reply.parseTime(timeval);
			// id
			res &= reply.parseInt(confNum);
			// prio
			res &= reply.parseInt(temp);

			if (!res)
			{
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			// read range
			if (komSession->getProtocolVersion() < 11)
			{
				// last read text
				res &= reply.parseInt(temp);
				res &= reply.parseSizeOnlyArray(temp);
			}
			else
			{
				res &= reply.parseSizeOnlyArray(temp);
			}
			// added by
			res &= reply.parseInt(temp);
			// time
			res &= reply.parseTime(timeval);
			// type (bitfield)
			unsigned int type;
			res &= reply.parseBitfield(type);
			if (!res)
			{
				error(KOM_ERROR_PROTOCOL);
				return true;
			}
			for (int c = 0; c < numTo; ++c)
			{
				if (to[c].conf == confNum)
					toFrom = true;
			}
		}


		KOMString rep;
		rep.appendFormat(KOM_GROUP_SEPARATOR "%d", to[0].conf);
		for (int c = 1; c < numTo; ++c)
			rep.appendFormat(KOM_RECORD_SEPARATOR "%d", to[c].conf);
		if (!toSelf)
		{
			rep.appendFormat(KOM_RECORD_SEPARATOR "%d", komSession->getUserID());
		}
		// FIXME: this should also check the aux item "send-comments-to"
		if (!toFrom)
		{
			rep.appendFormat(KOM_RECORD_SEPARATOR "%d", from);
		}
		ok(rep.c_string());
		return true;
	}
	return false;
}

void KOMGetReplyRecptCommand::toListFinished()
{
	// Remove duplicates and check for self
	for (int c = 0; c < numTo; ++c)
	{
		for (int c2 = c+1; c2 < numTo; ++c2)
		{
			if (to[c].conf == to[c2].conf)
			{
				for (int c3 = c2+1; c3 < numTo; ++c3)
					to[c3-1] = to[c3];
				--numTo;
			}
		}
		if (komSession->getConference(to[c].conf))
			toSelf = true;
	}
	if (toModified)
	{
		// check from
		fromCmd = komSession->getCommandID();
		KOMString req;
		if (komSession->getProtocolVersion()<11)
			req.appendFormat("%d 99 %d 0 1000 0\n", fromCmd, from);
		else
			req.appendFormat("%d 108 %d 0 1000 0 0\n", fromCmd, from);
		komSession->sendCommand(req.c_string());
		return;
	}
	KOMString rep;
	rep.appendFormat(KOM_GROUP_SEPARATOR "%d", to[0].conf);
	for (int c = 1; c < numTo; ++c)
		rep.appendFormat(KOM_RECORD_SEPARATOR "%d", to[c].conf);
	if (!toSelf)
	{
		rep.appendFormat(KOM_RECORD_SEPARATOR "%d", komSession->getUserID());
	}
	ok(rep.c_string());
}