/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#define _CRT_SECURE_NO_WARNINGS

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <stdio.h>
#include <stddef.h>
#include <ctype.h>

#include "komstring.h"

#define KOM_STRING_GROW_SIZE 128

KOMString::KOMString() : c_str(NULL), length(0), allocated_length(0),
	parser_position(0), stored_parser_position(0)
{}

KOMString::~KOMString()
{
	delete[] c_str;
}

int KOMString::findChar(char c, int startPos)
{
	while (startPos < length && c_str[startPos] != c)
		++startPos;
	return startPos;
}

static const char base64itoc[] = {
'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K',
'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g',
'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r',
's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2',
'3', '4', '5', '6', '7', '8', '9', '+', '/' 
};

bool KOMString::base64Encode()
{
	KOMString dest;
	unsigned int tmp;
	int state = 0;
	unsigned char* data = (unsigned char*)c_str;
	for (int pos = 0; pos < length; ++pos)
	{
		switch (state)
		{
		case 0:
			if (dest.append(base64itoc[data[pos]>>2]))
				return false;
			++state;
			// Intentional fallthough
		case 1:
			tmp = (data[pos]<<4)&0x30;
			if (pos+1 < length)
				tmp |= (data[pos+1]>>4);
			if (dest.append(base64itoc[tmp]))
				return false;
			++state;
			break;

		case 2:
			tmp = (data[pos]<<2)&0x3c;
			if (pos+1 < length)
				tmp |= (data[pos+1]>>6);
			if (dest.append(base64itoc[tmp]))
				return false;
			++state;
			break;
		case 3:
			if (dest.append(base64itoc[data[pos]&0x3f]))
				return false;
			state = 0;
			break;
		default:
			state = 0;
			break;
		}
	}
	if (state != 0)
	{
		// Append some padding to make sure the last character is parsed
		switch (state)
		{
		case 2:
			if (dest.append('='))
				return false;
			// Intentional fall-through
		case 3:
			if (dest.append('='))
				return false;
			break;
		default:
			break;
		}
	}
	clear();
	return append(dest.c_str)==0;
}

static const unsigned char numBytesTag[5] = { 0x00, 0x00, 0xC0, 0xE0, 0xF0 };
bool KOMString::latin1ToUTF8()
{
	if (!length)
		return true;
	bool need_convert = false;
	unsigned char* src = (unsigned char*)c_str;
	while (*src && !need_convert)
	{
		if (*src & 0x80)
			need_convert = true;
		++src;
	}
	if (!need_convert)
		return true;
	char* old_str = c_str;
	src = (unsigned char*)old_str;
	c_str = NULL;
	length = 0;
	allocated_length = 0;
	while (*src)
	{
		if (*src & 0x80)
		{
			append(numBytesTag[2] | ((*src) >> 6));
			append(0x80 | ((*src)&0x3f));
		}
		else
			append(*src);
		++src;
	}
	delete[] old_str;
	return true;
}

bool KOMString::UTF8ToLatin1()
{
	if (!length)
		return true;
	unsigned char* src = (unsigned char*)c_str;
	while (*src)
	{
		int bytes = 1;
		while (bytes < 4 && *src >= numBytesTag[bytes+1])
			++bytes;
		unsigned int ch = (*(src++))&(~numBytesTag[bytes]);
		for (int i = 1; i < bytes; ++i)
		{
			ch <<= 6;
			// Check for broken string
			if ((*src & 0xC0) == 0x80)
				ch |= ((*(src++))&0x3f);
		}
		if (ch > 0xff)
			return false;
	}
	src = (unsigned char*)c_str;
	unsigned char* dst = (unsigned char*)c_str;
	while (*src)
	{
		int bytes = 1;
		while (bytes < 4 && *src >= numBytesTag[bytes+1])
			++bytes;
		unsigned int ch = (*(src++))&(~numBytesTag[bytes]);
		for (int i = 1; i < bytes; ++i)
		{
			ch <<= 6;
			// Check for broken string
			if ((*src & 0xC0) == 0x80)
				ch |= ((*(src++))&0x3f);
		}
		*(dst++) = ch;
	}
	*dst = '\0';
	length = (ptrdiff_t)((char*)dst-c_str);
	return true;
}

bool KOMString::komToPlainText()
{
	char* text_start = c_str;
	while (*text_start && *text_start != '\n')
		++text_start;
	if (!*text_start)
		return true;
	++text_start;
	char* paragraph_start = text_start;
	while (*paragraph_start)
	{
		int num_lines = 0;
		int num_early_wrapped_lines = 0;
		// Skip all newlines at the start of a paragraph
		while (*paragraph_start && *paragraph_start == '\n')
			++paragraph_start;
		char* pos = paragraph_start;
		// Analyze paragraph by paragraph, if the paragraph could fit one more word within 70 chars on >= half the lines  it should not be reformated
		bool possible_wrap = true;
		while (*pos && (pos[0] != '\n' || pos[1] != '\n'))
		{
			++num_lines;
			char* line_start = pos;
			if (*line_start == '>' || *line_start == '*' || *line_start == '|' || 
				*line_start == '+' || *line_start == ' ')
			{
				possible_wrap = false;
			}
			int double_spaces = 0;
			while (*pos && *pos != '\n')
			{
				if (pos[0] == ' ' && pos[1] == ' ')
					++double_spaces;
				if (pos[0] == '\t')
					possible_wrap = false;
				if (pos[0] == '-' && pos[1] == '-' && pos[2] == '-')
					possible_wrap = false;
				++pos;
			}
			if (double_spaces > 1)
					possible_wrap = false;
			// Found the line length
			if (*pos && pos[1] != '\n')
			{
				++pos;
				char* word_pos = pos;
				while (*word_pos && *word_pos != ' ')
					++word_pos;
				if ((ptrdiff_t)(word_pos-line_start) <= 70)
					++num_early_wrapped_lines;
			}
		}
		if (possible_wrap && num_early_wrapped_lines*2 < num_lines)
		{
			pos = paragraph_start;
			char* prevpos = NULL;
			while (*pos && (pos[0] != '\n' || pos[1] != '\n'))
			{
				if (*pos == '\n')
					*pos = ' ';
				// Strip all dual spaces to make sure the monospace check in js does not trigger
				if (prevpos && *prevpos == ' ' && *pos == ' ')
				{
					char* srcpos = pos+1;
					char* dstpos = pos;
					while (*srcpos)
					{
						*dstpos = *srcpos;
						++srcpos;
						++dstpos;
					}
					*dstpos = '\0';
					--length;
				}
				else
				{
					prevpos = pos;
					++pos;
				}
			}
		}
		paragraph_start = pos;
	}
	return true;
}

void KOMString::stripConfName()
{
	for (int pos = 0; pos < length; ++pos)
	{
		if (c_str[pos] == '(')
		{
			int endpos = pos+1;
			int count = 1;
			while (endpos < length && count)
			{
				if (c_str[endpos] == ')')
					--count;
				else if (c_str[endpos] == '(')
					++count;
				++endpos;
			}
			if (!count)
			{
				int startpos = pos;
				int diff = endpos-pos;
				while (endpos <= length)
				{
					c_str[startpos] = c_str[endpos];
					++startpos;
					++endpos;
				}
				length -= diff;
				// Check same position again
				--pos;
			}
		}
	}
	int dstpos = 0;
	int pos = 0;
	while (c_str[pos] == ' ')
		++pos;
	while (pos < length)
	{
		while (c_str[pos] == ' ' && c_str[pos+1] == ' ')
			++pos;
		c_str[dstpos] = tolower(c_str[pos]);
		++dstpos;
		++pos;
	}
	while (dstpos > 0 && c_str[dstpos-1] == ' ')
		--dstpos;
	c_str[dstpos] = '\0';
	length = dstpos;
}

int KOMString::compare(const KOMString& other)
{
	return strcmp(c_str, other.c_str);
}


void KOMString::clear()
{
	if (length > 0)
	{
		*c_str = '\0';
		length = 0;
	}
}

void KOMString::terminateAt(char c)
{
	if (!length)
		return;
	int pos = findChar(c, 0);
	if (pos < length)
	{
		c_str[pos] = '\0';
		length = pos;
	}
}

void KOMString::advanceToAfter(char c)
{
	if (!length)
		return;
	int pos = findChar(c, 0);
	if (pos >= length)
	{
		clear();
		return;
	}
	++pos;
	int ofs;
	for (ofs = 0; pos+ofs < length; ++ofs)
		c_str[ofs] = c_str[pos+ofs];
	c_str[ofs] = '\0';
	length = ofs;
}

int KOMString::grow(int minlen)
{
	if (minlen + 1 > allocated_length)
	{
		int newlen = allocated_length;
		while (minlen + 1 > newlen)
			newlen += KOM_STRING_GROW_SIZE;
		char* nstr = new char[newlen];
		if (!nstr)
			return -1;
		if (c_str)
			memcpy(nstr, c_str, length+1);
		allocated_length = newlen;
		delete[] c_str;
		c_str = nstr;
	}
	return 0;
}

int KOMString::append(char c)
{
	if (grow(length+1))
		return -1;
	c_str[length] = c;
	++length;
	c_str[length] = '\0';
	return 0;
}

int KOMString::append(const char* str)
{
	int slen = strlen(str);
	if (grow(length+slen))
		return -1;
	strcpy(c_str+length, str);
	length += slen;
	return 0;
}

int KOMString::append(const char* str, unsigned int slen)
{
	if (grow(length+slen))
		return -1;
	memcpy(c_str+length, str, slen);
	length += slen;
	c_str[length] = '\0';
	return 0;
}

int KOMString::appendFormat(const char* fmt, ...)
{
	int slen = 0;
	va_list ap;
	int written = -1;
	while (written == -1 || written >= slen)
	{
		slen += 1024;
		if (c_str)
			c_str[length] = '\0';
		if (grow(length+slen))
		{
			if (c_str)
				c_str[length] = '\0';
			return -1;
		}
		va_start(ap, fmt);
		written = vsnprintf(c_str+length, slen, fmt, ap);
		va_end(ap);
	}
	length += written;
	return 0;
}

bool KOMString::operator==(const char* str)
{
	if (!length)
		return *str=='\0';
	return !strcmp(c_str, str);
}

bool KOMString::operator!=(const char* str)
{
	if (!length)
		return *str!='\0';
	return strcmp(c_str, str)!=0;
}

bool KOMString::hasCompleteReply()
{
	if (parser_position >= length)
		return false;
	// Find first newline and first string
	int newline = findChar('\n', parser_position);
	int string = findChar('H', parser_position);
	// If there is no newline this cannot be a command
	if (newline >= length)
		return false;
	// While there is a string before the newline
	while (string < newline)
	{
		// Parse the length of the string
		char* strpos = c_str+string-1;
		int slen = 0;
		int scale = 1;
		while (strpos > c_str+parser_position && (*strpos >= '0' && *strpos <= '9'))
		{
			slen += scale*(*strpos-'0');
			--strpos;
			scale *= 10;
		}
		++string;
		// Skip the string length if any
		if (scale > 1)
		{
			// If the string is incomplete the command is incomplete
			if (length-string < slen)
				return false;
			string += slen;
		}
		// if the new string position is after the newline, find a new newline
		if (string > newline)
		{
			newline = findChar('\n', string);
			if (newline>=length)
				return false;
		}
		// Find the next string
		string = findChar('H', string);
	}
	// No string was found before the newline, it must be a valid reply
	return true;
}

void KOMString::resetParserPosition()
{
	parser_position = 0;
	stored_parser_position = 0;
}

void KOMString::saveParserPosition()
{
	stored_parser_position = parser_position;
}

void KOMString::restoreParserPosition()
{
	parser_position = stored_parser_position;
}

void KOMString::deleteParsed()
{
	if (parser_position == 0)
		return;
	for (int ofs = 0; parser_position+ofs <= length; ++ofs)
		c_str[ofs] = c_str[ofs+parser_position];
	length -= parser_position;
	resetParserPosition();
}

bool KOMString::parseStatus(bool& is_ok)
{
	if (parser_position >= length)
		return false;
	skipWhitespace();
	if (c_str[parser_position] != '=' && c_str[parser_position] != '%')
		return false;
	is_ok = (c_str[parser_position] == '=');
	++parser_position;
	return true;
}

bool KOMString::parseAsync()
{
	if (parser_position >= length)
		return false;
	skipWhitespace();
	if (c_str[parser_position] != ':')
		return false;
	++parser_position;
	return true;
}

bool KOMString::parseProtocolError()
{
	if (parser_position >= length)
		return false;
	skipWhitespace();
	if (c_str[parser_position] != '%' || c_str[parser_position+1] != '%')
		return false;
	parser_position += 2;
	return true;
}

static const char* const KOMErrorCodes[] = 
{
	"No error",
	"Unknown error",
	"Internal error (not implemented)",
	"Internal error (obsolete call)",
	"Invalid password",
	"String too long",
	"Not logged in",
	"Login not allowed",
	"Internal error (conference zero)",
	"Internal error (undefined conference)",
	"Internal error (undefined person)",
	"Access denied",
	"Permission denied",
	"Not a member",
	"No such text",
	"Internal error (text zero)",
	"No such text",
	"Internal error (local text zero)",
	"Bad name",
	"Internal error (index out of range)",
	"Conference already exists",
	"Person already exists",
	"Secret requires a closed meeting",
	"Internal error (letterbox)",
	"Internal error (ldb error)",
	"Internal error (illegal misc)",
	"Internal error (illegal info type)",
	"Recipied added multiple times",
	"Text is already a comment",
	"Text is already a footnote",
	"Not recipient",
	"Not comment",
	"Not footnote",
	"Too many recipients",
	"Too many comments",
	"Too many footnotes",
	"Too many marks",
	"Not your own text",
	"No connection",
	"Out of memory",
	"Server is crazy",
	"Client is crazy",
	"Internal error (undefined session)",
	"Regexp error",
	"Not marked",
	"Temporary failure, try again later",
	"Too long array",
	"Anonymous texts not alowed",
	"Internal error (illegal aux item)",
	"Internal error (aux item permission denied)",
	"Internal error (unknown async)",
	"Internal sever error",
	"Feature disabled",
	"Message not sent",
	"Invalid membersip settings",
	"Internal error (invalid range)",
	"Internal error (invalid range list)",
	"Internal error (undefined measurement)",
	"Permission denied",
	"Permission denied",
	"Invalid weight",
	"Internal error (bad bool)"
};
const char* KOMString::parseErrorCode()
{
	int err;
	if (!parseInt(err) || err < 0 || (unsigned int)err >= sizeof(KOMErrorCodes)/sizeof(KOMErrorCodes[0]))
		return "Unknown error";
	return KOMErrorCodes[err];
}

bool KOMString::parseSizeOnlyArray(int& entries)
{
	if (parser_position >= length)
		return false;
	int oldpos = parser_position;
	if (!parseInt(entries))
		return false;
	skipWhitespace();
	if (c_str[parser_position] != '*')
	{
		parser_position = oldpos;
		return false;
	}
	++parser_position;
	return true;
}
bool KOMString::parseArrayStart(int& entries)
{
	if (parser_position >= length)
		return false;
	int oldpos = parser_position;
	if (!parseInt(entries))
		return false;
	skipWhitespace();
	if (c_str[parser_position] != (entries?'{':'*'))
	{
		parser_position = oldpos;
		return false;
	}
	++parser_position;
	return true;
}

bool KOMString::parseArrayEnd(int entries)
{
	if (parser_position >= length)
		return false;
	skipWhitespace();
	int oldpos = parser_position;
	if (entries && c_str[parser_position] != '}')
	{
		parser_position = oldpos;
		return false;
	}
	if (entries)
		++parser_position;
	return true;
}

bool KOMString::parseString(KOMString& string)
{
	if (parser_position >= length)
		return false;
	int oldpos = parser_position;
	int len;
	string.clear();
	if (!parseInt(len))
		return false;
	if (c_str[parser_position] != 'H')
	{
		parser_position = oldpos;
		return false;
	}
	++parser_position;
	while (len)
	{
		// FIXME: might not be best way to handle oom
		if (parser_position >= length || string.append(c_str[parser_position]))
		{
			parser_position = oldpos;
			return false;
		}
		++parser_position;
		--len;
	}
	return true;
}

bool KOMString::parseCommandString(KOMString& string)
{
	if (parser_position >= length)
		return false;
	int oldpos = parser_position;
	string.clear();
	skipWhitespace();
	if (c_str[parser_position] != *KOM_RECORD_SEPARATOR)
	{
		parser_position = oldpos;
		return false;
	}
	++parser_position;
	while (c_str[parser_position] != *KOM_RECORD_SEPARATOR)
	{
		// FIXME: might not be best way to handle oom
		if (parser_position >= length || string.append(c_str[parser_position]))
		{
			parser_position = oldpos;
			return false;
		}
		++parser_position;
	}
	++parser_position;
	return true;
}

bool KOMString::parseBitfield(unsigned int& bits)
{
	if (parser_position >= length)
		return false;
	int oldpos = parser_position;
	skipWhitespace();
	unsigned int bitData = 0;
	int numBits = 8;
	while (numBits > 0)
	{
		--numBits;
		if (c_str[parser_position] == '1')
			bitData |= (1<<numBits);
		else if (c_str[parser_position] != '0')
		{
			break;
		}
		++parser_position;
	}
	// Failuse on the 5:th bit is ok since some bitfields only use 4 numbers
	if (numBits && numBits != 3)
	{
		parser_position = oldpos;
		return false;
	}
	bits = bitData;
	return true;
}

bool KOMString::parseInt(int& num)
{
	if (parser_position >= length)
		return false;
	skipWhitespace();
	bool valid = false;
	bool neg = false;
	if (c_str[parser_position] == '-')
	{
		neg = true;
		++parser_position;
	}
	num = 0;
	while (parser_position < length && c_str[parser_position] >= '0' && c_str[parser_position] <= '9')
	{
		num *= 10;
		num += c_str[parser_position]-'0';
		++parser_position;
		valid = true;
	}
	if (neg)
		num = -num;
	return valid;
}

#ifdef _MSC_VER
time_t timegm(struct tm *tm)
{
	time_t ret;
	char *tz;

	tz = getenv("TZ");
	_putenv("TZ=UTC");
	_tzset();

	ret = mktime(tm);
	if(tz)
	{
		char buf[256];
		_snprintf(buf, sizeof(buf), "TZ=%s", tz);
		_putenv(buf);
	}
	else
		_putenv("TZ=");
	_tzset();

	return ret;
}
#endif // _MSC_VER
bool KOMString::parseTime(time_t& timeval)
{
	if (parser_position >= length)
		return false;
	struct tm t;
	int oldpos = parser_position;
	bool res = parseInt(t.tm_sec);
	res &= parseInt(t.tm_min);
	res &= parseInt(t.tm_hour);
	res &= parseInt(t.tm_mday);
	res &= parseInt(t.tm_mon);
	res &= parseInt(t.tm_year);
	res &= parseInt(t.tm_wday);
	res &= parseInt(t.tm_yday);
	res &= parseInt(t.tm_isdst);
	if (!res)
		parser_position = oldpos;
	else
		timeval = timegm(&t);
	return res;
}

bool KOMString::parseLiteral(const char* lit)
{
	if (parser_position >= length)
		return false;
	skipWhitespace();
	int llen = strlen(lit);
	if (!strncmp(c_str+parser_position, lit, llen))
	{
		parser_position += llen;
		return true;
	}
	return false;
}

