/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#ifndef KOMSTRING_H
#define KOMSTRING_H

#include <time.h>

#define KOM_GROUP_SEPARATOR "\x01d"
#define KOM_RECORD_SEPARATOR "\x01e"

/** String class used by kom for all parsing and formating of strings. */
class KOMString
{
public:
	KOMString();
	~KOMString();


	/** Get the c string representation of this string. */
	const char* c_string() const {return c_str?c_str:"";}
	/** Get the length of this string. */
	int len() const {return length;}

	/** Base64 encode the content of a string. This assumes the string data is binary (unsigned char). 
	 * @returns true on success, false on error. */
	bool base64Encode();
	/** Convert the content of the string from latin1 to utf-8.
	 * @returns true on success, false otherwise. */
	bool latin1ToUTF8();
	/** Convert the content of the string from utf-8 to latin1 if possible.
	 * @returns true on success, false if not possible or oom. */
	bool UTF8ToLatin1();

	/** Convert the string from the reformatable text/x-kom-basic format to
	 * text/plain.
	 * @returns true on success, false on error. */
	bool komToPlainText();

	/** Strip a name of parenthesis and leading/trailing/multiple whitespace. 
	 * Also turn it into lowercase to make it possible to sort. */
	void stripConfName();

	/** Do the equivalent of a strcmp(this, other). Only works with non binary strings. */
	int compare(const KOMString& other);

	/** Clear this string to an empty string. */
	void clear();
	/** Append a character to the string. */
	int append(char c);
	/** Append a c string to the end of this string. */
	int append(const char* str);
	/** Append a c string of a specified length to the end of this string. */
	int append(const char* str, unsigned int slen);
	/** Append a formated c string to the end of this string. */
	int appendFormat(const char* fmt, ...);

	/** Stop the string at the first occurance of a character. */
	void terminateAt(char c);
	/** Advance the string to start after the given char. If the char is
	 * not found the entre string is deleted. */
	void advanceToAfter(char c);

	/** Compare two strings. Only work fro non binary strings, so be careful. */
	bool operator==(const KOMString& str){return *this == str.c_string();}
	/** Compare two strings. Only work fro non binary strings, so be careful. */
	bool operator==(const char* str);
	/** Compare two strings. Only work fro non binary strings, so be careful. */
	bool operator!=(const char* str);

	// Parsing kom input
	/** Check if the current string from the parser position and forward 
	 * contains at least one complete kom reply. */
	bool hasCompleteReply();
	/** Reset the parser position to 0. Also resets the stored position. */
	void resetParserPosition();
	/** Save the current parser position in the string. Only one position 
 	 * can be stored. */
	void saveParserPosition();
	/** Restore the parset position stroed y saveParserPosition. */
	void restoreParserPosition();
	/** Delete all parsed data from the string. */
	void deleteParsed();
	/** Parse if a kom reply is an async message. */
	bool parseAsync();
	/** Parse if the reply is a protocol error. */
	bool parseProtocolError();
	/** Parse status of a KOM reply. */
	bool parseStatus(bool& is_ok);
	/** Parse an error status and translate it to a readable string. */
	const char* parseErrorCode();
	/** Parse a size only array. This means the array is of the form <size> * */
	bool parseSizeOnlyArray(int& entries);
	/** Parse the start of an array. */
	bool parseArrayStart(int& entries);
	/** Parse the end of an array after all entries in the array have been parsed. */
	bool parseArrayEnd(int entries);
	/** Parse a string in holerith format. The input string might be modified even if parsing fails*/
	bool parseString(KOMString& string);
	/** Parse a string surrounded by KOM_RECORD_SEPARATOR characters. The input string might be modified even if parsing fails*/
	bool parseCommandString(KOMString& string);
	/** Parse a bitfield. */
	bool parseBitfield(unsigned int& bits);
	/** Parse an integer. */
	bool parseInt(int& num);
	/** Parse a time value. Will be returned as a time_t with epoc matching the system. */
	bool parseTime(time_t& timeval);

	/* Parsing mainly used for comamnd parsing, not kom parsing. */
	/** Parse a literal string. The literal must not contain leading whitespace. */
	bool parseLiteral(const char* lit);
private:
	// Copy constructor is private since it would allocate memory
	KOMString(const KOMString& old);
	void operator=(const KOMString& str){}

	/** @ returns the first position character c is found on after (or on) startPos. 
	 * If the character is not found a value >= length is returned. */
	int findChar(char c, int startPos);

	/** Skip leading whitespace. Only used for parsing. */
	void skipWhitespace(){while (parser_position < length && c_str[parser_position] == ' ') ++parser_position;}
	/** Grow to fit at least minlen characters in the string. */
	int grow(int minlen);

	char* c_str;
	int length;
	int allocated_length;
	int parser_position;
	int stored_parser_position;
};

#endif // KOMSTRING_H

