#include "win32trayicon.h"

#include <shellapi.h>

#define WM_TRAYICON WM_USER+5

Win32TrayIcon::Win32TrayIcon(HWND w, TrayIconListener* l) : win(w), numBitmaps(0), 
	currentBitmap(0), isVisible(false), listener(l), popupMenu(NULL)
{}

Win32TrayIcon::~Win32TrayIcon()
{
	if (isVisible)
		show(0);
	// Release the bitmaps
	for (int i = 0; i < numBitmaps; ++i)
	{
		DestroyIcon(bitmaps[i]);
	}
	clearPopupMenu();
}

int Win32TrayIcon::addIcon(void* icon)
{
	if (numBitmaps >= MAX_ICONS)
		return -1;
	bitmaps[numBitmaps] = (HICON) LoadImage(GetModuleHandle(NULL),
									(LPCWSTR)icon,
									IMAGE_ICON,
									GetSystemMetrics(SM_CXSMICON),
									GetSystemMetrics(SM_CYSMICON),
									LR_DEFAULTCOLOR);
	if (!bitmaps[numBitmaps])
		return -1;
	++numBitmaps;
	return 0;
}

int Win32TrayIcon::setActiveIcon(int iconId)
{
	if (iconId >= numBitmaps || iconId < 0)
		return -1;
	if (iconId == currentBitmap)
		return 0;
	if (isVisible)
	{
		NOTIFYICONDATA icndata;
		memset(&icndata, 0, sizeof(icndata));
		icndata.cbSize = sizeof(icndata);
		icndata.hWnd = win;
		icndata.uID = (UINT)this;
		icndata.uFlags = NIF_ICON;
		icndata.hIcon = bitmaps[iconId];

		Shell_NotifyIcon(NIM_MODIFY, &icndata);
	}
	currentBitmap = iconId;
	return 0;
}

int Win32TrayIcon::show(int vis)
{
	if (!vis)
	{
		if (!isVisible)
			return 0;
		NOTIFYICONDATA icndata;
		memset(&icndata, 0, sizeof(icndata));
		icndata.cbSize = sizeof(icndata);
		icndata.hWnd = win;
		icndata.uID = (UINT)this;
		icndata.uFlags = 0;

		Shell_NotifyIcon(NIM_DELETE, &icndata);
		isVisible = false;
		return 0;
	}
	if (currentBitmap < 0 || currentBitmap >= numBitmaps)
		return -1;
	if (isVisible)
		return 0;

	NOTIFYICONDATA icndata;
	memset(&icndata, 0, sizeof(icndata));
	icndata.cbSize = sizeof(icndata);
	icndata.hWnd = win;
	icndata.uID = (UINT)this;
	icndata.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
	icndata.hIcon = bitmaps[currentBitmap];

	icndata.uCallbackMessage = WM_TRAYICON;

	icndata.szTip[0] = '\0';

	Shell_NotifyIcon(NIM_ADD, &icndata);
	isVisible = true;
	return 0;
}

void Win32TrayIcon::clearPopupMenu()
{
	while (popupMenu)
	{
		TrayPopupMenuItem* nm = popupMenu->next;
		delete[] popupMenu->text;
		delete popupMenu;
		popupMenu = nm;
	}
}

int Win32TrayIcon::addPopupMenuItem(const char* text, int id, bool def)
{
	TrayPopupMenuItem* item = new TrayPopupMenuItem;
	if (!item)
		return -1;
	item->next = NULL;
	item->id = id;
	item->text = new wchar_t[strlen(text)+1];
	if (!item->text)
	{
		delete item;
		return -1;
	}
	for (int i = 0; text[i]; ++i)
	{
		item->text[i] = text[i];
	}
	item->text[strlen(text)] = 0;
	item->def = def;

	TrayPopupMenuItem** nextp = &popupMenu;

	while (*nextp)
		nextp = &((*nextp)->next);
	*nextp = item;

	return 0;
}

int Win32TrayIcon::handleEvent(UINT msg, WPARAM wParam, LPARAM lParam)
{
	if (msg == WM_COMMAND)
	{
		listener->onAction(wParam);
		return 1;
	}
	else if (msg != WM_TRAYICON)
		return 0;
	if (lParam == WM_LBUTTONDBLCLK)
	{
		listener->onDefaultAction();
	}
	else if (lParam == WM_RBUTTONDOWN || lParam == WM_CONTEXTMENU)
	{
		HMENU hMenu;
		POINT pt;
		GetCursorPos(&pt);
		if (popupMenu && (hMenu = CreatePopupMenu()))
		{
			int def = -1;
			int num = 0;
			for (TrayPopupMenuItem* item = popupMenu; item; item = item->next)
			{
				if (def < 0 && item->def)
					def = num;
				InsertMenu(hMenu, -1, MF_BYPOSITION, item->id, item->text);
				++num;
			}
			//InsertMenu(hMenu, -1, MF_BYPOSITION|MF_SEPARATOR, WM_TRAYICON, L"--");
			if (def >= 0)
				SetMenuDefaultItem(hMenu, def, true);

			// must set window to the foreground or the menu won't disappear when it should
			SetForegroundWindow(win);

			TrackPopupMenu(hMenu, TPM_BOTTOMALIGN, pt.x, pt.y, 0, win, NULL);
			DestroyMenu(hMenu);
		}
	}
	return 1;
}

