#include "x11trayicon.h"

#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/time.h>

#include <X11/xpm.h>

#define SYSTEM_TRAY_REQUEST_DOCK	0
#define SYSTEM_TRAY_BEGIN_MESSAGE	1
#define SYSTEM_TRAY_CANCEL_MESSAGE	2

#define POPUP_BORDER_WIDTH 2

static const double trayicon_double_click_time = 0.5;

double get_time_ms()
{
	struct timeval t;
	gettimeofday(&t, NULL);
	return (t.tv_sec+((double)t.tv_usec)/1000000);
}

static int trapped_error_code = 0;
static int (*old_error_handler)(Display*, XErrorEvent*);

static int error_handler(Display *display, XErrorEvent *error)
{
	trapped_error_code = error->error_code;
	return 0;
}

void trap_errors()
{
	trapped_error_code = 0;
	old_error_handler = XSetErrorHandler(error_handler);
}

int untrap_errors()
{
	XSetErrorHandler(old_error_handler);
	return trapped_error_code;
}

X11TrayIcon::X11TrayIcon(Display* d, TrayIconListener* l) : disp(d), win(NULL), 
				gc(NULL), width(0), height(0), 
				lastClick(0.), numBitmaps(0), currentBitmap(0),
				isVisible(false), listener(l), popup_win(NULL), popup_gc(NULL),
				popupMenu(NULL)
{}

X11TrayIcon::~X11TrayIcon()
{
	closePopupWindow();
	clearPopupMenu();
	trap_errors();
	for (int i = 0; i < numBitmaps; ++i)
	{
		XFreePixmap(disp, bitmaps[i]);
		XFreePixmap(disp, masks[i]);
	}
	if (gc)
		XFreeGC(disp, gc);
	if (win)
		XDestroyWindow(disp, win);
	untrap_errors();
}

int X11TrayIcon::createWindow()
{
	if (!disp)
		return -1;
	width = 32;
	height = 32;
	// Create a window
	trap_errors();
	int blackColor = BlackPixel(disp, DefaultScreen(disp));
	win = XCreateSimpleWindow(disp, DefaultRootWindow(disp), 0, 0, 
				width, height, 0, blackColor, blackColor);
	if (win)
	{
		XSelectInput(disp, win, StructureNotifyMask | ExposureMask | ButtonPressMask);
		gc = XCreateGC(disp, win, 0, NULL);
	}
	if (untrap_errors() || !win || !gc)
		return -1;

	// Set the size hints
	XSizeHints sizehint;
	memset(&sizehint, 0, sizeof(sizehint));
	sizehint.flags = PMinSize | PMaxSize;
	sizehint.min_width = 16;
	sizehint.min_height = 16;
	sizehint.max_width = 48;
	sizehint.max_height = 48;
	trap_errors();
	XSetWMNormalHints(disp, win, &sizehint);
	if (untrap_errors())
		return -1;

#if 0
	// Set the class hint
	XClassHint classhint;
	classhint.res_name = "foo";
	classhint.res_class = "foo";
	XSetClassHint(disp, win, &classhint);

	// Set the name property
	char* name = "foo";
	XChangeProperty(disp, win, XInternAtom(disp, "_NET_WM_NAME", False),
			XInternAtom(disp, "UTF8_STRING", False),
			8, PropModeReplace, (unsigned char *)name,
			strlen(name));

	// Set the icon hint
	unsigned long icondata[16*16+2];
	int iconsize = 16*16+2;
	memset(icondata, 255, iconsize*4);
	icondata[0] = 16;
	icondata[1] = 16;
	XChangeProperty(disp, win, XInternAtom(disp, "_NET_WM_ICON", False),
			XA_CARDINAL, 32, PropModeReplace, (unsigned char*)icondata, iconsize);
#endif // 0
	return 0;
}

int X11TrayIcon::requestDock()
{
	// Set the embed info
	long data[2] = {0, 1}; // Version, docked
	trap_errors();
	Atom embed_info = XInternAtom(disp, "_XEMBED_INFO", False);
	XChangeProperty(disp, win, embed_info, embed_info, 32, PropModeReplace, 
			(unsigned char*)data, 2);
	if (untrap_errors())
		return -1;
	// Request dock
	int err = sendTrayMessage(SYSTEM_TRAY_REQUEST_DOCK, win);
	if (err)
		return err;
	return 0;
}

void X11TrayIcon::repaintIcon()
{
	XSetWindowBackgroundPixmap(disp, win, ParentRelative);
	XClearWindow(disp, win);
	if (currentBitmap < numBitmaps)
	{
		Window root_win;
		int x, y;
		unsigned int pmWidth, pmHeight;
		unsigned int border, depth;
		XGetGeometry(disp, bitmaps[currentBitmap], &root_win, 
			&x, &y, &pmWidth, &pmHeight, &border, &depth);
		XShapeCombineMask(disp, win, ShapeBounding, (width-pmWidth+1)/2, (height-pmHeight+1)/2, masks[currentBitmap], ShapeSet);
		XCopyArea(disp, bitmaps[currentBitmap], win, gc, 0, 0, pmWidth, pmHeight, (width-pmWidth+1)/2, (height-pmHeight+1)/2);
	}
	XSync(disp, False);
}

int X11TrayIcon::handleEvent(const XEvent &e)
{
	if (popup_win)
	{
		if (e.xany.window != popup_win)
			return 0;
		if (e.type == ButtonPress)
		{
			closePopupWindow();
			if (e.xbutton.button == 1 && e.xbutton.x >= POPUP_BORDER_WIDTH && e.xbutton.x < popup_width-2*POPUP_BORDER_WIDTH 
				&& e.xbutton.y >= POPUP_BORDER_WIDTH && e.xbutton.y < popup_height-2*POPUP_BORDER_WIDTH)
			{
				int itemNum = (e.xbutton.y-POPUP_BORDER_WIDTH)/(popup_font_ascent+popup_font_descent);
				TrayPopupMenuItem* item = popupMenu;
				while (item && itemNum>0)
				{
					item = item->next;
					--itemNum;
				}
				if (item && listener)
				{
					listener->onAction(item->id);
				}
			}
		}
		else if (e.type == Expose)
		{
			int whiteColor = WhitePixel(disp, DefaultScreen(disp));
			int blackColor = BlackPixel(disp, DefaultScreen(disp));
			XSetForeground(disp, popup_gc, whiteColor);
			XFillRectangle(disp, popup_win, popup_gc, 0, 0, popup_width, popup_height);
			XSetForeground(disp, popup_gc, blackColor);
			int xp = POPUP_BORDER_WIDTH;
			int yp = POPUP_BORDER_WIDTH;
			for (TrayPopupMenuItem* item = popupMenu; item; item = item->next)
			{
				XDrawString(disp, popup_win, popup_gc, xp, yp+popup_font_ascent, item->text, strlen(item->text));
				yp += popup_font_ascent+popup_font_descent;
			}
		}
		return 1;
	}
	if (e.xany.window != win)
		return 0;
	if (e.type == ClientMessage)
	{
		if (e.xclient.message_type != XInternAtom(disp, "_XEMBED", False))
		{
			//printf("Unknown client message\n");
		}
	}
	else if (e.type == ConfigureNotify)
	{
		width = e.xconfigure.width;
		height = e.xconfigure.height;
	}
	else if (e.type == Expose)
	{
		repaintIcon();
	}
	else if (e.type == ButtonPress)
	{
		if (e.xbutton.button == 3)
		{
			createPopupWindow(e.xbutton.x_root, e.xbutton.y_root);
		}
		else if (e.xbutton.button == 1)
		{
			// Left button
			double curClick = get_time_ms();
			if (curClick-lastClick < trayicon_double_click_time)
			{
				if (listener)
					listener->onDefaultAction();
				curClick = 0.;
			}
			lastClick = curClick;
		}
	}
	//else
	//	printf("Other event %d\n", e.type);
	return 1;
}

int X11TrayIcon::show(int vis)
{
	if (!vis && isVisible)
	{
		trap_errors();
		if (gc)
			XFreeGC(disp, gc);
		if (win)
			XDestroyWindow(disp, win);
		untrap_errors();
		win = NULL;
		gc = NULL;
		isVisible = false;
	}
	else if (vis && !isVisible)
	{
		int err = createWindow();
		if (err)
			return err;
		isVisible = true;
		err = requestDock();
		if (err)
		{
			show(0);
			return err;
		}
	}
	return 0;
}

int X11TrayIcon::addIcon(void* icon)
{
	if (numBitmaps >= MAX_ICONS)
		return -1;
	trap_errors();
	XpmCreatePixmapFromData(disp, DefaultRootWindow(disp), (char**)icon, &bitmaps[numBitmaps], &masks[numBitmaps], NULL);
	if (untrap_errors() || bitmaps[numBitmaps] == None || masks[numBitmaps] == None)
	{
		printf("Failed to add icon\n");
		return -1;
	}
	++numBitmaps;
	return 0;
}

int X11TrayIcon::setActiveIcon(int iconId)
{
	if (iconId < 0 || iconId >= numBitmaps)
		return -1;
	bool changed = (currentBitmap != iconId);
	currentBitmap = iconId;
	if (win && gc && changed)
		repaintIcon();
	return 0;
}

void X11TrayIcon::clearPopupMenu()
{
	// It's not good to modify the menu while it's visible, so close it
	closePopupWindow();
	while (popupMenu)
	{
		TrayPopupMenuItem* nm = popupMenu->next;
		delete[] popupMenu->text;
		delete popupMenu;
		popupMenu = nm;
	}
}

int X11TrayIcon::addPopupMenuItem(const char* text, int id, bool def)
{
	// It's not good to modify the menu while it's visible, so close it
	closePopupWindow();
	TrayPopupMenuItem* item = new TrayPopupMenuItem;
	if (!item)
		return -1;
	item->next = NULL;
	item->id = id;
	item->text = new char[strlen(text)+1];
	if (!item->text)
	{
		delete item;
		return -1;
	}
	for (int i = 0; text[i]; ++i)
	{
		item->text[i] = text[i];
	}
	item->text[strlen(text)] = 0;
	item->def = def;

	TrayPopupMenuItem** nextp = &popupMenu;

	while (*nextp)
		nextp = &((*nextp)->next);
	*nextp = item;

	return 0;
}

int X11TrayIcon::sendTrayMessage(long message, long data1, long data2, long data3)
{
	// Get the system tray, which should be owner of this selection
	trap_errors();
	Window stw = XGetSelectionOwner(disp, XInternAtom(disp, "_NET_SYSTEM_TRAY_S0", False));
	if (untrap_errors() || stw == None)
		return 1;

	// Send an event to the system tray
	XEvent ev;
	memset(&ev, 0, sizeof(ev));
	ev.xclient.type = ClientMessage;
	ev.xclient.window = stw;
	ev.xclient.message_type = XInternAtom(disp, "_NET_SYSTEM_TRAY_OPCODE", False);
	ev.xclient.format = 32;
	ev.xclient.data.l[0] = CurrentTime;
	ev.xclient.data.l[1] = message;
	ev.xclient.data.l[2] = data1;
	ev.xclient.data.l[3] = data2;
	ev.xclient.data.l[4] = data3;

	trap_errors();
	XSendEvent(disp, stw, False, NoEventMask, &ev);
	XSync(disp, False);
	if (untrap_errors())
	{
		return 1;
	}
	return 0;
}

void X11TrayIcon::createPopupWindow(int x, int y)
{
	if (!disp || !popupMenu || popup_win)
		return;
	// Create a window
	trap_errors();
	int blackColor = BlackPixel(disp, DefaultScreen(disp));
	popup_win = XCreateSimpleWindow(disp, DefaultRootWindow(disp), 0, 0, 
				32, 32, 0, blackColor, blackColor);
	if (popup_win)
	{
		XSelectInput(disp, popup_win, StructureNotifyMask | ExposureMask | ButtonPressMask);
		popup_gc = XCreateGC(disp, popup_win, 0, NULL);
	}
	if (untrap_errors() || !popup_win || !popup_gc)
	{
		closePopupWindow();
		return;
	}

	XSetWindowAttributes override;
	override.override_redirect = True;
	XChangeWindowAttributes(disp, popup_win, CWOverrideRedirect, &override);

	// calculate the size based on the content of the popup menu
	popup_width = 0;
	popup_height = POPUP_BORDER_WIDTH*2;
	for (TrayPopupMenuItem* item = popupMenu; item; item = item->next)
	{
		int direction;
		XCharStruct sizes;
		XQueryTextExtents(disp, XGContextFromGC(popup_gc), item->text, strlen(item->text), &direction, &popup_font_ascent, &popup_font_descent, &sizes);
		if (sizes.width > popup_width)
			popup_width = sizes.width;
		popup_height += popup_font_ascent+popup_font_descent;
	}
	popup_width += POPUP_BORDER_WIDTH*2;
	if (x > popup_width)
		x -= popup_width;
	if (y > popup_height)
		y -= popup_height;
	XMoveResizeWindow(disp, popup_win, x, y, popup_width, popup_height);

	XMapWindow(disp, popup_win);

	XGrabPointer(disp, popup_win, False, ButtonPressMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

}

void X11TrayIcon::closePopupWindow()
{
	if (popup_gc)
		XFreeGC(disp, popup_gc);
	if (popup_win)
	{
		XUngrabPointer(disp, CurrentTime);
		XUnmapWindow(disp, popup_win);
		XDestroyWindow(disp, popup_win);
	}
	popup_gc = NULL;
	popup_win = NULL;
}

