/* -*- Mode: c++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
**
** Copyright 2008 Tim Johansson.  All rights reserved.
**
*/

#define _CRT_SECURE_NO_WARNINGS

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <stddef.h>

#ifndef MACOSX
#define HTTP_IF_MODIFIED_SUPPORT
#endif // !MACOSX

#ifdef WIN32
#include <winsock.h>
#define close(s) closesocket(s)
#define ioctl(s, t, v) ioctlsocket(s, t, v)
#define UINT32 unsigned long
#include "../trayicon/win32trayicon.h"
#include "../../resource.h"
#define WM_SOCKET WM_USER
HWND g_win = NULL;
#else // !WIN32
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#define UINT32 u_int32_t
#ifdef MACOSX
#include "../trayicon/mactrayicon.h"
#else // !MACOSX
#include "../trayicon/x11trayicon.h"
#define static static const
#include "../../icon/icon_normal.xpm"
#include "../../icon/icon_unread.xpm"
#include "../../icon/icon_unread_msg.xpm"
#include "../../icon/icon_notconnected.xpm"
#undef static
#endif // !MACOSX
#endif // !WIN32
#include <time.h>

#include "../komclient.h"
#include "../komui.h"

#include "../trayicon/trayicon.h"
#include "versionchecker.h"

#include "../version.h"

#define DEFAULT_HTTP_PORT 8080
#define KOM_KEEPALIVE_TIMEOUT 60
#define KOM_RECONNECT_TIMEOUT 10

enum HTTPError
{
	HTTP_ERR_NOT_FOUND = 0,
	HTTP_ERR_BAD_REQUEST,
	HTTP_ERR_INTERNAL,
	HTTP_ERR_FORBIDDEN,
	HTTP_ERR_NOT_IMPLEMENTED,
	HTTP_ERR_NOT_MODIFIED
};
const int HTTPErrorNum[] = {
	404,
	400,
	500,
	403,
	501,
	304
};
const char* const HTTPErrorString[] = {
	"Not Found",
	"Bad Request",
	"Internal Error",
	"Forbidden",
	"Not Implemented",
	"Not Modified"
};

/** HTTP request class. Keeps track of HTTP requests from the client. */
class HTTPRequest
{
public:
	enum RequestType {REQ_GET, REQ_POST, REQ_HEAD};

	HTTPRequest(int socket) : url(NULL), content(NULL), 
#ifdef HTTP_IF_MODIFIED_SUPPORT
		modified_check(NULL), 
#endif // HTTP_IF_MODIFIED_SUPPORT
		socket(socket)
	{}
	~HTTPRequest()
	{
		if (socket >= 0)
			close(socket);
		delete[] url;
		delete[] content;
#ifdef HTTP_IF_MODIFIED_SUPPORT
		delete[] modified_check;
#endif // HTTP_IF_MODIFIED_SUPPORT
	}
	/** Read the request data and create an HTTP request from it. */
	int Construct();

	/** Reply ok to this request. */
	void replyOK(const char* msg, unsigned int len = 0, const char* type = "text/html", time_t last_modified = 0);
	/** Reply to this request with an error code. */
	void replyError(HTTPError err, const char* msg, const char* type = "text/html");

	RequestType getRequestType(){return requestType;}
	const char* getURL(){return url;}
	unsigned int getContentLength(){return contentLength;}
	char* getContent(){return content;}
#ifdef HTTP_IF_MODIFIED_SUPPORT
	char* getIfModifiedString(){return modified_check;}
#endif // HTTP_IF_MODIFIED_SUPPORT

	/** Replace \r\n with \n. */
	void stripLineFeed();

	int requestNum;
	HTTPRequest* next;

private:
	char* url;
	int http_major;
	int http_minor;
	RequestType requestType;
	int contentLength;
	char* content;
#ifdef HTTP_IF_MODIFIED_SUPPORT
	char* modified_check;
#endif // HTTP_IF_MODIFIED_SUPPORT
	int socket;
};

void HTTPRequest::replyOK(const char* msg, unsigned int len, const char* type, time_t last_modified)
{
	if (!len)
		len = strlen(msg);
	if (socket < 0)
		return;
	send(socket, "HTTP/1.0 200 OK\r\n", 17, 0);
	send(socket, "Content-Type: ", 14, 0);
	send(socket, type, strlen(type), 0);
	send(socket, "; charset=utf-8\r\n", 22, 0);
	char clen[1024];
	time_t now = time(NULL);
	sprintf(clen, "Date: %s", asctime(gmtime(&now)));
	clen[strlen(clen)-1] = '\0';
	strcat(clen, " GMT\r\n");
	send(socket, clen, strlen(clen), 0);
	if (last_modified)
	{
		sprintf(clen, "Last-Modified: %s", asctime(gmtime(&last_modified)));
		clen[strlen(clen)-1] = '\0';
		strcat(clen, " GMT\r\n");
		send(socket, clen, strlen(clen), 0);
#ifdef HTTP_IF_MODIFIED_SUPPORT
		sprintf(clen, "Expires: %s", asctime(gmtime(&now)));
		clen[strlen(clen)-1] = '\0';
		strcat(clen, " GMT\r\n");
		send(socket, clen, strlen(clen), 0);
#endif // HTTP_IF_MODIFIED_SUPPORT
	}
	sprintf(clen, "Content-Length: %d\r\n\r\n", len);
	send(socket, clen, strlen(clen), 0);
	if (requestType != REQ_HEAD)
		send(socket, msg, len, 0);
}

void HTTPRequest::replyError(HTTPError err, const char* msg, const char* type)
{
	if (socket < 0)
		return;
	char clen[1024];
	sprintf(clen, "HTTP/1.0 %d %s\r\n", HTTPErrorNum[err], HTTPErrorString[err]);
	send(socket, clen, strlen(clen), 0);
	if (err != HTTP_ERR_NOT_MODIFIED)
	{
		send(socket, "Content-Type: ", 14, 0);
		send(socket, type, strlen(type), 0);
		send(socket, "; charset=utf-8\r\n", 22, 0);
	}
	time_t now = time(NULL);
	sprintf(clen, "Date: %s", asctime(gmtime(&now)));
	clen[strlen(clen)-1] = '\0';
	strcat(clen, " GMT\r\n");
	send(socket, clen, strlen(clen), 0);
	if (err != HTTP_ERR_NOT_MODIFIED)
	{
		sprintf(clen, "Content-Length: %d\r\n\r\n", (int)strlen(msg));
		send(socket, clen, strlen(clen), 0);
		send(socket, msg, strlen(msg), 0);
	}
}

void HTTPRequest::stripLineFeed()
{
	unsigned char* src = (unsigned char*)content;
	unsigned char* dst = (unsigned char*)content;
	while (*src)
	{
		if (*src != '\r')
			*(dst++) = *src;
		++src;
	}
	*dst = '\0';
	contentLength = strlen(content);
}

int HTTPRequest::Construct()
{
	static int lastRequest = 0;
	requestNum = ++lastRequest;
	static char request[8*1024];
	int reqlen = recv(socket, request, sizeof(request)-1, 0);
	if (reqlen <= 0)
	{
		close(socket);
		socket = -1;
		return -1;
	}
	request[reqlen] = 0;
	const char* reqpos = strchr(request, '\n');
	if (!reqpos)
		return -1;
	delete[] url;
	url = new char[reqpos - request];
	if (!url)
		return -2;
	contentLength = 0;
	delete[] content;
	content = NULL;
	if (sscanf(request, "GET %[^ \r\n] HTTP/%d.%d", url, &http_major, &http_minor) == 3)
	{
		requestType = REQ_GET;
	}
	else if (sscanf(request, "POST %[^ \r\n] HTTP/%d.%d", url, &http_major, &http_minor) == 3)
	{
		// find the content-length and end of header so we can get the post data
		requestType = REQ_POST;
	}
	else if (sscanf(request, "HEAD %[^ \r\n] HTTP/%d.%d", url, &http_major, &http_minor) == 3)
	{
		requestType = REQ_HEAD;
	}
	else
	{
		return -1;
	}
	//printf("URL: \"%s\"\n", url);
	// parse the rest of the header stuff
	++reqpos;
	while (reqpos)
	{
		// break if the header is completed
		if (reqpos[0] == '\n')
		{
			++reqpos;
			break;
		}
		if (reqpos[0] == '\r' && reqpos[1] == '\n')
		{
			reqpos += 2;
			break;
		}
		const char* nextreq = strchr(reqpos, '\n');
		if (!nextreq)
		{
			// try to add more data to the request buffer
			reqlen = strlen(reqpos);
			// Make sure there is space for more data
			if ((unsigned int)reqlen >= sizeof(request)-1)
				return -1;
			memmove(request, reqpos, reqlen);
			// Make sure we could actually read some data
			int nreqlen = recv(socket, request+reqlen, sizeof(request)-1-reqlen, 0);
			if (nreqlen <= 0)
			{
				close(socket);
				socket = -1;
				return -1;
			}
			reqlen += nreqlen;
			reqpos = request;
			// retry
			continue;
		}
		++nextreq;
		// parse one line of request header
		int temp;
#ifdef HTTP_IF_MODIFIED_SUPPORT
		static const char* if_modified_since = "If-Modified-Since: ";
#endif // HTTP_IF_MODIFIED_SUPPORT
		if (sscanf(reqpos, "Content-Length: %d", &temp) == 1)
		{
			contentLength = temp;
		}
#ifdef HTTP_IF_MODIFIED_SUPPORT
		else if (!strncmp(reqpos, if_modified_since, strlen(if_modified_since))) 
		{
			const char* datepos = reqpos + strlen(if_modified_since);
			ptrdiff_t modlen = nextreq-datepos;
			delete[] modified_check;
			modified_check = new char[modlen];
			if (!modified_check)
				return -2;
			modlen = 0;
			while (*datepos != '\r' && *datepos != '\n')
				modified_check[modlen++] = *(datepos++);
			modified_check[modlen] = '\0';
		}
#endif // HTTP_IF_MODIFIED_SUPPORT

		reqpos = nextreq;
	}
	if (contentLength)
	{
		content = new char[contentLength+1];
		if (!content)
			return -2;
		reqlen = strlen(reqpos);
		if (reqlen > contentLength)
			reqlen = contentLength;
		memcpy(content, reqpos, reqlen);
		if (contentLength > reqlen)
		{
			int l = recv(socket, content+reqlen, contentLength-reqlen, 0);
			if (l < 0)
			{
				close(socket);
				socket = -1;
				return -1;
			}
			reqlen += l;
			if (reqlen != contentLength)
				return -1;
		}
		content[contentLength] = 0;
	}
	return 0;

}

/** Main webui class. Implements both the ui class and the command lsitener
 * so it can receive replies to kom commands. */
class KOMWebUi : public KOMUi, public KOMCommandListener, public TrayIconListener
{
public:
	KOMWebUi(int sock, int port) : serverSocket(sock), serverPort(port), sessionList(NULL), requestList(NULL), nextScheduledTimeout(0), icon(NULL)
#ifdef WIN32
		, win(NULL)
#elif defined(MACOSX)
		, quit(false)
#else // !WIN32 && !MACOSX
		, disp(NULL)
		, quit(false)
#endif // !WIN32 && !MACOSX
		, defaultPath(NULL)
	{}
	virtual ~KOMWebUi();
	/** Run the webui. */
	virtual void run();

	/** Reply received to a kom command. */
	virtual void onReply(int sid, const char* reply);

	virtual void onAction(int id);
	virtual void onDefaultAction();

	/** Check which tray icon should be displayed. */
	int getCurrentIconNum();

	/** Initialize the default path to the html resource files. */
	void initDefaultPath(const char* dataPath);
#ifdef WIN32
	LRESULT wndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
#endif // WIN32
private:
	struct KOMSessionList
	{
		KOMString url;
		KOMSession kom;
		time_t nextTimeout;
		bool dataSinceLastPing;
		KOMSessionList* next;
	};
	/** Send a file to the http request. */
	void sendFile(HTTPRequest* req);
	/** Handle a connection on the main server port. */
	void handleConnection();
	/** Find a kom session based on a requested url. */
	KOMSessionList* findSession(const char* url)
	{
		for (KOMSessionList* s = sessionList; s; s = s->next)
		{
			if (s->url == url)
			{
				return s;
			}
		}
		return NULL;
	}
	/** Start a new kom session with the given requested url. */
	KOMSessionList* startSession(const char* url)
	{
		KOMSessionList* s = new KOMSessionList();
		if (!s)
			return NULL;
		s->url.append(url),
		s->next = sessionList;
		s->nextTimeout = 0;
		s->dataSinceLastPing = false;
		sessionList = s;
		return s;
	}
	void handleDisconnected(KOMSessionList* es)
	{
		KOMSessionList** nextp = &sessionList;
		time_t now = time(NULL);
		for (KOMSessionList* s = sessionList; s; s = s->next)
		{
			if (s == es)
			{
				if (!s->kom.isConnected())
				{
					if (s->kom.isReconnecting())
					{
						if (!s->nextTimeout || s->nextTimeout > now+KOM_RECONNECT_TIMEOUT)
							rescheduleTimeout(s);
					}
					else
					{
						*nextp = s->next;
						delete s;
					}
				}
				return;
			}
			nextp = &s->next;
		}
	}
	void onTimeout(KOMSessionList* session)
	{
		if (session->kom.isReconnecting())
		{
			// Create a reconnect command
			if (!session->kom.isConnected())
				session->kom.createCommand(this, 0, "reconnect");
		}
		else if (session->kom.isConnected() && session->kom.isLoggedIn())
		{
			if (session->dataSinceLastPing)
				session->kom.pingServer();
			else
				// If no data was received since hte last ping was sent, the connection has probably been lost so try to reconnect
				session->kom.disconnect(true);
		}
		session->dataSinceLastPing = false;
		rescheduleTimeout(session);
		handleDisconnected(session);
	}
	void rescheduleTimeout(KOMSessionList* session)
	{
		time_t now = time(NULL);
		if (session && session->kom.isReconnecting())
			session->nextTimeout = now+KOM_RECONNECT_TIMEOUT;
		else if (session && session->kom.isLoggedIn())
			session->nextTimeout = now+KOM_KEEPALIVE_TIMEOUT;
		else if (session)
			session->nextTimeout = 0;
		time_t closestTimeout = 0;
		for (KOMSessionList* s = sessionList; s; s = s->next)
		{
			if (s && s->nextTimeout > 0)
			{
				if (!closestTimeout || closestTimeout > s->nextTimeout)
					closestTimeout = s->nextTimeout;
			}
		}
		nextScheduledTimeout = closestTimeout;
	}
	int serverSocket;
	int serverPort;
	KOMSessionList* sessionList;
	HTTPRequest* requestList;
	time_t nextScheduledTimeout;

	TrayIcon* icon;
#ifdef WIN32
	HWND win;
#elif defined(MACOSX)
	bool quit;
#else // !WIN32 && !MACOSX
	Display* disp;
	bool quit;
#endif // !WIN32 && !MACOSX
	char* defaultPath;
	VersionChecker verCheck;
};

KOMWebUi* g_webui = NULL;

KOMWebUi::~KOMWebUi()
{
	delete icon;

	// Shutting down
	close(serverSocket);

	while (sessionList)
	{
		KOMSessionList* ns = sessionList->next;
		delete sessionList;
		sessionList = ns;
	}
	while (requestList)
	{
		HTTPRequest* nr = requestList->next;
		delete requestList;
		requestList = nr;
	}

#ifdef WIN32
	if (win)
		CloseWindow(win);
#elif defined (MACOSX)
#else // !WIN32 && !MACOSX
	if (disp)
		XCloseDisplay(disp);
#endif // !WIN32 && !MACOSX

	delete[] defaultPath;
}

int KOMUi::Create(KOMUi** ui, const char* dataPath, int argc, char** argv)
{
	int port = DEFAULT_HTTP_PORT;
	for (int arg = 1; arg < argc; ++arg)
	{
		if ((!strcmp(argv[arg], "--port") || !strcmp(argv[arg], "-p")) && arg+1<argc)
		{
			port = atoi(argv[arg+1]);
			++arg;
		}
	}

	// Create a kom ui, in this case a webui
	// TODO: run as daemon
	int serverSocket;
	struct sockaddr_in servAddr;

	serverSocket = socket(PF_INET, SOCK_STREAM, 0);
	if (serverSocket < 0)
	{
		printf("Could not create socket\n");
		return -1;
	}
#ifndef WIN32 
	int one = 1;
	setsockopt(serverSocket, SOL_SOCKET, SO_REUSEADDR, &one, sizeof (int));
#endif
	memset(&servAddr, 0, sizeof(servAddr));
	servAddr.sin_family = AF_INET;
	servAddr.sin_addr.s_addr = htonl((INADDR_LOOPBACK));
	servAddr.sin_port = htons(port);
	if (bind(serverSocket, (struct sockaddr*)&servAddr, sizeof(servAddr))<0)
	{
		close(serverSocket);
		printf("Could not create socket\n");
		return -1;
	}
	if (listen(serverSocket, SOMAXCONN)<0)
	{
		close(serverSocket);
		printf("Could not create socket\n");
		return -1;
	}

	// Ready to accept connections
	*ui = new KOMWebUi(serverSocket, port);
	if (!*ui)
		return -1;
	g_webui = (KOMWebUi*)*ui;
	g_webui->initDefaultPath(dataPath);
	return 0;

}

int KOMWebUi::getCurrentIconNum()
{
	int iconNum = 3;
	for (KOMSessionList* s = sessionList; s; s = s->next)
	{
		if (s->kom.isLoggedIn())
		{
			if (iconNum == 3)
				iconNum = 0;
			if (s->kom.hasUnreadInstantMsg())
				iconNum = 2;
			else if (iconNum == 0 && s->kom.hasUnread())
				iconNum = 1;
		}
		// Always use the "not connected" icon if a session is reconnecting
		else if (s->kom.isReconnecting())
			return 3;
	}
	return iconNum;
}

void KOMWebUi::initDefaultPath(const char* dataPath)
{
	delete[] defaultPath;
	defaultPath = new char[6+strlen(dataPath)];
	if (defaultPath)
	{
		strcpy(defaultPath, dataPath);
		if (defaultPath[strlen(defaultPath)-1] != '/' && defaultPath[strlen(defaultPath)-1] != '\\')
			strcat(defaultPath, "/");
		strcat(defaultPath, "html");
	}
}

static const char* const FileMimeTypes[] = {
	".css", "text/css",
	".js", "application/x-javascript",
	".jpg", "image/jpeg",
	".jpeg", "image/jpeg",
	".gif", "image/gif",
	".png", "image/png",
	".ico", "image/vnd.microsoft.icon"
};

void KOMWebUi::sendFile(HTTPRequest* req)
{
	if (req->getRequestType() != HTTPRequest::REQ_GET && req->getRequestType() != HTTPRequest::REQ_HEAD)
	{
		req->replyError(HTTP_ERR_NOT_IMPLEMENTED, "This server has a very limited feature set and should only be used with the KOMclean software.");
		return;
	}
	char* path;
	const char* default_file_name = "index.html";
	path = new char[(defaultPath?strlen(defaultPath):0)+strlen(req->getURL())+strlen(default_file_name)+1];
	if (!path || *req->getURL() != '/')
	{
		req->replyError(HTTP_ERR_INTERNAL, "Out Of Memory");
		delete[] path;
		return;
	}
	if (defaultPath)
		strcpy(path, defaultPath);
	else
		*path = '\0';
	strcat(path, req->getURL());
	if (path[strlen(path)-1] == '/')
		strcat(path, default_file_name);
	struct stat modified;
	if (stat(path, &modified))
	{
		req->replyError(HTTP_ERR_NOT_FOUND, "The requested file was not found");
		delete[] path;
		return;
	}
#ifdef HTTP_IF_MODIFIED_SUPPORT
	if (req->getIfModifiedString())
	{
		char mstr[1024];
		strcpy(mstr, asctime(gmtime(&modified.st_mtime)));
		mstr[strlen(mstr)-1] = '\0';
		strcat(mstr, " GMT");
		if (!strcmp(mstr, req->getIfModifiedString()))
		{
			req->replyError(HTTP_ERR_NOT_MODIFIED, "");
			delete[] path;
			return;
		}
	}
#endif // HTTP_IF_MODIFIED_SUPPORT
	FILE* f = fopen(path, "rb");
	if (!f)
	{
		req->replyError(HTTP_ERR_NOT_FOUND, "The requested file was not found");
		delete[] path;
		return;
	}
	const char* type = "text/html";
	for (unsigned int i = 0; i < sizeof(FileMimeTypes)/sizeof(FileMimeTypes[0]); i += 2)
	{
		if (strlen(path) >= strlen(FileMimeTypes[i]) && !strcmp(path+strlen(path)-strlen(FileMimeTypes[i]), FileMimeTypes[i]))
		{
			type = FileMimeTypes[i+1];
			break;
		}
	}
	// Check file size
	fseek(f, 0, SEEK_END);
	int fileSize = ftell(f);
	fseek(f, 0, SEEK_SET);
	char *filebuf = NULL;
	if (req->getRequestType() == HTTPRequest::REQ_HEAD)
	{
		req->replyOK(NULL, fileSize, type, modified.st_mtime);
	}
	else
	{
		filebuf = new char[fileSize+1];
		if (filebuf)
		{
			int len = fread(filebuf, 1, fileSize, f);
			req->replyOK(filebuf, len, type, modified.st_mtime);
		}
		else
		{
			req->replyError(HTTP_ERR_INTERNAL, "Out Of Memory");
		}
	}
	fclose(f);
	delete[] filebuf;
	delete[] path;
}

void KOMWebUi::handleConnection()
{
	int sessionSock = accept(serverSocket, NULL, NULL);
#ifdef WIN32
	// Disable async select and set back to blocking mode
	unsigned long blockdis = 0;
	WSAAsyncSelect(sessionSock, win, WM_SOCKET, 0);
	ioctlsocket(sessionSock, FIONBIO, &blockdis);
#endif // WIN32
	HTTPRequest* req = new HTTPRequest(sessionSock);
	if (!req)
	{
		close(sessionSock);
		return;
	}
	if (req->Construct() < 0)
	{
		req->replyError(HTTP_ERR_BAD_REQUEST, "Broken request or server error");
		delete req;
		return;
	}
	if (req->getRequestType() != HTTPRequest::REQ_POST)
	{
		sendFile(req);
		delete req;
		return;
	}
	if (req->getContentLength()>=7 && !strncmp(req->getContent(), "version", 7))
	{
		req->replyOK("OK" KOM_GROUP_SEPARATOR KOMCLEAN_VERSION_STRING);
		delete req;
		return;
	}
	if (req->getContentLength()>=15 && !strncmp(req->getContent(), "upgrade_message", 15))
	{
		KOMString rep;
		rep.appendFormat("OK" KOM_GROUP_SEPARATOR "%s", verCheck.getUpgradeMessage());
		req->replyOK(rep.c_string());
		delete req;
		return;
	}
	KOMSessionList* session = findSession(req->getURL()+1);
	if (session && !session->kom.isConnected())
	{
		handleDisconnected(session);
		session = findSession(req->getURL()+1);
	}
	// Never send error - reconnecting as response to logout
	if (session && session->kom.isReconnecting() &&
		(req->getContentLength()<6 || strncmp(req->getContent(), "logout", 6)))
	{
		req->replyOK("ERROR" KOM_GROUP_SEPARATOR "Reconnecting");
		delete req;
		return;
	}
	if (!session)
	{
		if (req->getContentLength()>=6 && !strncmp(req->getContent(), "login ", 6))
		{
			session = startSession(req->getURL()+1);
		}
		if (!session)
		{
			req->replyOK("ERROR" KOM_GROUP_SEPARATOR "Not connected");
			delete req;
			return;
		}
	}
	// Add to request list before creating command in case the reply is instant
	req->next = requestList;
	requestList = req;
	req->stripLineFeed();
	if (session->kom.createCommand(this, req->requestNum, req->getContent()))
	{
		handleDisconnected(session);
		req->replyError(HTTP_ERR_INTERNAL, "Out Of Memory");
		requestList = requestList->next;
		delete req;
	}
}

#ifdef WIN32
static LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return g_webui->wndProc(hwnd, uMsg, wParam, lParam);
}

LRESULT KOMWebUi::wndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if (((Win32TrayIcon*)icon)->handleEvent(uMsg, wParam, lParam))
		return 0;
	if (uMsg == WM_SOCKET)
	{
		if (wParam == serverSocket)
		{
			handleConnection();
		}
		else if (verCheck.getCurrentSocket() >= 0 && wParam == verCheck.getCurrentSocket())
		{
			int param = WSAGETSELECTEVENT(lParam);
			if (WSAGETSELECTEVENT(lParam) == FD_CLOSE)
				verCheck.disconnect();
			else if (WSAGETSELECTEVENT(lParam) == FD_CONNECT)
			{
				if (WSAGETSELECTERROR(lParam)!=0)
					verCheck.disconnect();
			}
			else
				verCheck.processData();
		}
		else
		{
			time_t oldTimeout = nextScheduledTimeout;
			time_t now = time(NULL);
			// Check all others
			KOMSessionList* nexts = NULL;
			for (KOMSessionList* s = sessionList; s; s = nexts)
			{
				nexts = s->next;
				if (s->kom.isConnected() && wParam == s->kom.socket)
				{
					if(WSAGETSELECTEVENT(lParam) == FD_CLOSE)
						s->kom.disconnect(true);
					else //if (WSAGETSELECTEVENT(lParam) == FD_READ)
					{
						s->kom.processData();
						s->dataSinceLastPing = true;
						rescheduleTimeout(s);
					}
					// Check if the session has died
					handleDisconnected(s);
				}
			}
			if (nextScheduledTimeout && nextScheduledTimeout != oldTimeout)
			{
				SetTimer(win, 1, (UINT)(nextScheduledTimeout-now)*1000, NULL);
			}
			else if (!nextScheduledTimeout && oldTimeout)
			{
				KillTimer(win, 1);
			}

		}
		icon->setActiveIcon(getCurrentIconNum());
		return 0;
	}
	else if (uMsg == WM_TIMER)
	{
		time_t now = time(NULL);
		for (KOMSessionList* s = sessionList; s;)
		{
			KOMSessionList* ns = s->next;
			if (s->nextTimeout <= now)
			{
				onTimeout(s);
			}
			s = ns;
		}
		rescheduleTimeout(NULL);
		if (nextScheduledTimeout)
			SetTimer(win, 1, (UINT)(nextScheduledTimeout-now)*1000, NULL);
		return 0;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}
#endif // WIN32

void KOMWebUi::run()
{
#ifdef WIN32
	WNDCLASS wclass;
	memset(&wclass, 0, sizeof(wclass));
	wclass.lpfnWndProc = WndProc;
	wclass.lpszClassName = L"KOMclean";
	RegisterClass(&wclass);
	win = CreateWindow(L"KOMclean", L"KOMclean", WS_BORDER, 
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
		HWND_MESSAGE, NULL, NULL, NULL);
	g_win = win;

	icon = new Win32TrayIcon(win, this);
	icon->addIcon(MAKEINTRESOURCE(IDI_ICONNORMAL));
	icon->addIcon(MAKEINTRESOURCE(IDI_ICONUNREAD));
	icon->addIcon(MAKEINTRESOURCE(IDI_ICONUNREADMSG));
	icon->addIcon(MAKEINTRESOURCE(IDI_ICONNOTCONNECTED));

	icon->addPopupMenuItem("Open WebUI", 0, true);
	icon->addPopupMenuItem("Exit", -1, true);

	icon->show(1);
	icon->setActiveIcon(3);
	WSAAsyncSelect(serverSocket, win, WM_SOCKET, FD_ACCEPT);
	verCheck.sendRequest();
	MSG msg;
	while (GetMessage(&msg, win, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return;
#else // !WIN32
#ifdef MACOSX
	icon = new MacTrayIcon(this);
	icon->addIcon((void*)"icon_mac_normal");
	icon->addIcon((void*)"icon_mac_unread");
	icon->addIcon((void*)"icon_mac_unread_msg");
	icon->addIcon((void*)"icon_mac_notconnected");
	icon->addPopupMenuItem("Open WebUI", 0, true);
	icon->addPopupMenuItem("Quit", -1, true);
	icon->show(1);
	int systemsock = ((MacTrayIcon*)icon)->getSystemSocket();
#else // !MACOSX
	disp = XOpenDisplay(NULL);
	int systemsock = -1;
	if (disp)
	{
		icon = new X11TrayIcon(disp, this);
		icon->addIcon((void*)icon_normal_xpm);
		icon->addIcon((void*)icon_unread_xpm);
		icon->addIcon((void*)icon_unread_msg_xpm);
		icon->addIcon((void*)icon_notconnected_xpm);
		icon->addPopupMenuItem("Open WebUI", 0, true);
		icon->addPopupMenuItem("Exit", -1, true);
		icon->show(1);
		systemsock = ConnectionNumber(disp);
	}
#endif // !MACOSX
	icon->setActiveIcon(3);
	verCheck.sendRequest();
	time_t now = time(NULL);
	while (!quit)
	{
		int maxfd = serverSocket;
		fd_set readSet;
		fd_set writeSet;
		int numReady;
		FD_ZERO(&readSet);
		FD_ZERO(&writeSet);
		FD_SET(serverSocket, &readSet);
		if (systemsock >= 0)
		{
			FD_SET(systemsock, &readSet);
			if (systemsock > maxfd)
				maxfd = systemsock;
		}
		if (verCheck.getCurrentSocket() >= 0)
		{
			FD_SET(verCheck.getCurrentSocket(), &readSet);
			FD_SET(verCheck.getCurrentSocket(), &writeSet);
			if (verCheck.getCurrentSocket() > maxfd)
				maxfd = verCheck.getCurrentSocket();
		}
		// include the socket of all running sessions
		for (KOMSessionList* s = sessionList; s; s = s->next)
		{
			if (s->kom.isConnected())
			{
				FD_SET(s->kom.socket, &readSet);
				if (s->kom.socket > maxfd)
					maxfd = s->kom.socket;
			}
		}
		struct timeval* timeout = NULL;
		struct timeval _timeout;
		if (nextScheduledTimeout)
		{
			_timeout.tv_sec = nextScheduledTimeout-now;
			_timeout.tv_usec = 0;
			timeout = &_timeout;
		}
		numReady = select(maxfd+1, &readSet, &writeSet, NULL, timeout);
		if (numReady < 0)
		{
			// unexpected error!
			//fprintf(stderr, "Error waiting for socket\n");
			return;
		}
		if (numReady > 0)
		{
			if (FD_ISSET(serverSocket, &readSet))
			{
				handleConnection();
				// Someone connected
				--numReady;
			}
			if (systemsock >= 0 && FD_ISSET(systemsock, &readSet))
			{
#ifdef MACOSX
				void* icncmd;
				if (read(systemsock, &icncmd, sizeof(void*))==sizeof(void*))
				{
					((MacTrayIcon*)icon)->handleEvent(icncmd);
				}
#else
				XEvent e;
				while (XPending(disp))
				{
					XNextEvent(disp, &e);
					((X11TrayIcon*)icon)->handleEvent(e);
				}
#endif // !MACOSX
				--numReady;
			}
			if (verCheck.getCurrentSocket() >= 0)
			{
				if (FD_ISSET(verCheck.getCurrentSocket(), &readSet))
				{
					verCheck.processData();
					--numReady;
				}
				if (verCheck.getCurrentSocket() >= 0 && FD_ISSET(verCheck.getCurrentSocket(), &writeSet))
				{
					verCheck.processData();
					--numReady;
				}
			}
			for (KOMSessionList* s = sessionList; s && numReady; s = s->next)
			{
				if (s->kom.isConnected() && FD_ISSET(s->kom.socket, &readSet))
				{
					s->kom.processData();
					s->dataSinceLastPing = true;
					rescheduleTimeout(s);
					// Check if the session has died
					handleDisconnected(s);
					--numReady;
				}
			}
			icon->setActiveIcon(getCurrentIconNum());
		}
		now = time(NULL);
		for (KOMSessionList* s = sessionList; s;)
		{
			KOMSessionList* ns = s->next;
			if (s->nextTimeout <= now)
			{
				onTimeout(s);
			}
			s = ns;
		}
	}
#endif // !WIN32
}

void KOMWebUi::onReply(int sid, const char* reply)
{
	// Received a reply from kom, find which request it belongs to
	HTTPRequest** nextp = &requestList;
	for (HTTPRequest* req = requestList; req; req = req->next)
	{
		if (req->requestNum == sid)
		{
			// Found it, reply to it
			req->replyOK(reply);
			// And delete it from the list
			*nextp = req->next;
			delete req;
			return;
		}
		nextp = &req->next;
	}
}

void KOMWebUi::onAction(int id)
{
	if (id < 0)
	{
#ifdef WIN32
		PostQuitMessage(0);
#else // !WIN32
		quit = true;
#endif // !WIN32
	}
	else if (id == 0)
		onDefaultAction();
}

#ifdef WIN32
static inline void SystemOpenURL(const char* url)
{
	ShellExecuteA(NULL, "open", url, NULL, NULL, SW_SHOWNORMAL);
}
#elif defined(MACOSX)
// Defined in objective c (mactrayicon.h/mm)
#else // !WIN32 && !MACOSX
static void SystemOpenURL(const char* url)
{
	KOMString cmd;
	if (getenv("GNOME_DESKTOP_SESSION_ID"))
		cmd.append("gnome-open ");
	else if (getenv("KDE_FULL_SESSION"))
		cmd.append("kfmclient openURL ");
	else if (getenv("BROWSER"))
	{
		cmd.append(getenv("BROWSER"));
		cmd.append(" ");
	}
	else if (!system("which opera"))
		cmd.append("opera ");
	else if (!system("which firefox"))
		cmd.append("firefox ");
	else
	{
		printf("No browser found, please open %s\n", url);
		return;
	}
	cmd.append(url);
	cmd.append(" &");
	system(cmd.c_string());
}
#endif // !WIN32 && !MACOSX

void KOMWebUi::onDefaultAction()
{
	KOMSessionList* nexts = NULL;
	for (KOMSessionList* s = sessionList; s; s = nexts)
	{
		nexts = s->next;
		handleDisconnected(s);
	}
	if (sessionList)
	{
		for (KOMSessionList* s = sessionList; s; s = s->next)
		{
			KOMString cmd;
			cmd.appendFormat("http://127.0.0.1:%d/index.html#", serverPort);
			cmd.append(s->url.c_string());
			SystemOpenURL(cmd.c_string());
		}
	}
	else
	{
		KOMString cmd;
		cmd.appendFormat("http://127.0.0.1:%d/login.html", serverPort);
		SystemOpenURL(cmd.c_string());
	}
}

